/* $Id: aomidctl.c 322 2010-01-14 06:37:17Z ken $ */
/*
 * Copyright (c) 2009 KAMADA Ken'ichi.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <arpa/inet.h>
#include <errno.h>
#include <inttypes.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "missing/missing.h"

#include "pathnames.h"

#include "aomidctl.h"

static void ctl_reload(void);
static void ctl_stat(void);

static int connect_aomidctl(void);
static void *send_recv(int s, size_t *replenp, void *cmdbuf, size_t cmdlen);
static void errmsg(const char *fmt, ...);
static void usage(void);

static const char *ctl_file = CONTROL_FILE;

int
main(int argc, char *argv[])
{
	int c;

	setprogname(argv[0]);
	while ((c = getopt(argc, argv, "c:h")) != -1) {
		switch (c) {
		case 'c':
			ctl_file = optarg;
			break;
		case 'h':
			usage();
			/* NOTREACHED */
		default:
			usage();
			/* NOTREACHED */
		}
	}
	argc -= optind;
	argv += optind;

	if (argc == 1 && strcmp(argv[0], "reload") == 0)
		ctl_reload();
	else if (argc == 1 && strcmp(argv[0], "stat") == 0)
		ctl_stat();
	else
		usage();
	exit(0);
}

static void
ctl_reload(void)
{
	struct ctl_msg_generic req, *rep;
	size_t len;
	int s;

	if ((s = connect_aomidctl()) == -1)
		exit(1);
	memset(&req, 0, sizeof(req));
	req.h.version = 0;
	req.h.command = CTL_RELOAD;
	req.h.reserved = 0;
	rep = send_recv(s, &len, &req, sizeof(req));
	if (len < sizeof(*rep)) {
		errmsg("reply too short");
		exit(1);
	}
	if (rep->h.version != 0) {
		errmsg("control message version mismatch");
		exit(1);
	}
	if (rep->h.command == CTL_EBADCMD) {
		errmsg("RELOAD not supported by the server");
		exit(1);
	}
	if (rep->h.command != CTL_OK) {
		errmsg("unknown error from the server");
		exit(1);
	}
	free(rep);
	close(s);
}

static void
ctl_stat(void)
{
	struct ctl_msg_generic req;
	struct ctl_msg_rep_stat *rep;
	size_t len;
	int s;

	if ((s = connect_aomidctl()) == -1)
		exit(1);
	memset(&req, 0, sizeof(req));
	req.h.version = 0;
	req.h.command = CTL_STAT;
	req.h.reserved = 0;
	rep = send_recv(s, &len, &req, sizeof(req));
	if (len < sizeof(*rep)) {
		errmsg("reply too short");
		exit(1);
	}
	if (len > sizeof(*rep))
		errmsg("unknown extra entries from the server are ignored");
	if (rep->h.version != 0) {
		errmsg("control message version mismatch");
		exit(1);
	}
	if (rep->h.command == CTL_EBADCMD) {
		errmsg("RELOAD not supported by the server");
		exit(1);
	}
	if (rep->h.command != CTL_OK) {
		errmsg("unknown error from the server");
		exit(1);
	}
	free(rep);
	close(s);

	printf("Incoming requests:\n"
	    "  query=%" PRIu32 ", success=%" PRIu32 ", nodata=%" PRIu32
	    ", nxdomain=%" PRIu32 ", servfail=%" PRIu32 "\n",
	    ntohl(rep->d_query), ntohl(rep->d_success), ntohl(rep->d_nodata),
	    ntohl(rep->d_nxdomain), ntohl(rep->d_servfail));

	printf("Memory allocation:\n");
	printf("  pbuf requested=%" PRIu32 ", released=%" PRIu32
	    ", failed=%" PRIu32 ", max=%" PRIu32 "\n",
	    ntohl(rep->obj_pbuf_req), ntohl(rep->obj_pbuf_rel),
	    ntohl(rep->obj_pbuf_fail), htonl(rep->obj_pbuf_max));
	printf("  udpsock requested=%" PRIu32 ", released=%" PRIu32
	    ", failed=%" PRIu32 ", max=%" PRIu32 "\n",
	    ntohl(rep->obj_udpsock_req), ntohl(rep->obj_udpsock_rel),
	    ntohl(rep->obj_udpsock_fail), htonl(rep->obj_udpsock_max));
	printf("  tcpsock requested=%" PRIu32 ", released=%" PRIu32
	    ", failed=%" PRIu32 ", max=%" PRIu32 "\n",
	    ntohl(rep->obj_tcpsock_req), ntohl(rep->obj_tcpsock_rel),
	    ntohl(rep->obj_tcpsock_fail), htonl(rep->obj_tcpsock_max));
	printf("  qtask requested=%" PRIu32 ", released=%" PRIu32
	    ", failed=%" PRIu32 ", max=%" PRIu32 "\n",
	    ntohl(rep->obj_qtask_req), ntohl(rep->obj_qtask_rel),
	    ntohl(rep->obj_qtask_fail), htonl(rep->obj_qtask_max));
}

static int
connect_aomidctl(void)
{
	struct sockaddr_un su;
	socklen_t sulen;
	int s;

	memset(&su, 0, sizeof(su));
	su.sun_family = AF_LOCAL;
	if (strlen(ctl_file) >= sizeof(su.sun_path)) {
		errmsg("%s: path is too long for sockaddr_un", ctl_file);
		return -1;
	}
	strcpy(su.sun_path, ctl_file);
	sulen = su.sun_path - (char *)&su + strlen(su.sun_path);
#ifdef HAVE_SA_LEN
	su.sun_len = sulen;
#endif

	if ((s = socket(PF_LOCAL, SOCK_STREAM, 0)) == -1) {
		errmsg("socket: %s", strerror(errno));
		return -1;
	}
	if (connect(s, (struct sockaddr *)&su, sulen) == -1) {
		errmsg("connect: %s: %s", ctl_file, strerror(errno));
		close(s);
		return -1;
	}
	return s;
}

static void *
send_recv(int s, size_t *replenp, void *cmdbuf, size_t cmdlen)
{
	uint16_t wirelen;
	char *buf, *p, *end;
	size_t len;
	ssize_t ret;

	/* cmdlen is much smaller than 2^16. */
	wirelen = htons(cmdlen);
	p = (char *)&wirelen;
	end = p + sizeof(wirelen);
	while (p < end) {
		if ((ret = send(s, p, end - p, 0)) == -1)
			goto send_fail;
		p += ret;
	}
	p = cmdbuf;
	end = p + cmdlen;
	while (p < end) {
		if ((ret = send(s, p, end - p, 0)) == -1)
			goto send_fail;
		p += ret;
	}

	p = (char *)&wirelen;
	end = p + sizeof(wirelen);
	while (p < end) {
		if ((ret = recv(s, p, end - p, 0)) == -1)
			goto recv_fail;
		if (ret == 0)
			goto recv_eof;
		p += ret;
	}
	len = ntohs(wirelen);
	if ((buf = (char *)malloc(len)) == NULL) {
		errmsg("%s", strerror(errno));
		exit(1);
	}
	p = buf;
	end = p + len;
	while (p < end) {
		if ((ret = recv(s, p, end - p, 0)) == -1)
			goto recv_fail;
		if (ret == 0)
			goto recv_eof;
		p += ret;
	}

	*replenp = len;
	return buf;

send_fail:
	errmsg("send: %s", strerror(errno));
	exit(1);
recv_fail:
	errmsg("recv: %s", strerror(errno));
	exit(1);
recv_eof:
	errmsg("unexpected close by the server");
	exit(1);
}

static void
errmsg(const char *fmt, ...)
{
	char buf[1024];
	va_list ap;

	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);
	va_end(ap);
	fprintf(stderr, "%s: %s\n", getprogname(), buf);
}

static void
usage(void)
{
	printf("usage: %s [-c socket] command [arg ...]\n"
	    "  reload\n"
	    "  stat\n", getprogname());
	exit(1);
}
