/* $Id: cache.c 216 2007-11-15 13:25:38Z ken $ */
/*
 * Copyright (c) 2006 KAMADA Ken'ichi.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#include <stddef.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <time.h>

#include "var.h"
#include "util.h"
#include "stats.h"
#include "cache.h"


struct alignouter {
	char c;
	union {
		void *p;
		void (*fp)(void);
		intmax_t i;
	} aligninner;
};
#define ALIGN(p) ALIGN_N(p, offsetof(struct alignouter, aligninner))
#define ALIGN_N(p, n) (char *)(((intptr_t)(p) + ((n) - 1)) & ~((n) - 1))

static void *c_ext(size_t len);
static void *c_ext_align(size_t len);
static char *next_space(size_t len, int align);
static void reclaim(char *from, size_t len);

time_t curtime;

static char *cache, *cacheend;
static struct cache *oldest, *latest;
static char *front;
static unsigned int generation;

int
c_init(size_t size)
{
	if (size < MIN_CACHE_SIZE) {
		LWARN("cache size too small");
		size = MIN_CACHE_SIZE;
	}

	if ((cache = (char *)malloc(size)) == NULL) {
		LERR("%m");
		return -1;
	}
	cacheend = cache + size;

	oldest = latest = NULL;
	front = cache;
	generation = 0;

	return 0;
}

void *
c_new(size_t len)
{
	struct cache *c;
	char *ext;

	if (oldest == NULL) {
		/* first invocation of c_new */
		oldest = latest = (struct cache *)cache;
		ext = cache;
	} else {
		ext = next_space(len, 1);
		reclaim(ext, len);
	}

	/*
	 * In the first invocation, latest == c here; so c->next must be
	 * assigned after latest->next.
	 */
	c = (struct cache *)ext;
	latest->next = c;
	/* XXX <*c> will be reinitialized by the caller. */
	c->next = NULL;
	c->eol = 0;

	latest = c;
	front = ext + len;
	return ext;
}

void *
c_extcpy(const void *ptr, size_t len)
{
	void *dst;

	dst = c_ext(len);
	return memcpy(dst, ptr, len);
}

static void *
c_ext(size_t len)
{
	char *ext;

	ext = next_space(len, 0);
	reclaim(ext, len);
	front = ext + len;
	return ext;
}

void *
c_extcpy_align(const void *ptr, size_t len)
{
	void *dst;

	dst = c_ext_align(len);
	return memcpy(dst, ptr, len);
}

static void *
c_ext_align(size_t len)
{
	char *ext;

	ext = next_space(len, 1);
	reclaim(ext, len);
	front = ext + len;
	return ext;
}

static char *
next_space(size_t len, int align)
{
	char *ext;

	if (cacheend - cache < len) {
		LERR("fatal: cache too small");
		abort();
	}

	ext = align ? ALIGN(front) : front;
	if (ext <= cacheend - len)
		return ext;
	generation++;
	LDEBUG2("generation %u", generation);
	return cache;
}

static void
reclaim(char *from, size_t len)
{
	struct cache *next;
	char *to;

	if (oldest == NULL) {
		LERR("SANITY: reclaim() before c_new()");
		abort();
	}

	to = from + len;
	while ((from <= (char *)oldest && (char *)oldest < to) ||
	    (from < front && front <= (char *)oldest)) {
		next = oldest->next;
		if (next == NULL) {
			LERR("fatal: too large data to cache");
			abort();
		}
		if (oldest->destructor == 0)
			STAT_CACHE_WASTED();
		else if (oldest->destructor == &cache_expired)
			STAT_CACHE_EXPIRED();
		else if (oldest->destructor == &cache_replaced)
			STAT_CACHE_REPLACED();
		else {
			if (oldest->eol < curtime)	/* include eol == 0 */
				STAT_CACHE_EXPIRED();
			else
				STAT_CACHE_RECLAIMED();
			(*oldest->destructor)(oldest);
		}
		oldest = next;
	}
}

void
c_checkttl(struct cache *c, uint32_t ttl)
{
	time_t eol;

	eol = curtime + ttl;
	if (eol != c->eol)
		LDEBUG2("TTL mismatch");
	if (eol < c->eol)
		c->eol = eol;
}

void
c_bumptime(void)
{
	curtime = time(NULL);
}


#ifdef NOTUSED
void
c_refer(struct cacheref *cr, struct cache *c)
{
	cr->generation = generation;
	cr->c = c;
}

int
c_isvalid(struct cacheref *cr)
{
	if (cr->c == NULL) {
		LERR("SANITY: NULL reference");
		return 0;
	}

	if (cr->generation == generation)
		return 1;
	if (cr->generation + 1 == generation && oldest <= cr->c)
		return 1;
	return 0;
}
#endif


void
cache_expired(struct cache *c)
{
	/* empty */
}

void
cache_replaced(struct cache *c)
{
	/* empty */
}


void
info_cache(void)
{
	LINFO("Cache Info:");
	LINFO(" size=%td, front=%td, generation=%d",
	    cacheend - cache, front - cache, generation);
	LINFO(" rear eol=%lu, front eol=%lu",
	    (unsigned long)(oldest == NULL ? 0 : oldest->eol),
	    (unsigned long)(latest == NULL ? 0 : latest->eol));
}
