/* $Id: ctlmsg.c 335 2011-05-22 16:54:06Z ken $ */
/*
 * Copyright (c) 2009 KAMADA Ken'ichi.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <arpa/inet.h>
#include <errno.h>
#include <stdint.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>

#include "version.h"
#include "var.h"			/* req'd by format.h */
#include "util.h"
#include "stats.h"
#include "format.h"			/* pbuf */
#include "net.h"

#include "aomidctl.h"

static size_t proc_stat(struct pbuf *pb, char *buf, size_t len);

int
open_aomidctl(const char *path)
{
	struct stat sb;
	struct sockaddr_un su;
	socklen_t sulen;
	int s;

	memset(&su, 0, sizeof(su));
	su.sun_family = AF_LOCAL;
	if (strlen(path) >= sizeof(su.sun_path)) {
		LERR("%s: path is too long for sockaddr_un", path);
		return -1;
	}
	strcpy(su.sun_path, path);
	sulen = su.sun_path - (char *)&su + strlen(su.sun_path);
#ifdef HAVE_SA_LEN
	su.sun_len = sulen;
#endif

	/* Remove a stale socket. */
	if (stat(path, &sb) == 0 && S_ISSOCK(sb.st_mode)) {
		/* Ensure that no one is listening on it. */
		if ((s = socket(PF_LOCAL, SOCK_STREAM, 0)) == -1) {
			LERR("socket: %m");
			return -1;
		}
		if (connect(s, (struct sockaddr *)&su, sulen) == 0) {
			LERR("active socket at %s", path);
			close(s);
			return -1;
		}
		if (errno != ECONNREFUSED) {
			LERR("connect: %s: %m", path);
			close(s);
			return -1;
		}
		close(s);
		LDEBUG("removing a stale socket at %s", path);
		(void)unlink(path);
	}

	/* Open a new control socket. */
	if ((s = socket(PF_LOCAL, SOCK_STREAM, 0)) == -1) {
		LERR("socket: %m");
		return -1;
	}
	if (bind(s, (struct sockaddr *)&su, sulen) == -1) {
		LERR("bind: %s: %m", path);
		close(s);
		return -1;
	}
	if (chmod(path, S_IRUSR | S_IWUSR) == -1) {
		LERR("chmod: %s: %m", path);
		close(s);
		return -1;
	}
	if (listen(s, 1) == -1) {
		LERR("listen: %m");
		close(s);
		return -1;
	}
	LDEBUG("control socket at %s", path);
	return s;
}

/*
 * Return value:
 *  >0  Length of reply.
 *   0  Error.
 */
size_t
upcall_ctl_msg(struct transport *t, char *buf, size_t len)
{
	struct ctl_msg_header *h;
	struct pbuf *pb;

	if (len < sizeof(*h)) {
		LWARN("control message too short");
		return 0;
	}

	pb = get_pbuf_to_reply(t);
	if (pb->size < sizeof(*h)) {
		LERR("SANITY: message buffer too short");
		return 0;
	}
	h = (struct ctl_msg_header *)PBUF_DATA(pb);
	*h = *(struct ctl_msg_header *)buf;

	if (h->version != 0) {
		LDEBUG("control message version %d unsupported", h->version);
		h->version = 0;
		h->command = CTL_EBADCMD;
		return sizeof(*h);
	}
	h->reserved = 0;

	switch (h->command) {
	case CTL_RELOAD:
		inject_sighup();
		h->command = CTL_OK;
		return sizeof(*h);
	case CTL_STAT:
		return proc_stat(pb, buf, len);
	default:
		h->command = CTL_EBADCMD;
		return sizeof(*h);
	}
	/* NOTREACHED */
}

static size_t
proc_stat(struct pbuf *pb, char *buf, size_t len)
{
	struct ctl_msg_rep_stat *rep;

	if (pb->size < sizeof(*rep)) {
		LERR("SANITY: message buffer too short");
		return 0;
	}
	rep = (struct ctl_msg_rep_stat *)PBUF_DATA(pb);
	rep->h.command = CTL_OK;
	rep->d_query = htonl(stats.d_query);
	rep->d_success = htonl(stats.d_success);
	rep->d_nodata = htonl(stats.d_nodata);
	rep->d_nxdomain = htonl(stats.d_nxdomain);
	rep->d_servfail = htonl(stats.d_servfail);
	rep->obj_pbuf_req = htonl(stats.obj_pbuf_alloc + stats.obj_pbuf_fail);
	rep->obj_pbuf_rel = htonl(stats.obj_pbuf_alloc - stats.obj_pbuf);
	rep->obj_pbuf_fail = htonl(stats.obj_pbuf_fail);
	rep->obj_pbuf_max = htonl(stats.obj_pbuf_max);
	rep->obj_udpsock_req = htonl(stats.obj_udpsock_alloc + stats.obj_udpsock_fail);
	rep->obj_udpsock_rel = htonl(stats.obj_udpsock_alloc - stats.obj_udpsock);
	rep->obj_udpsock_fail = htonl(stats.obj_udpsock_fail);
	rep->obj_udpsock_max = htonl(stats.obj_udpsock_max);
	rep->obj_tcpsock_req = htonl(stats.obj_tcpsock_alloc + stats.obj_tcpsock_fail);
	rep->obj_tcpsock_rel = htonl(stats.obj_tcpsock_alloc - stats.obj_tcpsock);
	rep->obj_tcpsock_fail = htonl(stats.obj_tcpsock_fail);
	rep->obj_tcpsock_max = htonl(stats.obj_tcpsock_max);
	rep->obj_qtask_req = htonl(stats.obj_qtask_alloc + stats.obj_qtask_fail);
	rep->obj_qtask_rel = htonl(stats.obj_qtask_alloc - stats.obj_qtask);
	rep->obj_qtask_fail = htonl(stats.obj_qtask_fail);
	rep->obj_qtask_max = htonl(stats.obj_qtask_max);
	return sizeof(*rep);
}
