/* $Id: db.c 256 2008-04-21 13:54:49Z ken $ */
/*
 * Copyright (c) 2006 KAMADA Ken'ichi.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <time.h>			/* XXX calc_rtt */

#include "missing/missing.h"

#include "var.h"
#include "splaytree.h"
#include "util.h"
#include "format.h"
#include "net.h"
#include "cache.h"
#include "db.h"

#define UNKNOWN_RTT	30	/* ms */

static int hintcmp(const struct splay_node *n1, const struct splay_node *n2);
static void unlink_rrset(struct cache *c);
static int rrsetcmp(const struct splay_node *n1, const struct splay_node *n2);
static void unlink_authority(struct cache *c);
static int authoritycmp(const struct splay_node *n1,
    const struct splay_node *n2);
static void unlink_glue(struct cache *c);
static int gluecmp(const struct splay_node *n1, const struct splay_node *n2);
static unsigned int calc_rtt(const struct timespec *start);
static struct servcap *get_or_new_servcap(const struct sockaddr *sa);
static void unlink_servcap(struct cache *c);
static int servcapcmp(const struct splay_node *n1,
    const struct splay_node *n2);

static struct splay_descr authority_descr = { .cmp = &authoritycmp };
static struct splay_descr rrset_descr = { .cmp = &rrsetcmp };
static struct splay_descr glue_descr = { .cmp = &gluecmp };
static struct splay_descr servcap_descr = { .cmp = &servcapcmp };


struct hint *
alloc_hint(const char *zone, unsigned int class)
{
	struct hint *h, tmpl;
	struct splay_node *found;
	int new_sac;

	if (conf_load.hint == NULL) {
		if ((conf_load.hint = (struct splay_descr *)
		    malloc(sizeof(*conf_load.hint))) == NULL) {
			LERR("%m");
			return NULL;
		}
		*conf_load.hint = (struct splay_descr) { .cmp = &hintcmp };
	}

	tmpl.zone = zone;
	tmpl.class = class;
	found = splay_find(conf_load.hint, &tmpl.tree);
	if (found != NULL) {
		h = MBR2STRUCT(found, struct hint, tree);
		if (h->sac < h->max_sac)
			return h;
		/* resize if necessary */
		new_sac = h->max_sac + h->max_sac / 2 + 1;
		if ((h = (struct hint *)
		    realloc(h, offsetof(struct hint, sav[new_sac]))) == NULL) {
			LERR("%m");
			return NULL;
		}
		h->max_sac = new_sac;
		/* XXX depend on the internal of splaytree */
		conf_load.hint->root = &h->tree;
		return h;
	}
	/* allocate new hint */
	if ((h = (struct hint *)
	    malloc(offsetof(struct hint, sav[1]))) == NULL) {
		LERR("%m");
		return NULL;
	}
	if ((h->zone = domdup(zone)) == NULL) {
		LERR("%m");
		free(h);
		return NULL;
	}
	h->class = class;
	h->local = 0;
	h->local_soa = NULL;
	h->max_sac = 1;
	h->sac = 0;
	(void)splay_replace(conf_load.hint, &h->tree);
	return h;
}

int
fixate_hints(int cancel)
{
	struct splay_descr *tofree;
	struct hint *h;
	int i;

	if (cancel)
		tofree = conf_load.hint;
	else {
		tofree = conf_run.hint;
		conf_run.hint = conf_load.hint;
	}
	conf_load.hint = NULL;

	if (tofree != NULL) {
		while (tofree->root != NULL) {
			h = MBR2STRUCT(tofree->root, struct hint, tree);
			(void)splay_remove(tofree, tofree->root);
			free(UNCONST(char *, h->zone));
			free(h->local_soa);
			for (i = 0; i < h->sac; i++)
				free(h->sav[i]);
			free(h);
		}
		free(tofree);
	}
	return 0;
}

struct hint *
get_hint(const char **superzone, const char *zone, unsigned int class)
{
	struct hint *h, tmpl;
	int ret;

	*superzone = zone + domlen(zone);
	tmpl.zone = zone;
	tmpl.class = class;

	/* conf_run.hint must be non-empty due to built-in hints. */
	for (;;) {
		ret = splay_find_leq(conf_run.hint, &tmpl.tree);
		h = MBR2STRUCT(conf_run.hint->root, struct hint, tree);
		if (ret == 0) {		/* if found */
			*superzone = tmpl.zone;
			return h;
		} else if (ret > 0)		/* only greater was found */
			return NULL;
		else if (h->class != class)  /* underflowed to another class */
			return NULL;
		tmpl.zone = domcommon(h->zone, tmpl.zone);
	}
}

static int
hintcmp(const struct splay_node *n1, const struct splay_node *n2)
{
	const struct hint *h1, *h2;
	int diff;

	/*
	 * get_hint() requires that the domain name is least significant
	 * and sorted by dictionary order.
	 */
	h1 = MBR2STRUCT(n1, const struct hint, tree);
	h2 = MBR2STRUCT(n2, const struct hint, tree);
	if ((diff = CMP16(h1->class, h2->class)) != 0)
		return diff;
	return domcmp_nbr(h1->zone, h2->zone);
}


/* new_rrset never fails */
struct rrset *
new_rrset(const char *domain, unsigned int type, unsigned int class,
    uint32_t ttl, int count)
{
	static const struct rrset r0;
	struct rrset *r;
	size_t len;

	r = (struct rrset *)c_new(sizeof(*r) + sizeof(r->datav[0]) * count);
	*r = r0;

	r->c.eol = curtime + ttl;
	len = domlen(domain);
	r->domain = c_extcpy(domain, len);
	r->type = type;
	r->class = class;

	return r;
}

struct rrset *
get_rrset(const char *domain, unsigned int type, unsigned int class)
{
	struct splay_range range;
	struct rrset *r, tmpl, min, max;
	struct splay_node *found;

	if (type == QTYPE_ANY) {
		min.domain = max.domain = domain;
		min.type = 0;
		max.type = 65535;
		min.class = max.class = class;
		found = splay_range_start(&rrset_descr, &range,
		    &min.tree, &max.tree);
		while (found != NULL) {
			r = MBR2STRUCT(found, struct rrset, tree);
			/*
			 * Ignore NoData unless it is for QTYPE_ANY.  If all
			 * cached RRsets are NoData, we want to query ANY.
			 * If it is also NoData, it is cached on QTYPE_ANY.
			 */
			if (curtime <= r->c.eol &&
			    !(r->rcode == 0 && r->datac == 0 &&
			    r->type != QTYPE_ANY))
				return r;
			found = splay_range_next(&rrset_descr, &range);
		}
		return NULL;
	}

	tmpl.domain = domain;
	tmpl.type = type;
	tmpl.class = class;
	found = splay_find(&rrset_descr, &tmpl.tree);
	if (found == NULL)
		return NULL;
	r = MBR2STRUCT(found, struct rrset, tree);
	if (curtime <= r->c.eol)
		return r;
	LDEBUG2("discard %p (rrset %s,%u,%u)",
	    r, dom2str(r->domain), r->type, r->class);
	(void)splay_remove(&rrset_descr, &r->tree);
	r->c.destructor = &cache_expired;
	return NULL;
}

int
foreach_rrset_any(struct rrset *head,
    int (*callback)(struct rrset *, void *), void *arg)
{
	struct splay_range range;
	struct rrset *r, max;
	struct splay_node *found;
	int ret;

	max.domain = head->domain;
	max.type = 65535;
	max.class = head->class;
	found = splay_range_start(&rrset_descr, &range, &head->tree, &max.tree);
	while (found != NULL) {
		r = MBR2STRUCT(found, struct rrset, tree);
		if (curtime <= r->c.eol) {
			ret = (*callback)(r, arg);
			if (ret != 0)
				return ret;
		}
		found = splay_range_next(&rrset_descr, &range);
	}
	return 0;
}

void
replace_rrset(struct rrset *r)
{
	struct rrset *del;
	struct splay_node *found;

	found = splay_replace(&rrset_descr, &r->tree);
	r->c.destructor = &unlink_rrset;
	if (found != NULL) {
		del = MBR2STRUCT(found, struct rrset, tree);
#if 0
		LDEBUG2("replace %p (rrset %s,%u,%u)",
		    del, dom2str(del->domain), del->type, del->class);
#endif
		del->c.destructor =
		    curtime <= del->c.eol ? &cache_replaced : &cache_expired;
	}
}

void
replace_rrset_as_glue(struct rrset *r)
{
	/* struct rrset and struct glue has the same format. */
	replace_glue((struct glue *)r);
}

static void
unlink_rrset(struct cache *c)
{
	struct rrset *r;

	r = (struct rrset *)c;
	LDEBUG2("%s %p (rrset %s,%d,%d)",
	    curtime <= r->c.eol ? "reclaim" : "collect", r,
	    dom2str(r->domain), r->class, r->type);
	(void)splay_remove(&rrset_descr, &r->tree);
	r->c.destructor = NULL;
}

static int
rrsetcmp(const struct splay_node *n1, const struct splay_node *n2)
{
	const struct rrset *r1, *r2;
	int diff;

	r1 = MBR2STRUCT(n1, const struct rrset, tree);
	r2 = MBR2STRUCT(n2, const struct rrset, tree);
	if ((diff = CMP16(r1->class, r2->class)) != 0)
		return diff;
	if ((diff = domcmp(r1->domain, r2->domain)) != 0)
		return diff;
	/* TYPE is least significant in order to support QTYPE_ANY. */
	return CMP16(r1->type, r2->type);
}


struct authority *
new_authority(const char *zone, unsigned int class, uint32_t ttl, int count)
{
	static const struct authority a0;
	struct authority *a;
	size_t len;

	a = (struct authority *)c_new(sizeof(*a) + sizeof(a->nsv[0]) * count);
	*a = a0;

	a->c.eol = curtime + ttl;
	len = domlen(zone);
	a->zone = c_extcpy(zone, len);
	a->class = class;

	return a;
}

struct authority *
get_authority(const char **superzone, const char *zone, unsigned int class)
{
	struct authority *a, tmpl;
	int ret;

	*superzone = zone + domlen(zone);
	tmpl.zone = zone;
	tmpl.class = class;
	for (;;) {
		if (authority_descr.root == NULL)
			return NULL;
		ret = splay_find_leq(&authority_descr, &tmpl.tree);
		a = MBR2STRUCT(authority_descr.root, struct authority, tree);
		if (ret == 0) {		/* if found */
			if (curtime <= a->c.eol) {
				*superzone = tmpl.zone;
				return a;
			}
			LDEBUG2("discard %p (authority %s,%u)",
			    a, dom2str(a->zone), a->class);
			(void)splay_remove(&authority_descr, &a->tree);
			a->c.destructor = &cache_expired;
		} else if (ret > 0)		/* only greater was found */
			return NULL;
		else if (a->class != class)  /* underflowed to another class */
			return NULL;
		/* <a> may be unlinked but still valid */
		tmpl.zone = domcommon(a->zone, tmpl.zone);
	}
}

void
replace_authority(struct authority *a)
{
	struct authority *del;
	struct splay_node *found;

	found = splay_replace(&authority_descr, &a->tree);
	a->c.destructor = &unlink_authority;
	if (found != NULL) {
		del = MBR2STRUCT(found, struct authority, tree);
#if 0
		LDEBUG2("replace %p (authority %s,%u)",
		    del, dom2str(del->zone), del->class);
#endif
		del->c.destructor =
		    curtime <= del->c.eol ? &cache_replaced : &cache_expired;
	}
}

static void
unlink_authority(struct cache *c)
{
	struct authority *a;

	a = (struct authority *)c;
	LDEBUG2("%s %p (authority %s,%d)",
	    curtime <= a->c.eol ? "reclaim" : "collect", a,
	    dom2str(a->zone), a->class);
	(void)splay_remove(&authority_descr, &a->tree);
	a->c.destructor = NULL;
}

static int
authoritycmp(const struct splay_node *n1, const struct splay_node *n2)
{
	const struct authority *a1, *a2;
	int diff;

	/*
	 * get_authority() requires that the domain name is least significant
	 * and sorted by dictionary order.
	 */
	a1 = MBR2STRUCT(n1, const struct authority, tree);
	a2 = MBR2STRUCT(n2, const struct authority, tree);
	if ((diff = CMP16(a1->class, a2->class)) != 0)
		return diff;
	return domcmp_nbr(a1->zone, a2->zone);
}


struct glue *
new_glue(const char *domain, unsigned int type, unsigned int class,
    uint32_t ttl, int count)
{
	static const struct glue g0;
	struct glue *g;
	size_t len;

	g = (struct glue *)c_new(sizeof(*g) + sizeof(g->datav[0]) * count);
	*g = g0;

	g->c.eol = curtime + ttl;
	len = domlen(domain);
	g->domain = c_extcpy(domain, len);
	g->type = type;
	g->class = class;

	return g;
}

struct glue *
get_glue(const char *domain, unsigned int type, unsigned int class)
{
	struct glue *g, tmpl;
	struct splay_node *found;

	tmpl.domain = domain;
	tmpl.type = type;
	tmpl.class = class;
	found = splay_find(&glue_descr, &tmpl.tree);
	if (found == NULL)
		return NULL;
	g = MBR2STRUCT(found, struct glue, tree);
	if (curtime <= g->c.eol)
		return g;
	LDEBUG2("discard %p (glue %s,%u,%u)",
	    g, dom2str(g->domain), g->type, g->class);
	(void)splay_remove(&glue_descr, &g->tree);
	g->c.destructor = &cache_expired;
	return NULL;
}

void
replace_glue(struct glue *g)
{
	struct glue *del;
	struct splay_node *found;

	found = splay_replace(&glue_descr, &g->tree);
	g->c.destructor = &unlink_glue;
	if (found != NULL) {
		del = MBR2STRUCT(found, struct glue, tree);
#if 0
		LDEBUG2("replace %p (glue %s,%u,%u)",
		    del, dom2str(del->domain), del->type, del->class);
#endif
		del->c.destructor =
		    curtime <= del->c.eol ? &cache_replaced : &cache_expired;
	}
}

static void
unlink_glue(struct cache *c)
{
	struct glue *g;

	g = (struct glue *)c;
	LDEBUG2("%s %p (glue %s,%d,%d)",
	    curtime <= g->c.eol ? "reclaim" : "collect", g,
	    dom2str(g->domain), g->class, g->type);
	(void)splay_remove(&glue_descr, &g->tree);
	g->c.destructor = NULL;
}

static int
gluecmp(const struct splay_node *n1, const struct splay_node *n2)
{
	const struct glue *g1, *g2;
	int diff;

	g1 = MBR2STRUCT(n1, const struct glue, tree);
	g2 = MBR2STRUCT(n2, const struct glue, tree);
	if ((diff = CMP16(g1->type, g2->type)) != 0)
		return diff;
	if ((diff = CMP16(g1->class, g2->class)) != 0)
		return diff;
	return domcmp(g1->domain, g2->domain);
}


unsigned int
get_servcap_rtt(const struct sockaddr *sa)
{
	struct servcap t;
	struct servcap *sc;
	struct splay_node *found;
	unsigned int rtt;

	t.sa = sa;
	found = splay_find(&servcap_descr, &t.tree);
	if (found == NULL)
		return UNKNOWN_RTT;
	sc = MBR2STRUCT(found, struct servcap, tree);
	if (sc->active.tv_sec == 0)
		return sc->rtt;
	/*
	 * If there is an outstanding packet but no response yet,
	 * penalize it early.
	 */
	rtt = calc_rtt(&sc->active);
	if (rtt < sc->rtt)
		return sc->rtt;
	else
		return (sc->rtt + rtt) / 2;
}

void
add_servcap_rtt(const struct sockaddr *sa, unsigned int add)
{
	struct servcap *sc;

	sc = get_or_new_servcap(sa);
	LDEBUG2("rtt: %s: %u-->%u (add)", sa2str(sa), sc->rtt, sc->rtt + add);
	sc->rtt += add;
	if (15000 < sc->rtt)
		sc->rtt = 15000;
}

void
decay_servcap_rtt(const struct sockaddr *sa)
{
	struct servcap *sc;

	sc = get_or_new_servcap(sa);
	sc->rtt -= (sc->rtt + 31) / 32;
}

void
start_servcap_rtt(const struct sockaddr *sa, struct timespec *start)
{
	struct servcap *sc;

	sc = get_or_new_servcap(sa);
	if (sc->active.tv_sec == 0)
		sc->active = *start;
}

void
stop_servcap_rtt(const struct sockaddr *sa, struct timespec *start, int cancel)
{
	struct servcap *sc;
	unsigned int rtt;

	sc = get_or_new_servcap(sa);
	sc->active.tv_sec = 0;
	rtt = calc_rtt(start);

	/* If timed-out or interrupted by another query, don't reduce RTT. */
	if (cancel && rtt <= sc->rtt) {
		LDEBUG2("rtt: %s: %u (not updated by %u)",
		    sa2str(sa), sc->rtt, rtt);
		return;
	}
	rtt = (sc->rtt + rtt) / 2;
	LDEBUG2("rtt: %s: %u-->%u", sa2str(sa), sc->rtt, rtt);
	sc->rtt = rtt;
}

static unsigned int
calc_rtt(const struct timespec *start)
{
	struct timespec now, diff;

	(void)clock_gettime(CLOCK_MONOTONIC, &now);
	diff.tv_sec = now.tv_sec - start->tv_sec;
	diff.tv_nsec = now.tv_nsec - start->tv_nsec;
	if (diff.tv_nsec < 0) {
		diff.tv_sec--;
		diff.tv_nsec += 1000000000L;
	}
	if (0 <= diff.tv_sec && diff.tv_sec < 15)
		return diff.tv_sec * 1000 + diff.tv_nsec / 1000000;
	else
		return 15000;
}

int
get_servcap_edns_version(const struct sockaddr *sa)
{
	struct servcap t;
	struct splay_node *found;

	t.sa = sa;
	found = splay_find(&servcap_descr, &t.tree);
	if (found != NULL)
		return MBR2STRUCT(found, struct servcap, tree)->edns_version;
	return EDNS_UNKNOWN;
}

void
set_servcap_edns_version(const struct sockaddr *sa, int edns_version)
{
	struct servcap *sc;

	sc = get_or_new_servcap(sa);
	sc->edns_version = edns_version;
}

struct servcap *
get_or_new_servcap(const struct sockaddr *sa)
{
	static const struct servcap sc0;
	struct servcap *sc, t;
	struct splay_node *found;

	t.sa = sa;
	found = splay_find(&servcap_descr, &t.tree);
	if (found != NULL)
		return MBR2STRUCT(found, struct servcap, tree);

	sc = (struct servcap *)c_new(sizeof(*sc));
	*sc = sc0;

	sc->sa = (struct sockaddr *)c_extcpy_align(sa, SA_LEN(sa));
	sc->rtt = UNKNOWN_RTT;
	sc->edns_version = EDNS_UNKNOWN;
	sc->active.tv_sec = 0;
	/* always insert */
	(void)splay_replace(&servcap_descr, &sc->tree);
	sc->c.destructor = &unlink_servcap;
	/* sc->c.eol = 0 */

	return sc;
}

static void
unlink_servcap(struct cache *c)
{
	(void)splay_remove(&servcap_descr, &((struct servcap *)c)->tree);
	c->destructor = NULL;
}

static int
servcapcmp(const struct splay_node *n1, const struct splay_node *n2)
{
	return sacmp(MBR2STRUCT(n1, const struct servcap, tree)->sa,
	    MBR2STRUCT(n2, const struct servcap, tree)->sa);
}
