/* $Id: format.c 341 2011-06-02 00:50:59Z ken $ */
/*
 * Copyright (c) 2006 KAMADA Ken'ichi.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#include <arpa/inet.h>
#include <ctype.h>
#include <inttypes.h>
#include <limits.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>

#include "var.h"
#include "util.h"
#include "stats.h"
#include "format.h"

#define NUM2CHAR(n)	((n) + '0')
#define CHAR2NUM(c)	((c) - '0')

static int preparse_section(struct pktdec *pd, int section);
static int preparse_rr(struct pktdec *pd, int section);
static int rrcmp_pkt(const char *pkt, const char *r1,
    const char *d2, const char *tc2);
static int parse_opt(struct pktdec *pd);
static int parse_dname(char *dst, size_t dstlen,
    const char *pkt, const char *name, const char *end);
static int skip_dname(const char *name, const char *end);

static int compose_rdata(struct pktenc *pe,
    uint16_t type, const char *rd, uint16_t rdlen);
static int compress_dname(struct pktenc *pe, const char *domain);
static const char **make_label_array(const char **array, const char *p);
static int labelcmp(const char *p1, const char *p2);

int
dec_init(struct pktdec *pd, const char *buf, size_t len,
    const char *log_prefix)
{
	static const struct pktdec pd0;

	if (len < sizeof(*pd->hd)) {
		LDEBUG("%s truncated header", log_prefix);
		return -1;
	}
	*pd = pd0;
	pd->buf = buf;
	pd->p = buf + sizeof(*pd->hd);
	pd->end = buf + len;
	pd->hd = (const struct header *)buf;
	pd->opt.version = -1;
	return 0;
}

int
dec_question(struct pktdec *pd)
{
	int ret, count;

	/* Decode question section. */
	count = ntohs(pd->hd->qdcount);
	if (count != 1) {
		LDEBUG("%d entries in question section", count);
		return -1;
	}
	if ((ret = parse_dname(pd->name, sizeof(pd->name),
	    pd->buf, pd->p, pd->end)) == -1)
		return -1;
	pd->p += ret;
	if (pd->end - pd->p < 4) {
		LDEBUG("truncated record");
		return -1;
	}
	pd->type = load16(pd->p);
	pd->class = load16(pd->p + 2);
	pd->p += 4;
	return 0;
}

int
dec_preparse(struct pktdec *pd)
{
	int i;

	pd->countv[SECTION_AN] = ntohs(pd->hd->ancount);
	pd->countv[SECTION_NS] = ntohs(pd->hd->nscount);
	pd->countv[SECTION_AR] = ntohs(pd->hd->arcount);
	pd->idx = pd->idxbuf;
	pd->sortend = pd->sortbuf;

	/* Skip answer, authority, and additional sections. */
	for (i = 0; i < SECTION_MAX; i++) {
		pd->idxv[i] = pd->idx;
		pd->sortv[i] = pd->sortend;
		pd->last_rrset = -1;
		if (preparse_section(pd, i) == -1)
			return -1;
		if (pd->idxv[i] == pd->idx)
			pd->idxv[i] = NULL;
	}

	if (pd->p < pd->end)
		LDEBUG("extra %td octet(s)", pd->end - pd->p);
	return 0;
}

/*
 * Rough scan to locate the beginning of each section.
 * If OPT RR is found, decode it.
 */
static int
preparse_section(struct pktdec *pd, int section)
{
	int count;

	count = pd->countv[section];
	while (count-- > 0)
		if (preparse_rr(pd, section) == -1)
			return -1;
	return 0;
}

static int
preparse_rr(struct pktdec *pd, int section)
{
	const char *rr;
	uint16_t rdlen;
	struct pktdec_sort *sortv;
	int i, upper, lower;
	int ret;

	rr = pd->p;
	if ((ret = parse_dname(pd->name, sizeof(pd->name),
	    pd->buf, pd->p, pd->end)) == -1)
		return -1;
	pd->p += ret;
	if (pd->end - pd->p < 10) {
		LDEBUG("truncated record");
		return -1;
	}
	rdlen = load16(pd->p + 8);
	if (pd->end - pd->p - 10 < rdlen) {
		LDEBUG("truncated record");
		return -1;
	}

	if (load16(pd->p) == PTYPE_OPT) {
		pd->countv[section]--;		/* remove OPT virtually */
		if (section != SECTION_AR) {
			LDEBUG("OPT not in additional section");
			return -1;		/* [RFC2671 4.1] */
		}
		if (parse_opt(pd) != 0)
			return -1;
		pd->p += 10 + rdlen;
		return 0;
	}

	/*
	 * Build sorted index
	 */
	if (endof(pd->idxbuf) <= pd->idx) {
		LERR("SANITY: out of idx");
		return -1;
	}
	pd->idx->rr = rr;
	pd->idx->next = NULL;
	pd->idx->count = 1;

	sortv = pd->sortv[section];
	if (endof(pd->sortbuf) <= pd->sortend) {
		LERR("SANITY: out of sortbuf");
		return -1;
	}
	/* binary search */
	upper = pd->sortend - sortv;
	lower = 0;
	i = upper / 2;
	while (lower != upper) {
		ret = rrcmp_pkt(pd->buf, sortv[i].rr, pd->name, pd->p);
		if (ret == 0)
			break;
		if (ret <= 0)
			lower = i + 1;
		else
			upper = i;
		i = (upper + lower) / 2;
	}
	if (lower != upper) {
		/* found */
		sortv[i].first_idx->count++;
		pd->idx->next = sortv[i].last_idx->next;
		sortv[i].last_idx->next = pd->idx;
		sortv[i].last_idx = pd->idx;
	} else {
		/* not found; insert */
		if (pd->last_rrset >= 0)
			sortv[pd->last_rrset].last_idx->next = pd->idx;
		memmove(&sortv[i + 1], &sortv[i],
		    (char *)pd->sortend - (char *)&sortv[i]);
		pd->sortend++;
		sortv[i].rr = rr;
		sortv[i].first_idx = pd->idx;
		sortv[i].last_idx = pd->idx;
		pd->last_rrset = i;
	}
	pd->idx++;

	pd->p += 10 + rdlen;
	return 0;
}

/*
 * Compare two RRs <r1> and <d2, tc2>.
 * r1: in-packet format (may have pointers)
 * d2: in-core format (must not have pointers)
 */
static int
rrcmp_pkt(const char *pkt, const char *r1, const char *d2, const char *tc2)
{
	const char *tc1;
	int diff;

	tc1 = NULL;

	/* compare NAME */
	for (;;) {
		while (((unsigned char)*r1 >> 6) == 3) {
			if (tc1 == NULL)
				tc1 = r1 + 2;
			r1 = pkt + (((unsigned char)*r1 & 0x3f) << 8) +
			    (unsigned char)*(r1 + 1);
		}
		if (*r1 == 0 && *d2 == 0)
			break;		/* NAME matched */
		if ((diff = labelcmp(r1, d2)) != 0)
			return diff;
		r1 += 1 + (unsigned char)*r1;
		d2 += 1 + (unsigned char)*d2;
	}
	if (tc1 == NULL)
		tc1 = r1 + 1;

	/* compare TYPE and CLASS */
	return memcmp(tc1, tc2, 4);
}

/*
 * Parse OPT RR without NAME field.
 * The VERSION field is not checked in this function, because the responder
 * returns its highest implementation level. [RFC2671 4.6]
 */
static int
parse_opt(struct pktdec *pd)
{
	uint16_t rdlen;
	uint16_t optcode, optlen;
	const char *p;

	/* Two or more OPTs are not allowed [RFC2671 4.1]. */
	if (pd->opt.version != -1) {
		LDEBUG("multiple OPT RRs");
		return -1;
	}
	pd->opt.udpsize = load16(pd->p + 2);
	pd->opt.extrcode = (unsigned char)*(pd->p + 4);
	pd->opt.version = (unsigned char)*(pd->p + 5);
	pd->opt.extflags = load16(pd->p + 6);

	/* Parse variable options.  The caller should have checked rdlen. */
	rdlen = load16(pd->p + 8);
	p = pd->p + 10;
	while (rdlen > 0) {
		if (rdlen < 4) {
			LDEBUG("truncated option");
			return -1;
		}
		optcode = load16(p);
		optlen = load16(p + 2);
		if (rdlen - 4 < optlen) {
			LDEBUG("truncated option");
			return -1;
		}
		rdlen -= 4 + optlen;
		p += 4 + optlen;
		/*
		 * Ignore unknown options
		 * [draft-ietf-dnsext-rfc2671bis-edns0-05 6.2].
		 * History: RFC 2671 did not specify this.
		 * draft-msawyer-dnsext-edns-attributes discussed this,
		 * but it did not proceed.
		 */
		LDEBUG("unknown option %u", optcode);
	}
	return 0;
}

/* Start decoding from the specified section. */
void
dec_from_section(struct pktdec *pd, int section)
{
	pd->idx = pd->idxv[section];
	pd->count = 0;
}

int
dec_is_end(struct pktdec *pd)
{
	/* If there are unused RRs, skip them. */
	if (pd->count != 0) {
		LDEBUG2("skipping %d RR(s)", pd->count);
		do {
			pd->idx = pd->idx->next;
		} while (--pd->count > 0);
	}
	return pd->idx == NULL;
}

int
dec_next_rrset(struct pktdec *pd)
{
	int ret;

	if (pd->count != 0) {
		LERR("SANITY: unfinished RRset");
		return -1;
	}
	if (pd->idx->count == 0) {
		LERR("SANITY: RRset out of sync");
		return -1;
	}

	pd->count = pd->idx->count;
	pd->p = pd->idx->rr;
	if ((ret = parse_dname(pd->name, sizeof(pd->name),
	    pd->buf, pd->p, pd->end)) == -1)
		return -1;
	pd->p += ret;
	/*
	 * This is safe because preparse routines already checked the length.
	 * OPTs have been removed.
	 */
	pd->type = load16(pd->p);
	pd->class = load16(pd->p + 2);
	pd->ttl = load32(pd->p + 4);	/* TTL of the 1st RR, tentatively */
	return 0;
}

int
dec_parse_rr(struct pktdec *pd)
{
	int ret;

	if (pd->count <= 0 || pd->idx == NULL) {
		LERR("SANITY: no RR to parse");
		return -1;
	}
	pd->p = pd->idx->rr;
	pd->idx = pd->idx->next;
	pd->count--;

	if ((ret = skip_dname(pd->p, pd->end)) == -1)
		return -1;
	pd->p += ret;
	pd->ttl = load32(pd->p + 4);
	pd->rdlen = load16(pd->p + 8);
	pd->rd = pd->p + 10;

	if (pd->ttl > DNS_MAX_TTL) {
		LDEBUG("truncating too long TTL: %" PRIu32, pd->ttl);
		pd->ttl = DNS_MAX_TTL;
	}

	return 0;
}

/* Decode domain-name(s) in RDATA. */
int
dec_parse_rdata(struct pktdec *pd)
{
	static const char rdata_format[][6] = {
		[TYPE_A] = { 4, 0 },	/* A and AAAA are used by myself. */
		[TYPE_AAAA] = { 16, 0, },

		[TYPE_NS] = { -1, 0 },
		[TYPE_MD] = { -1, 0 },
		[TYPE_MF] = { -1, 0 },
		[TYPE_CNAME] = { -1, 0 },
		[TYPE_SOA] = { -1, -1, 20, 0 },	/* <dn> <dn> 32 32 32 32 32 */
		[TYPE_MB] = { -1, 0 },
		[TYPE_MG] = { -1, 0 },
		[TYPE_MR] = { -1, 0 },
		[TYPE_PTR] = { -1, 0 },
		[TYPE_MINFO] = { -1, -1, 0 },	/* <dn> <dn> */
		[TYPE_MX] = { 2, -1, 0 },	/* 16 <dn> */
		[TYPE_RP] = { -1, -1, 0 },	/* <dn> <dn> [RFC1183] */
		[TYPE_AFSDB] = { 2, -1, 0 },	/* 16 <dn> [RFC1183] */
		[TYPE_RT] = { 2, -1, 0 },	/* 16 <dn> [RFC1183] */
		[TYPE_SIG] = { 18, -1, 0 },
				/* 16 8 8 32 32 32 16 <dn> var [RFC2535] */
		[TYPE_PX] = { 2, -1, -1, 0 },	/* 16 <dn> <dn> [RFC2163] */
		[TYPE_NXT] = { -1, 0 },		/* <dn> var [RFC2535] */
		[TYPE_SRV] = { 6, -1, 0 },	/* 16 16 16 <dn> [RFC2782] */
		[TYPE_NAPTR] = { 4, -2, -2, -2, -1, 0 },
				/* 16 16 <cs> <cs> <cs> <dn> [RFC2915] */
	};
	const char *pat;	/* -1=dn, -2=cs, positive=skip octets, 0=end */
	const char *src, *srcend;
	char *dst, *dstend;
	int used;

	/*
	 * All well-known RRs (defined in RFC 1035) MUST be decompressed.
	 * RP, AFSDB, RT, SIG, PX, NXT, NAPTR, and SRV SHOULD be also
	 * decompressed [RFC3597 4].
	 * All well-known RRs have class-invariant formats [RFC1035 3.3].
	 */
	if (pd->type >= lengthof(rdata_format))
		return 0;
	if (*(pat = rdata_format[pd->type]) == 0)
		return 0;

	src = pd->rd;
	srcend = pd->rd + pd->rdlen;
	dst = pd->rd_tmp;
	dstend = pd->rd_tmp + sizeof(pd->rd_tmp);

	for (; *pat != 0; pat++) {
		switch (*pat) {
		case -1:
			if ((used = parse_dname(dst, dstend - dst,
			    pd->buf, src, srcend)) == -1)
				return -1;
			src += used;
			dst += skip_dname(dst, dstend);		/* XXX waste */
			break;
		case -2:
			if (srcend - src < 1)
				goto trunc;
			used = (unsigned char)*src + 1;
			if (srcend - src < used)
				goto trunc;
			if (dstend - dst < used)
				goto nobuf;
			memcpy(dst, src, used);
			src += used;
			dst += used;
			break;
		default:
			if (srcend - src < *pat)
				goto trunc;
			if (dstend - dst < *pat)
				goto nobuf;
			memcpy(dst, src, *pat);
			src += *pat;
			dst += *pat;
			break;
		}
	}

	/* copy remaining octets */
	used = srcend - src;
	if (dstend - dst < used)
		goto nobuf;
	memcpy(dst, src, used);
	src += used;
	dst += used;

	pd->rd = pd->rd_tmp;
	pd->rdlen = dst - pd->rd_tmp;
	return 0;
trunc:
	LDEBUG("truncated RDATA");
	return -1;
nobuf:
	LDEBUG("out of RDATA buffer");
	return -1;
}

int
dec_parse_soa_minimum(struct pktdec *pd)
{
	const char *p, *end;
	int ret;

	p = pd->rd;
	end = pd->rd + pd->rdlen;
	/* skip mname, rname */
	if ((ret = skip_dname(p, end)) == -1)
		return -1;
	p += ret;
	if ((ret = skip_dname(p, end)) == -1)
		return -1;
	p += ret;
	if (end - p < 20)
		goto trunc;
	LDEBUG2("SOA serial=%u refresh=%u retry=%u expire=%u minimum=%u",
	    load32(p), load32(p + 4), load32(p + 8),
	    load32(p + 12), load32(p + 16));
	pd->soa_minimum = load32(p + 16);
	return 0;
trunc:
	LDEBUG("truncated SOA");
	return -1;
}

static int
parse_dname(char *dst, size_t dstlen,
    const char *pkt, const char *name, const char *end)
{
	const char *p, *used;
	size_t clen;
	int pointer;

	/*
	 * Don't import too long names, because domcpy(), which is used for
	 * in-core buffers, doesn't check buffer length.
	 */
	if (dstlen > DNS_MAX_NAME_LEN)
		dstlen = DNS_MAX_NAME_LEN;

	used = NULL;
	pointer = 50;
	p = name;
	for (;;) {
		if (end - p < 1)
			goto trunc;
		clen = (unsigned char)*p++;
		switch (clen >> 6) {
		case 0:			/* label */
			if (dstlen < clen + 1)
				goto nobuf;
			dstlen -= clen + 1;
			*dst++ = clen;
			if (clen == 0)		/* root */
				return (used == NULL ? p : used) - name;
			if (end - p < clen)
				goto trunc;
			do {
				*dst++ = *p++;
			} while (--clen > 0);
			break;
		case 3:			/* pointer */
			if (--pointer < 0)	/* avoid infinite loop */
				goto manypointer;
			if (end - p < 1)
				goto trunc;
			clen = ((clen & 0x3f) << 8) + (unsigned char)*p++;
			/* reparse from the offset */
			if (used == NULL)
				used = p;
			if (end - pkt < clen)
				goto trunc;
			p = pkt + clen;
			break;
		case 1:			/* EDNS extended label [RFC2671] */
			/* No type is implemented. */
		default:
			LDEBUG("unknown label type");
			return -1;
		}
	}
manypointer:
	LDEBUG("too many pointers");
	return -1;
trunc:
	LDEBUG("truncated name");
	return -1;
nobuf:
	LDEBUG("too short name buffer");
	return -1;
}

static int
skip_dname(const char *name, const char *end)
{
	const char *p;
	size_t clen;

	p = name;
	for (;;) {
		if (end - p < 1)
			goto trunc;
		clen = (unsigned char)*p++;
		if (clen == 0)
			return p - name;
		switch (clen >> 6) {
		case 0:			/* label */
			if (end - p < clen)
				goto trunc;
			p += clen;
			break;
		case 3:			/* pointer */
			/* omit to expand pointers */
			if (end - p < 1)
				goto trunc;
			return p + 1 - name;
		case 1:			/* EDNS extended label */
		default:
			LDEBUG("unknown label type");
			return -1;
		}
	}
trunc:
	LDEBUG("truncated name");
	return -1;
}


int
enc_init(struct pktenc *pe, char *buf, size_t len,
    int edns_version, size_t peers_udpsize, int is_tcp)
{
	static const struct pktenc pe0;
	static const struct header header0;

	if (len < sizeof(*pe->hd))
		return -1;
	*pe = pe0;
	pe->buf = buf;
	pe->p = buf + sizeof(*pe->hd);
	pe->end = buf + len;
	pe->hd = (struct header *)buf;

	*pe->hd = header0;

	if (edns_version > 0)
		edns_version = 0;
	pe->opt.version = edns_version;
	/* In TCP mode, leave pe->peers_udpsize 0. */
	if (!is_tcp) {
		pe->peers_udpsize = peers_udpsize > DNS_MAX_UDP_SIZE ?
		    peers_udpsize : DNS_MAX_UDP_SIZE;
	}

	return 0;
}

void
enc_header(struct pktenc *pe, int rcode)
{
	pe->hd->flags |= htons(rcode & RCODE_MASK);
	pe->opt.extrcode = (unsigned int)rcode >> EXT_RCODE_SHIFT;

	if (pe->peers_udpsize != 0) {
		/* 11 is space for OPT RR. */
		if (pe->opt.version >= 0 ||
		    pe->opt.extrcode != 0 || pe->opt.extflags != 0)
			pe->peers_udpsize -= 11;
	}
}

int
enc_fini(struct pktenc *pe)
{
	/* Add OPT RR. */
	if (pe->opt.version >= 0 ||
	    pe->opt.extrcode != 0 || pe->opt.extflags != 0) {
		if (pe->opt.version < 0) {
			LERR("SANITY: OPT to a non-EDNS node");
			return -1;
		}
		if (pe->end - pe->p < 11) {
			LINFO("no space for OPT");
			return -1;
		}
		*pe->p = 0;					/* NAME */
		store16(pe->p + 1, PTYPE_OPT);			/* TYPE */
		store16(pe->p + 3, DNS_MAX_EDNS_UDP_SIZE);	/* CLASS */
		*(pe->p + 5) = pe->opt.extrcode;		/* TTL */
		*(pe->p + 6) = pe->opt.version;
		store16(pe->p + 7, pe->opt.extflags);
		store16(pe->p + 9, 0);				/* RDLEN */
		pe->p += 11;
		pe->countv[SECTION_AR]++;
	}

	pe->hd->ancount = htons(pe->countv[SECTION_AN]);
	pe->hd->nscount = htons(pe->countv[SECTION_NS]);
	pe->hd->arcount = htons(pe->countv[SECTION_AR]);

	return 0;
}

int
enc_question(struct pktenc *pe,
    const char *domain, uint16_t type, uint16_t class)
{
	return enc_rr(pe, SECTION_QD, domain, type, class, 0, NULL, 0);
}

int
enc_rr(struct pktenc *pe, int section,
    const char *domain, uint16_t type, uint16_t class, uint32_t ttl,
    const char *rd, size_t rdlen)
{
	char *rollback;

	if (pe->ctl.stop_encoding)
		return 0;

	rollback = pe->p;
	if (compress_dname(pe, domain) == 0)
		return -1;
	if (pe->end - pe->p < 8)
		return -1;
	store16(pe->p, type);
	store16(pe->p + 2, class);

	if (section == SECTION_QD) {
		pe->p += 4;
		pe->hd->qdcount = htons(ntohs(pe->hd->qdcount) + 1);
		/*
		 * Just return without peers_udpsize check.  Any question
		 * section will fit 512 octets.
		 */
		return 0;
	}

	store32(pe->p + 4, ttl);
	pe->p += 8;
	if (compose_rdata(pe, type, rd, rdlen) == -1)
		return -1;

	/* If the response is too large for UDP, truncate it [RFC2181 9]. */
	if (pe->peers_udpsize > 0 &&
	    pe->p - pe->buf > pe->peers_udpsize) {
		/*
		 * XXX If TC bit is not set, partial RRsets are not
		 * allowed [RFC2181 9].  Aomidoro currently returns only
		 * required RRsets, so it is OK to rollback one RR here.
		 */
		pe->p = rollback;
		if (!pe->ctl.optional)
			pe->hd->flags |= htons(TC_BIT);
		pe->ctl.stop_encoding = 1;
		LDEBUG2("truncating to %u (peer's recv size %u)%s",
		    pe->p - pe->buf, pe->peers_udpsize,
		    pe->ctl.optional ? "" : ", TC");
		return 0;
	}

	pe->countv[section]++;

	return 0;
}

/* Encode RDLENGTH and RDATA, compressing domain-name(s) in RDATA. */
static int
compose_rdata(struct pktenc *pe, uint16_t type, const char *rd, uint16_t rdlen)
{
	char *rdlen_ptr;
	size_t consumed;

	rdlen_ptr = pe->p;
	if (pe->end - pe->p < 2)
		return -1;
	pe->p += 2;

	/*
	 * Encode only well-known RRs [RFC3597 4].
	 */
	switch (type) {
	case TYPE_CNAME:
	case TYPE_MB:
	case TYPE_MD:
	case TYPE_MF:
	case TYPE_MG:
	case TYPE_MR:
	case TYPE_NS:
	case TYPE_PTR:
		if ((consumed = compress_dname(pe, rd)) == 0)
			return -1;
		rd += consumed;
		rdlen -= consumed;
		break;
	case TYPE_MINFO:	/* <dn> <dn> */
	case TYPE_SOA:		/* <dn> <dn> 32 32 32 32 32 */
		if ((consumed = compress_dname(pe, rd)) == 0)
			return -1;
		rd += consumed;
		rdlen -= consumed;
		if ((consumed = compress_dname(pe, rd)) == 0)
			return -1;
		rd += consumed;
		rdlen -= consumed;
		break;
	case TYPE_MX:		/* 16 <dn> */
		if (pe->end - pe->p < 2)
			return -1;
		memcpy(pe->p, rd, 2);
		pe->p += 2;
		rd += 2;
		rdlen -= 2;
		if ((consumed = compress_dname(pe, rd)) == 0)
			return -1;
		rd += consumed;
		rdlen -= consumed;
		break;
	default:
		break;
	}

	/* copy remaining octets */
	if (pe->end - pe->p < rdlen)
		return -1;
	memcpy(pe->p, rd, rdlen);
	pe->p += rdlen;

	store16(rdlen_ptr, pe->p - (rdlen_ptr + 2));
	return 0;
}

/*
 * Return value:
 *  >0  Consumed octets.
 *   0  Error (out of buffer).
 */
static int
compress_dname(struct pktenc *pe, const char *domain)
{
	const char *newv[1 + (DNS_MAX_NAME_LEN - 1) / 2], **newp;
	const char *oldv[1 + (DNS_MAX_NAME_LEN - 1) / 2], **oldp;
	int i, len;
	uint16_t suffix;

	/*
	 * When a domain is given, <suffix> (or <pointer>) is uniquely
	 * decided.  That is the offset where that domain appears first.
	 */
	suffix = 0;
	newp = make_label_array(newv, domain);

	/*
	 * newv[0] is a pointer to the last (unused in this loop), but
	 * is a guard against newp becoming an invalid pointer as well.
	 */
	for (i = 0; i < pe->namec; i++) {
		/* If the suffix mismatches, no hope to be compressed. */
		if (suffix != pe->namev[i].pointer)
			continue;

		oldp = make_label_array(oldv, pe->buf + pe->namev[i].offset);

		/* Compress labels. */
		while (--newp > newv && --oldp > oldv &&
		    labelcmp(*newp, *oldp) == 0)
			suffix = *oldp - pe->buf;
		if (newp++ == newv)
			break;			/* no more label to compress */
	}

	/* Put uncompressed label(s) and remember it. */
	if (newp - 1 > newv) {
		if (pe->namec >= lengthof(pe->namev)) {
			LINFO("out of name compression pointer");
			pe->namec--;
		}
		pe->namev[pe->namec].offset = pe->p - pe->buf;
		pe->namev[pe->namec].pointer = suffix;
		pe->namec++;

		len = 0;
		while (--newp > newv)
			len += (unsigned char)**newp + 1;
		if (pe->end - pe->p < len)
			return 0;
		memcpy(pe->p, domain, len);
		pe->p += len;
	}

	/* Put the root or a pointer. */
	if (suffix == 0) {
		if (pe->end - pe->p < 1)
			return 0;
		*pe->p++ = 0;
	} else {
		if (pe->end - pe->p < 2)
			return 0;
		*pe->p++ = 0xc0 + (suffix >> 8);
		*pe->p++ = suffix;
	}
	return *newv + 1 - domain;
}

/*
 * Set pointers of the beginning of all labels until a pointer (or the root)
 * is found.  The input <p> must be from the trusted source and well-formed.
 * The result is stored in array[1..].  array[0] is special and points to
 * the end (a pointer or the root).
 */
static const char **
make_label_array(const char **array, const char *p)
{
	const char **a;
	unsigned int clen;

	a = array + 1;
	while ((clen = (unsigned char)*p++) != 0) {
		switch (clen >> 6) {
		case 0:
			*a++ = p - 1;
			p += clen;
			break;
		case 3:
			goto end;
		default:
			/* NOTREACHED */
			goto end;
		}
	}
end:
	array[0] = p - 1;
	return a;
}


/*
 * Return value:
 *   0  Buffer is too short.
 *  >0  Encoded length.
 */
size_t
enc_make_soa(char *buf, size_t len,
    const char *mname, const char *rname,
    uint32_t serial, uint32_t refresh, uint32_t retry, uint32_t expire,
    uint32_t minimum)
{
	size_t needlen, mlen, rlen;

	mlen = domlen(mname);
	rlen = domlen(rname);
	needlen = mlen + rlen + 20;
	if (len < needlen)
		return 0;

	memcpy(buf, mname, mlen);
	buf += mlen;
	memcpy(buf, rname, rlen);
	buf += rlen;
	store32(buf, serial);
	store32(buf + 4, refresh);
	store32(buf + 8, retry);
	store32(buf + 12, expire);
	store32(buf + 16, minimum);
	return needlen;
}


/* If d1 is the parent of d2, return the pointer in d2 where d1 is found. */
const char *
domparent(const char *d1, const char *d2)
{
	const char *p;
	int n1, n2;

	n1 = n2 = 0;
	p = d1;
	while ((unsigned char)*p != 0) {
		p += 1 + (unsigned char)*p;
		n1++;
	}
	p = d2;
	while ((unsigned char)*p != 0) {
		p += 1 + (unsigned char)*p;
		n2++;
	}
	if (n2 < n1)
		return 0;
	for (n2 -= n1; n2 > 0; n2--)
		d2 += 1 + (unsigned char)*d2;
	return domcmp(d1, d2) == 0 ? d2 : NULL;
}

const char *
domcommon(const char *d1, const char *d2)
{
	const char *v1[1 + (DNS_MAX_NAME_LEN - 1) / 2], **p1;
	const char *v2[1 + (DNS_MAX_NAME_LEN - 1) / 2], **p2, **orig2;

	p1 = make_label_array(v1, d1);
	p2 = make_label_array(v2, d2);
	orig2 = p2;
	while (p1--, p2--, p1 > v1 && p2 > v2)
		if (labelcmp(*p1, *p2) != 0)
			break;
	if (++p2 == orig2)		/* root label is not at the end */
		return v2[0];
	return *p2;
}

int
domcmp(const char *d1, const char *d2)
{
	int diff;

	while (*d1 != 0 || *d2 != 0) {
		if ((diff = labelcmp(d1, d2)) != 0)
			return diff;
		d1 += (unsigned char)*d1 + 1;
		d2 += (unsigned char)*d2 + 1;
	}
	return 0;
}

/*
 * All subdomains in a domain are adjacent.
 * A domain comes before all its subdomains.
 */
int
domcmp_nbr(const char *d1, const char *d2)
{
	const char *v1[1 + (DNS_MAX_NAME_LEN - 1) / 2], **p1;
	const char *v2[1 + (DNS_MAX_NAME_LEN - 1) / 2], **p2;
	int diff;

	p1 = make_label_array(v1, d1);
	p2 = make_label_array(v2, d2);
	while (p1--, p2--, p1 > v1 && p2 > v2)
		if ((diff = labelcmp(*p1, *p2)) != 0)
			return diff;
	if (p1 > v1)
		return 1;
	if (p2 > v2)
		return -1;
	return 0;
}

static int
labelcmp(const char *p1, const char *p2)
{
	int diff, len;

	if ((diff = (unsigned char)*p1 - (unsigned char)*p2) != 0)
		return diff;
	for (len = (unsigned char)*p1; len > 0; len--) {
		p1++;
		p2++;
		if ((diff = tolower((unsigned char)*p1) -
		    tolower((unsigned char)*p2)) != 0)
			return diff;
	}
	return 0;
}

const char *
domcpy(char *d, const char *s)
{
	size_t len;

	len = domlen(s);
	return memcpy(d, s, len);
}

size_t
domlen(const char *d)
{
	const char *p;

	p = d;
	while ((unsigned char)*p != 0)
		p += 1 + (unsigned char)*p;
	return p + 1 - d;
}

char *
domdup(const char *d)
{
	char *dup;
	size_t len;

	len = domlen(d);
	if ((dup = (char *)malloc(len)) == NULL)
		return NULL;
	return memcpy(dup, d, len);
}

int
str2dom(char *d, size_t len, const char *str)
{
	char *dot;
	int labellen, c;

	if (len > DNS_MAX_NAME_LEN)
		len = DNS_MAX_NAME_LEN;
	if (len-- < 1)
		return -1;
	dot = d++;
	if (*str == '\0')
		return -1;
	if (*str == '.' && *(str + 1) == '\0') {
		*dot = 0;
		return 0;
	}
	while (*str != '\0') {
		if (len-- < 1)		/* one byte is written in any case */
			return -1;
		switch ((c = (unsigned char)*str++)) {
		case '.':
			labellen = d - 1 - dot;
			if (labellen <= 0 || 63 < labellen)
				return -1;	/* empty or too long label */
			*dot = labellen;
			dot = d++;
			break;
		case '\\':
			if (isdigit((unsigned char)*str)) {
				if (!isdigit((unsigned char)str[1]) ||
				    !isdigit((unsigned char)str[2]))
					return -1;
				c = CHAR2NUM(str[0]) * 100 +
				    CHAR2NUM(str[1]) * 10 +
				    CHAR2NUM(str[2]);
				if (c > 255)
					return -1;
				str += 3;
			} else {
				/*
				 * The lexer guarantees that this isn't
				 * the terminating null character.
				 */
				c = *str++;
			}
			/* FALLTHROUGH */
		default:
			*d++ = c;
		}
	}
	if (d - 1 - dot != 0)
		return -1;			/* not a FQDN */
	*dot = 0;
	return 0;
}


struct pbuf *
alloc_pbuf(size_t size)
{
	struct pbuf *pb;

	if (size > SIZE_MAX - sizeof(*pb)) {
		LERR("pbuf too large");
		return NULL;
	}
	STAT_OBJ_PBUF_REQ();
	if ((pb = (struct pbuf *)malloc(sizeof(*pb) + size)) == NULL) {
		LERR("%m");
		STAT_OBJ_PBUF_FAIL();
		return NULL;
	}
	STAT_OBJ_PBUF_ALLOC();
	pb->size = size;
	pb->len = 0;
	pb->done = 0;
	return pb;
}

void
free_pbuf(struct pbuf *pb)
{
	/* Do not decrement the counter if allocation failed (i.e., NULL). */
	if (pb == NULL)
		return;
	free(pb);
	STAT_OBJ_PBUF_FREE();
}


uint16_t
load16(const char *buf)
{
	return ((uint16_t)(unsigned char)buf[0] << 8) + (unsigned char)buf[1];
}

uint32_t
load32(const char *buf)
{
	return ((uint32_t)(unsigned char)buf[0] << 24) +
	    ((uint32_t)(unsigned char)buf[1] << 16) +
	    ((uint32_t)(unsigned char)buf[2] << 8) +
	    (unsigned char)buf[3];
}

void
store16(char *buf, uint16_t val)
{
	buf[0] = val >> 8;
	buf[1] = val;
}

void
store32(char *buf, uint32_t val)
{
	buf[0] = val >> 24;
	buf[1] = val >> 16;
	buf[2] = val >> 8;
	buf[3] = val;
}


/*
 * debug
 */

const char *
dom2str(const char *d)
{
	/* DNS_MAX_NAME_LEN is enough if no escape occurs. */
	/* +2 is for the truncation mark.  +3 is escape margin. */
	static char cyclicbuf[2][DNS_MAX_NAME_LEN + 2 + 3];
	static int cyclicindex = 0;
	char *buf, *p, *end;
	int len, c;

	if (*d == 0)
		return ".";

	buf = cyclicbuf[cyclicindex];
	cyclicindex = (cyclicindex + 1) % lengthof(cyclicbuf);

	p = buf;
	end = p + sizeof(cyclicbuf[0]) - 3;	/* null char and trunc mark */
	while ((len = (unsigned char)*d++) != 0) {
		while (len-- > 0) {
			if (end - p < 4)
				goto nobuf;
			c = (unsigned char)*d++;
			if (c == '\\' || c == '.') {
				*p++ = '\\';
				*p++ = c;
			} else if (c < 0x21 || 0x7e < c) {
				*p++ = '\\';
				*p++ = NUM2CHAR(c / 100);
				*p++ = NUM2CHAR(c % 100 / 10);
				*p++ = NUM2CHAR(c % 10);
			} else
				*p++ = c;
		}
		if (end - p < 1)
			goto nobuf;
		*p++ = '.';
	}
	*p = '\0';
	return buf;
nobuf:
	*p++ = '.';
	*p++ = '.';
	*p = '\0';
	return buf;
}
