/* $Id: main.c 344 2011-06-02 01:39:19Z ken $ */
/*
 * Copyright (c) 2006 KAMADA Ken'ichi.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <signal.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <syslog.h>
#include <unistd.h>

#include "missing/missing.h"

#include "version.h"
#include "var.h"
#include "util.h"
#include "stats.h"
#include "format.h"
#include "net.h"
#include "cache.h"
#include "db.h"
#include "preload.h"
#include "ctlmsg.h"
#include "serve.h"

static void usage(void);
static void version(void);

struct option option = {
	.bind_port = NUM2STR(DNS_DEFAULT_PORT),
	.debug_level = 0,
	.foreground = 0,
	.user = NULL,
	.ctl_file = NULL,
};
struct conf conf_load, conf_run;

int
main(int argc, char *argv[])
{
	struct timeval tv, *tvp;
	int c, signo;

	setprogname(argv[0]);
	while ((c = getopt(argc, argv, "c:dFhp:u:V")) != -1) {
		switch (c) {
#if 0
		case 'c':
			option.ctl_file = optarg;
			break;
#endif
		case 'd':
			option.debug_level++;
			break;
		case 'F':
			option.foreground = 1;
			break;
		case 'h':
			usage();
			/* NOTREACHED */
		case 'p':
			option.bind_port = optarg;
			break;
		case 'u':
			option.user = optarg;
			break;
		case 'V':
			version();
			/* NOTREACHED */
		default:
			usage();
			/* NOTREACHED */
		}
	}
	argc -= optind;
	argv += optind;
	if (argc != 0)
		usage();

#ifdef LOG_PERROR
	openlog(NULL, LOG_PERROR, LOG_DAEMON);
#else
	openlog(NULL, 0, LOG_DAEMON);
	logmsg_stderr(1);
#endif
	if (option.debug_level == 0)
		setlogmask(LOG_UPTO(LOG_INFO));
	init_random();
	init_stats();
	if (load_config() != 0 ||
	    init_builtin_rrset() != 0 ||
	    c_init(conf_run.cache_size) != 0 ||
	    net_init() != 0)
		exit(1);
#if 0
	if (option.ctl_file != NULL &&
	    net_add_aomidctl(open_aomidctl(option.ctl_file)) != 0)
		exit(1);
#endif

	if (!option.foreground && make_pidfile() != 0)
		exit(1);
	if (option.user != NULL && run_as(option.user) != 0) {
		cleanup_pidfile();
		exit(1);
	}
	if (!option.foreground) {
		if (daemon(0, 0) == -1) {
			LERR("daemon: %m");
			cleanup_pidfile();
			exit(1);
		}
#ifdef LOG_PERROR
		openlog(NULL, 0, LOG_DAEMON);
#else
		logmsg_stderr(0);
#endif
	}
	if (!option.foreground && write_pidfile() != 0) {
		cleanup_pidfile();
		exit(1);
	}

	LNOTICE(PROGVERSION " started");
	for (;;) {
		tvp = schedule(&tv);
		invoke_spontaneous();
		signo = wait_socket(tvp);
		switch (signo) {
		case -1:
			LNOTICE(PROGVERSION " aborting");
			exit(1);
			/* NOTREACHED */
		case 0:
			break;
		case SIGHUP:
			LNOTICE(PROGVERSION " reloading");
			if (load_config() == 0)
				LNOTICE(PROGVERSION " reloaded");
			else
				LNOTICE(PROGVERSION " failed to reload");
			break;
		case SIGINT:
		case SIGTERM:
			LNOTICE(PROGVERSION " exiting");
			cleanup_pidfile();
			exit(0);
			/* NOTREACHED */
#ifdef SIGINFO
		case SIGINFO:
			info_query();
			info_stats();
			info_cache();
			break;
#endif
		default:
			LERR("unexpected signal %d", signo);
			break;
		}
	}
	/* NOTREACHED */
}

static void
usage(void)
{
	printf("usage: %s [-dFhV] [-p port] [-u user]\n",
	    getprogname());
	exit(1);
}

static void
version(void)
{
	printf(PROGVERSION "\n");
	exit(1);
}
