/* $Id: net.c 344 2011-06-02 01:39:19Z ken $ */
/*
 * Copyright (c) 2006 KAMADA Ken'ichi.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#ifdef HAVE_SYS_QUEUE_H
# include <sys/queue.h>
#else
# include "missing/queue.h"
#endif
#include <sys/socket.h>
#include <arpa/inet.h>			/* XXX inet_ntop */
#include <netinet/in.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <signal.h>
#include <stdio.h>			/* snprintf */
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>

#include "var.h"
#include "util.h"
#include "stats.h"
#include "format.h"
#include "net.h"

#define L_NET(...)							\
	logmsg(is_errno_operational() ? LOG_DEBUG : LOG_ERR,		\
	__VA_ARGS__)

/* DNS server sockets */
struct listen_on {
	int fd;
	int socktype;
	int protocol;
	struct sockaddr *sa;
	struct udpsock_serv {
		struct transport t;
		int fd;
	} udps;
	struct listen_on *next;
};

struct allow_query {
	int prefixlen;
	struct sockaddr *sa;
	struct allow_query *next;
};

/* used by UDP server function. */
struct udpsock_serv;

/* used by UDP client function. */
struct udpsock {
	struct transport t;
	int fd;
	TAILQ_ENTRY(udpsock) next;
};

/* used by TCP server and client functions. */
struct tcpsock {
	struct transport t;
	int fd;
	struct pbuf *sendbuf;
	struct pbuf *recvbuf;
	/*
	 * Reading is disabled from when a message (query) is read until
	 * a message (reply) is written _completely_.  If we want to read and
	 * write in parallel, we need to care about shutdown from the client.
	 */
	int (*io)(struct tcpsock *);
	void (*process)(struct tcpsock *);	/* switch client and server */
	void (*ioclose)(struct tcpsock *);
	int initial_reply_done;

	TAILQ_ENTRY(tcpsock) next;
};

static void sighandler(int signo);
static int add_listen_on(struct addrinfo *r);
static void set_v6only(int fd);
static void set_reuseaddr(int fd);
static void get_ephemeral_port_range(void);
static int bind_random_port(int fd, int family);

#if 0
static int accept_aomidctl(int s);
static void wait_next_ctl_msg(struct tcpsock *tcp);
static void recv_ctl_msg(struct tcpsock *tcp);
static void error_or_close_aomidctl(struct tcpsock *tcp);
#endif

static int recv_udp_query(struct udpsock_serv *udps);
static int recv_udp_reply(struct udpsock *udp);
static struct transport *open_udp_query(const struct sockaddr *sa);
static int send_udp_query(struct udpsock *udp, size_t len);
static int send_udp_reply(struct udpsock_serv *udps,
    const struct sockaddr *sa, size_t len);

static int accept_tcp_socket(int listen_fd);
static void recv_tcp_query(struct tcpsock *tcp);
static void recv_tcp_reply(struct tcpsock *tcp);
static struct transport *open_tcp_query(const struct sockaddr *sa);
static int send_tcp_query(struct tcpsock *tcp, size_t len);
static int send_tcp_reply(struct tcpsock *tcp, size_t len);
static void wait_tcp_reply(struct tcpsock *tcp);
static void wait_tcp_next_query(struct tcpsock *tcp);
static void error_tcp_query(struct tcpsock *tcp);
static void error_or_close_tcp_client(struct tcpsock *tcp);
static int read_tcp_socket(struct tcpsock *tcp);
static int pickup_tcp_msg(struct tcpsock *tcp);
static int write_tcp_socket(struct tcpsock *tcp);
static int set_nonblocking(int fd);

static struct udpsock *alloc_udp(int fd);
static void free_udp(struct udpsock *udp);
static struct tcpsock *alloc_tcp(int fd, int allocate_pbuf);
static void free_tcp(struct tcpsock *tcp);
static int serve_new_fd(int fd, fd_set *set);
static int is_errno_operational(void);

static int check_allow_query(const struct sockaddr *sa);
static int cmp_bits_be(const void *p1, const void *p2, int bits);

static int sigxfr[2];
#if 0
static int aomidctl = -1;
#endif
static unsigned int ephemeral_min, ephemeral_max;
static TAILQ_HEAD(, udpsock) udp_head = TAILQ_HEAD_INITIALIZER(udp_head);
static TAILQ_HEAD(, tcpsock) tcp_head = TAILQ_HEAD_INITIALIZER(tcp_head);
static struct udpsock *udp_next = NULL;
static struct tcpsock *tcp_next = NULL;
static fd_set rfds0, wfds0, rfds, wfds;
static int maxfd;
static int ntcp_clients;
/*
 * pbuf allocation rule:
 * -  UDP always uses transient_pbuf.
 * -  TCP from/to a client uses two dedicated pbufs for "send" and "recv".
 * -  TCP to/from an authoritative server uses one pbuf.
 *    Initially, it has no pbuf and uses transient_pbuf to compose a query.
 *    Before sending the query, a dedicated pbuf is allocated for "send".
 *    After sending the query, the pbuf is reused for "recv".
 */
static struct pbuf *transient_pbuf;


int
net_init(void)
{
	static const struct sigaction sa0;
	struct sigaction sa;
	struct listen_on *lstn;

	if (pipe(sigxfr) == -1) {
		LERR("pipe: %m");
		return -1;
	}
	if (set_nonblocking(sigxfr[1]) == -1) {
		close(sigxfr[0]);
		close(sigxfr[1]);
		return -1;
	}
	sa = sa0;
	sa.sa_handler = &sighandler;
	sa.sa_flags = SA_RESTART;
	(void)sigaction(SIGHUP, &sa, NULL);
	(void)sigaction(SIGINT, &sa, NULL);
	(void)sigaction(SIGTERM, &sa, NULL);
#ifdef SIGINFO
	(void)sigaction(SIGINFO, &sa, NULL);
#endif
	sa.sa_handler = SIG_IGN;
	(void)sigaction(SIGPIPE, &sa, NULL);

	get_ephemeral_port_range();

	if ((transient_pbuf = alloc_pbuf(DNS_MSG_BUF_SIZE)) == NULL)
		return -1;

	FD_ZERO(&rfds0);
	FD_ZERO(&wfds0);
	maxfd = 0;
	if (serve_new_fd(sigxfr[0], &rfds0) != 0)
		return -1;
	for (lstn = conf_run.listen_on; lstn != NULL; lstn = lstn->next)
		if (lstn->fd != -1 && serve_new_fd(lstn->fd, &rfds0) != 0)
			return -1;
	return 0;
}

#if 0
int
net_add_aomidctl(int s)
{
	if (s == -1)
		return -1;
	if (serve_new_fd(s, &rfds0) != 0)
		return -1;
	aomidctl = s;
	return 0;
}
#endif

void
inject_sighup(void)
{
	sighandler(SIGHUP);
}

static void
sighandler(int signo)
{
	/*
	 * Partial writes will never happen here.
	 * If the O_NONBLOCK flag is set on a pipe and write size is
	 * no more than PIPE_BUF, write() shall transfer all the data or
	 * fail with EAGAIN.  [SUSv3, XSH, write()]
	 * The minimum value of PIPE_BUF is 512 (_POSIX_PIPE_BUF) and
	 * sizeof(int) is quite reasonably expected to be less than 512.
	 * [SUSv3, XBD, <limits.h>]
	 */
	(void)write(sigxfr[1], &signo, sizeof(signo));
}

int
load_listen_on(const char *addr, const char *port)
{
	struct addrinfo hints, *res, *res0;
	int gairet;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_DGRAM;
	hints.ai_protocol = IPPROTO_UDP;
#ifdef AI_NUMERICSERV
	hints.ai_flags = AI_PASSIVE | AI_NUMERICHOST | AI_NUMERICSERV;
#else
	hints.ai_flags = AI_PASSIVE | AI_NUMERICHOST;
#endif
	gairet = getaddrinfo(addr, port, &hints, &res0);
	if (gairet != 0) {
		/* XXX yyerror */
		LERR("%s %s: %s", addr, port, gai_strerror(gairet));
		return -1;
	}
	for (res = res0; res != NULL; res = res->ai_next) {
		if (add_listen_on(res) == -1) {
			freeaddrinfo(res0);
			return -1;
		}
	}
	freeaddrinfo(res0);

	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = IPPROTO_TCP;
	gairet = getaddrinfo(addr, port, &hints, &res0);
	if (gairet != 0) {
		/* XXX yyerror */
		LERR("%s %s: %s", addr, port, gai_strerror(gairet));
		return -1;
	}
	for (res = res0; res != NULL; res = res->ai_next) {
		if (add_listen_on(res) == -1) {
			freeaddrinfo(res0);
			return -1;
		}
	}
	freeaddrinfo(res0);

	return 0;
}

static int
add_listen_on(struct addrinfo *r)
{
	static const struct listen_on lstn0;
	struct listen_on *lstn;

	if ((lstn = (struct listen_on *)malloc(sizeof(*lstn))) == NULL) {
		LERR("%m");
		return -1;
	}
	*lstn = lstn0;
	lstn->fd = -1;
	lstn->socktype = r->ai_socktype;
	lstn->protocol = r->ai_protocol;
	if ((lstn->sa = sadup(r->ai_addr)) == NULL) {
		LERR("%m");
		free(lstn);
		return -1;
	}
	lstn->next = conf_load.listen_on;
	conf_load.listen_on = lstn;
	return 0;
}

int
fixate_listen_on(int cancel)
{
	struct listen_on *lstn, *cmp;
	int s;

	if (!cancel) {
		if (conf_run.listen_on == NULL) {
			for (lstn = conf_load.listen_on; lstn != NULL;
			    lstn = lstn->next) {
				s = socket(lstn->sa->sa_family,
				    lstn->socktype, lstn->protocol);
				if (s == -1) {
					LERR("socket: %m");
					continue;
				}
				if (lstn->sa->sa_family == PF_INET6)
					set_v6only(s);
				if (lstn->socktype == SOCK_STREAM) {
					set_reuseaddr(s);
					/*
					 * When a TCP socket is closed before
					 * accept() (e.g., by RST), it blocks
					 * on (old) Berkeley-derived
					 * implementations.  [unpv12e 15.6]
					 */
					set_nonblocking(s);
				}
				if (bind(s, lstn->sa, SA_LEN(lstn->sa)) == -1) {
					LERR("bind: %m");
					close(s);
					continue;
				}
				if (lstn->socktype == SOCK_STREAM &&
				    listen(s, 5) == -1) {
					LERR("listen: %m");
					close(s);
					continue;
				}
				if (lstn->socktype == SOCK_DGRAM)
					lstn->udps.fd = s;
				lstn->fd = s;
			}
			conf_run.listen_on = conf_load.listen_on;
			conf_load.listen_on = NULL;
			return 0;
		}
		lstn = conf_run.listen_on;
		cmp = conf_load.listen_on;
		while (lstn != NULL && cmp != NULL) {
			if (lstn->socktype != cmp->socktype ||
			    lstn->protocol != cmp->protocol ||
			    sacmp(lstn->sa, cmp->sa) != 0)
				break;
			lstn = lstn->next;
			cmp = cmp->next;
		}
		if (lstn != NULL || cmp != NULL)
			LWARN("reloading listen-on not supported");
	}
	while ((lstn = conf_load.listen_on) != NULL) {
		conf_load.listen_on = conf_load.listen_on->next;
		if (lstn->fd != -1)
			close(lstn->fd);
		free(lstn->sa);
		free(lstn);
	}
	return 0;
}

static void
set_v6only(int fd)
{
	int on = 1;

	if (setsockopt(fd, IPPROTO_IPV6, IPV6_V6ONLY, &on, sizeof(on)) == -1) {
		LERR("setsockopt(V6ONLY): %m");
		exit(1);
	}
}

static void
set_reuseaddr(int fd)
{
	int on = 1;

	if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) == -1) {
		LERR("setsockopt(REUSEADDR): %m");
		exit(1);
	}
}

static void
get_ephemeral_port_range(void)
{
	/*
	 * Should look at
	 * - net.inet.ip.anonportmin/anonportmax and
	 *   net.inet6.ip6.anonportmin/anonportmax (NetBSD),
	 * - net.inet.ip.portrange.first/last/randomized (FreeBSD),
	 * etc.
	 */
	ephemeral_min = EPHEMERAL_PORT_MIN;
	ephemeral_max = EPHEMERAL_PORT_MAX;

#if 0
	if (ephemeral_min < 1 || ephemeral_max > 65535 ||
	    ephemeral_min > ephemeral_max) {
		LERR("invalid ephemeral port range %u..%u",
		    ephemeral_min, ephemeral_max);
		exit(1);
	}
#endif
}

static int
bind_random_port(int fd, int family)
{
	struct addrinfo hints, *res0;
	char portstr[6];
	unsigned int port;
	int ret, i;

	for (i = 0; i < 5; i++) {
		port = ephemeral_min +
		    random_secure(ephemeral_max - ephemeral_min + 1);

		memset(&hints, 0, sizeof(hints));
		hints.ai_family = family;
		hints.ai_socktype = SOCK_DGRAM;
		hints.ai_protocol = IPPROTO_UDP;
#ifdef AI_NUMERICSERV
		hints.ai_flags = AI_PASSIVE | AI_NUMERICHOST | AI_NUMERICSERV;
#else
		hints.ai_flags = AI_PASSIVE | AI_NUMERICHOST;
#endif
		(void)snprintf(portstr, sizeof(portstr), "%u", port);
		if ((ret = getaddrinfo(NULL, portstr, &hints, &res0)) != 0) {
			LERR("getaddrinfo: %s", gai_strerror(ret));
			return -1;
		}
		ret = bind(fd, res0->ai_addr, res0->ai_addrlen);
		if (ret == -1 && errno != EADDRINUSE) {
			LERR("bind: %m");
			freeaddrinfo(res0);
			return -1;
		}
		freeaddrinfo(res0);
		if (ret != -1)
			return 0;
		LDEBUG2("port %u is in use", port);
	}
	LINFO("failed to bind random UDP port");
	return -1;
}


/*
 * Return value:
 *  -1  error
 *   0  reschedule
 *  >0  signal number
 */
int
wait_socket(struct timeval *tvp)
{
	struct udpsock *udp;
	struct tcpsock *tcp;
	struct listen_on *lstn;
	int ret;
	int signo;

	rfds = rfds0;
	wfds = wfds0;
	ret = select(maxfd + 1, &rfds, &wfds, NULL, tvp);
	if (ret == -1) {
		if (errno == EINTR)
			return 0;
		LERR("select: %m");
		return -1;
	}

	/* receive signal */
	if (FD_ISSET(sigxfr[0], &rfds)) {
		if (read(sigxfr[0], &signo, sizeof(signo)) == -1) {
			LERR("read: %m");
			return -1;
		}
		/* Signal numbers are positive.  [SUSv3, XBD, <signal.h>] */
		return signo;
	}

#if 0
	/* accept control connection */
	if (aomidctl >=0 && FD_ISSET(aomidctl, &rfds))
		accept_aomidctl(aomidctl);
#endif

	for (lstn = conf_run.listen_on; lstn != NULL; lstn = lstn->next) {
		if (lstn->fd == -1)
			continue;
		if (!FD_ISSET(lstn->fd, &rfds))
			continue;
		if (lstn->socktype == SOCK_STREAM)
			accept_tcp_socket(lstn->fd);
		else
			recv_udp_query(&lstn->udps);
	}

	/* XXX unfair */
	for (udp = TAILQ_FIRST(&udp_head); udp != NULL; udp = udp_next) {
		udp_next = TAILQ_NEXT(udp, next);
		if (FD_ISSET(udp->fd, &rfds)) {
			if (recv_udp_reply(udp) != 0) {
				upcall_drop_server(&udp->t);
				free_udp(udp);
			}
		}
	}
	for (tcp = TAILQ_FIRST(&tcp_head); tcp != NULL; tcp = tcp_next) {
		tcp_next = TAILQ_NEXT(tcp, next);
		if (FD_ISSET(tcp->fd, &rfds) || FD_ISSET(tcp->fd, &wfds)) {
			if ((*tcp->io)(tcp) != 0) {
				/* I/O error or a successful client close */
				(*tcp->ioclose)(tcp);
				free_tcp(tcp);
			}
		}
	}

	return 0;
}

struct transport *
open_query(void *tag, const struct sockaddr *sa, int use_tcp)
{
	struct transport *t;

	if (use_tcp)
		t = open_tcp_query(sa);
	else
		t = open_udp_query(sa);
	if (t != NULL)
		t->tag = tag;
	return t;
}

int
send_query(struct transport *t, size_t len)
{
	if (t->is_tcp)
		return send_tcp_query((struct tcpsock *)t, len);
	else
		return send_udp_query((struct udpsock *)t, len);
}

int
send_reply(struct transport *t, const struct sockaddr *sa, size_t len)
{
	if (t->is_tcp)
		return send_tcp_reply((struct tcpsock *)t, len);
	else
		return send_udp_reply((struct udpsock_serv *)t, sa, len);
}


#if 0
static int
accept_aomidctl(int s)
{
	struct sockaddr_storage ss;
	socklen_t sslen;
	struct tcpsock *tcp;
	int fd;

	sslen = sizeof(ss);
	if ((fd = accept(s, (struct sockaddr *)&ss, &sslen)) == -1) {
		LERR("accept: %m");
		return -1;
	}

	if ((tcp = alloc_tcp(fd, 1)) == NULL) {
		close(fd);
		return -1;
	}
	tcp->io = &read_tcp_socket;
	tcp->process = &recv_ctl_msg;
	tcp->ioclose = &error_or_close_aomidctl;
	return 0;
}

static void
wait_next_ctl_msg(struct tcpsock *tcp)
{
	LDEBUG2("wait_next_ctl_msg");
	tcp->initial_reply_done = 1;
	tcp->sendbuf->len = 0;
	tcp->sendbuf->done = 0;
	tcp->io = &read_tcp_socket;
	tcp->process = &recv_ctl_msg;
	tcp->ioclose = &error_or_close_aomidctl;
	FD_CLR(tcp->fd, &wfds0);
	FD_SET(tcp->fd, &rfds0);
}

static void
recv_ctl_msg(struct tcpsock *tcp)
{
	size_t len;

	FD_CLR(tcp->fd, &rfds0);

	if ((len = upcall_ctl_msg(&tcp->t,
	    PBUF_DATA(tcp->recvbuf), tcp->recvbuf->len - 2)) == 0) {
		free_tcp(tcp);
		return;
	}
	/* Self-generated messages are much smaller than 2^16 bytes. */
	store16(PBUF_TCP_PTR(tcp->sendbuf), len);
	tcp->sendbuf->len = len + 2;
	if (tcp->sendbuf->done != 0)
		abort();

	tcp->io = &write_tcp_socket;
	tcp->process = &wait_next_ctl_msg;
	tcp->ioclose = &error_or_close_aomidctl;
	FD_SET(tcp->fd, &wfds0);
	return;
}

static void
error_or_close_aomidctl(struct tcpsock *tcp)
{
	/* nothing to do */
}
#endif


static int
recv_udp_query(struct udpsock_serv *udps)
{
	struct sockaddr_storage ss;
	socklen_t sslen;
	ssize_t ret;

	sslen = sizeof(ss);
	ret = recvfrom(udps->fd,
	    PBUF_DATA(transient_pbuf), transient_pbuf->size, 0,
	    (struct sockaddr *)&ss, &sslen);
	if (ret == -1) {
		LERR("recvfrom: %m");
		return -1;
	}
	/* XXX check trunc */

	if (check_allow_query((struct sockaddr *)&ss) != 0) {
		LDEBUG2("%s: stranger UDP", sa2str((struct sockaddr *)&ss));
		return 0;
	}

	(void)upcall_query(&udps->t, (struct sockaddr *)&ss,
	    PBUF_DATA(transient_pbuf), ret);
	return 0;
}

static int
recv_udp_reply(struct udpsock *udp)
{
	struct sockaddr_storage ss;
	socklen_t sslen;
	ssize_t ret;

	sslen = sizeof(ss);
	ret = recvfrom(udp->fd,
	    PBUF_DATA(transient_pbuf), transient_pbuf->size, 0,
	    (struct sockaddr *)&ss, &sslen);
	if (ret == -1) {
		L_NET("recvfrom: %m");
		return -1;
	}
	/* XXX check trunc */

	upcall_reply(&udp->t, (struct sockaddr *)&ss,
	    PBUF_DATA(transient_pbuf), ret);
	return 0;
}

static struct transport *
open_udp_query(const struct sockaddr *sa)
{
	struct udpsock *udp;
	int fd;

	if ((fd = socket(sa->sa_family, SOCK_DGRAM, IPPROTO_UDP)) == -1) {
		LERR("socket: %m");
		return NULL;
	}
	(void)bind_random_port(fd, sa->sa_family);
	if (connect(fd, sa, SA_LEN(sa)) == -1) {
		L_NET("connect: %m");
		close(fd);
		return NULL;
	}
	if ((udp = alloc_udp(fd)) == NULL) {
		close(fd);
		return NULL;
	}
	return &udp->t;
}

static int
send_udp_query(struct udpsock *udp, size_t len)
{
	if (send(udp->fd, PBUF_DATA(transient_pbuf), len, 0) == -1) {
		L_NET("send: %m");
		return -1;
	}
	return 0;
}

static int
send_udp_reply(struct udpsock_serv *udps, const struct sockaddr *sa,
    size_t len)
{
	if (sendto(udps->fd, PBUF_DATA(transient_pbuf), len, 0,
	    sa, SA_LEN(sa)) == -1) {
		LERR("sendto: %m");
		return -1;
	}
	return 0;
}


static int
accept_tcp_socket(int listen_fd)
{
	struct sockaddr_storage ss;
	socklen_t sslen;
	struct tcpsock *tcp;
	int fd;

	sslen = sizeof(ss);
	if ((fd = accept(listen_fd, (struct sockaddr *)&ss, &sslen)) == -1) {
		/*
		 * If connection is aborted before accept(), ECONNABORTED
		 * will be returned.  Some old OSes may return EWOULDBLOCK
		 * (Berkeley-derived) or EPROTO (SVR4).
		 */
		LERR("accept: %m");
		return -1;
	}

	if (check_allow_query((struct sockaddr *)&ss) != 0) {
		LDEBUG2("%s: stranger TCP", sa2str((struct sockaddr *)&ss));
		close(fd);
		return 0;
	}

	if ((tcp = alloc_tcp(fd, 1)) == NULL) {
		close(fd);
		return -1;
	}
	tcp->io = &read_tcp_socket;
	tcp->process = &recv_tcp_query;
	tcp->ioclose = &error_or_close_tcp_client;
	if (++ntcp_clients > MAX_TCP_CLIENTS) {
		/* Drop the oldest client. */
		LINFO("MAX_TCP_CLIENTS exceeded");
		/* tcp is inserted from the tail, so search from the head. */
		TAILQ_FOREACH(tcp, &tcp_head, next) {
			if (tcp->ioclose != &error_or_close_tcp_client)
				continue;	/* this is downstream */
			upcall_drop_tcp_client(&tcp->t);
			ntcp_clients--;
			free_tcp(tcp);
			break;
		}
	}
	return 0;
}

static void
recv_tcp_query(struct tcpsock *tcp)
{
	FD_CLR(tcp->fd, &rfds0);
	if (upcall_query(&tcp->t, NULL,
	    PBUF_DATA(tcp->recvbuf), tcp->recvbuf->len - 2) == -1) {
		ntcp_clients--;
		free_tcp(tcp);
		return;
	}
	/*
	 * XXX A consumed query message (in recvbuf) is removed after the
	 * reply is sent (see write_tcp_socket()).
	 * There is a case where recvbuf must be adjusted before
	 * upcall_query() above returns; recv_tcp_query --> upcall_query
	 * --> send_reply --> write_tcp_socket --> pickup_tcp_msg.
	 */
}

static void
recv_tcp_reply(struct tcpsock *tcp)
{
	struct pbuf *pb;

	FD_CLR(tcp->fd, &rfds0);
	/*
	 * Keep pbuf so that the received packet can be accessed after
	 * free_transport().
	 */
	pb = tcp->recvbuf;
	tcp->recvbuf = NULL;		/* no free by free_transport() ... */
	upcall_reply(&tcp->t, NULL, PBUF_DATA(pb), pb->len - 2);
	free_pbuf(pb);			/* ... and free here */
}

static struct transport *
open_tcp_query(const struct sockaddr *sa)
{
	struct tcpsock *tcp;
	int fd;

	if ((fd = socket(sa->sa_family, SOCK_STREAM, IPPROTO_TCP)) == -1) {
		LERR("socket: %m");
		return NULL;
	}
	if (set_nonblocking(fd) == -1) {
		close(fd);
		return NULL;
	}
	if (connect(fd, sa, SA_LEN(sa)) == -1 &&
	    !(errno == EINPROGRESS || errno == EWOULDBLOCK)) {
		L_NET("connect: %m");
		close(fd);
		return NULL;
	}

	if ((tcp = alloc_tcp(fd, 0)) == NULL) {
		close(fd);
		return NULL;
	}
	return &tcp->t;
}

static int
send_tcp_query(struct tcpsock *tcp, size_t len)
{
	struct pbuf *pb;

	/* allocate a pbuf; exchange it with transient_pbuf to avoid a copy */
	if ((pb = alloc_pbuf(DNS_MSG_BUF_SIZE)) == NULL)
		return -1;
	if (tcp->sendbuf != NULL)
		abort();
	tcp->sendbuf = transient_pbuf;
	transient_pbuf = pb;

	store16(PBUF_TCP_PTR(tcp->sendbuf), len);
	tcp->sendbuf->len = len + 2;
	if (tcp->sendbuf->done != 0)
		abort();

	FD_CLR(tcp->fd, &rfds0);	/* XXX not here */
	FD_SET(tcp->fd, &wfds0);
	/* special handler for connect may be better to get the correct errno */
	tcp->io = &write_tcp_socket;
	tcp->process = &wait_tcp_reply;
	tcp->ioclose = &error_tcp_query;
	return 0;
}

static int
send_tcp_reply(struct tcpsock *tcp, size_t len)
{
	store16(PBUF_TCP_PTR(tcp->sendbuf), len);
	tcp->sendbuf->len = len + 2;
	if (tcp->sendbuf->done != 0)
		abort();

	tcp->io = &write_tcp_socket;
	tcp->process = &wait_tcp_next_query;
	tcp->ioclose = &error_or_close_tcp_client;

	FD_SET(tcp->fd, &wfds0);
	/* The first write will not block. */
	if (!tcp->initial_reply_done && write_tcp_socket(tcp) != 0) {
		ntcp_clients--;
		free_tcp(tcp);
		return -1;
	}
	return 0;
}

static void
wait_tcp_reply(struct tcpsock *tcp)
{
	LDEBUG2("wait_tcp_reply");
	tcp->sendbuf->len = 0;
	tcp->sendbuf->done = 0;
	/*
	 * This function is only used for querying socket, which shares
	 * a pbuf to send/recv.
	 */
	if (tcp->recvbuf != NULL)
		abort();
	tcp->recvbuf = tcp->sendbuf;
	tcp->sendbuf = NULL;
	tcp->io = &read_tcp_socket;
	tcp->process = &recv_tcp_reply;
	tcp->ioclose = &error_tcp_query;
	FD_CLR(tcp->fd, &wfds0);
	FD_SET(tcp->fd, &rfds0);
}

static void
wait_tcp_next_query(struct tcpsock *tcp)
{
	LDEBUG2("wait_tcp_next_query");
	tcp->initial_reply_done = 1;
	tcp->sendbuf->len = 0;
	tcp->sendbuf->done = 0;
	tcp->io = &read_tcp_socket;
	tcp->process = &recv_tcp_query;
	tcp->ioclose = &error_or_close_tcp_client;
	FD_CLR(tcp->fd, &wfds0);
	FD_SET(tcp->fd, &rfds0);
}

static void
error_tcp_query(struct tcpsock *tcp)
{
	upcall_drop_server(&tcp->t);
}

static void
error_or_close_tcp_client(struct tcpsock *tcp)
{
	ntcp_clients--;
}


static int
read_tcp_socket(struct tcpsock *tcp)
{
	struct pbuf *pb;
	ssize_t ret;

	pb = tcp->recvbuf;

	ret = recv(tcp->fd,
	    PBUF_TCP_PTR(pb) + pb->done, pb->size + 2 - pb->done, 0);
	if (ret == -1) {
		LERR("tcp recv: %m");
		return -1;
	}
	if (ret == 0) {
		/*
		 * initial_reply_done is always false on sockets to
		 * authoritative servers.
		 */
		if (pb->done != 0 || !tcp->initial_reply_done)
			LWARN("tcp %d premature close", tcp->fd);
		return -1;
	}

	pb->done += ret;
	return pickup_tcp_msg(tcp);
}

static int
pickup_tcp_msg(struct tcpsock *tcp)
{
	struct pbuf *pb;

	pb = tcp->recvbuf;

	if (pb->len == 0 && pb->done >= 2) {
		pb->len = load16(PBUF_TCP_PTR(pb));
		if (pb->len > pb->size) {
			LWARN("tcp %d too large: %zu", tcp->fd, pb->len);
			return -1;
		}
		pb->len += 2;
	}
	if (pb->len > 0) {
		LDEBUG2("tcp %d %zu/%zu read", tcp->fd, pb->done, pb->len);
		if (pb->done >= pb->len)
			(*tcp->process)(tcp);
	} else
		LDEBUG2("tcp %d %zu/NA read", tcp->fd, pb->done);
	return 0;
}

static int
write_tcp_socket(struct tcpsock *tcp)
{
	struct pbuf *pb;
	ssize_t ret;

	pb = tcp->sendbuf;
	ret = send(tcp->fd, PBUF_TCP_PTR(pb) + pb->done,
	    pb->len - pb->done, 0);
	if (ret == -1) {
		LERR("tcp send: %m");
		return -1;
	}
	pb->done += ret;
	LDEBUG2("tcp %d %zu/%zu written", tcp->fd, pb->done, pb->len);
	if (pb->done >= pb->len) {
		(*tcp->process)(tcp);
		/*
		 * XXX We adjust tcp->recvbuf here, but this should be done
		 * in recv_tcp_query().  We should NOT know the internal
		 * of tcp->process.  Luckily, this does not do harm to
		 * sockets to authoritative servers.
		 */
		memmove(PBUF_TCP_PTR(tcp->recvbuf),
		    PBUF_TCP_PTR(tcp->recvbuf) + tcp->recvbuf->len,
		    tcp->recvbuf->done - tcp->recvbuf->len);
		tcp->recvbuf->done -= tcp->recvbuf->len;
		tcp->recvbuf->len = 0;
		/*
		 * Multiple queries may have been read by one read().
		 * Check if any query is ready before returning to select().
		 */
		return pickup_tcp_msg(tcp);
	}
	return 0;
}

static int
set_nonblocking(int fd)
{
	/* Strictly speaking, we need to do F_GETFL and modify the flag. */
	if (fcntl(fd, F_SETFL, O_NONBLOCK) == -1) {
		LERR("fcntl(O_NONBLOCK): %m");
		return -1;
	}
	return 0;
}


struct pbuf *
get_pbuf_to_query(void)
{
	/* TCP query also uses this. */
	return transient_pbuf;
}

struct pbuf *
get_pbuf_to_reply(struct transport *t)
{
	if (t->is_tcp)
		return ((struct tcpsock *)t)->sendbuf;
	else
		return transient_pbuf;
}

/*
 * This function should be called only for an upstream transport,
 * because ntcp_clients is not taken care of.
 */
void
free_transport(struct transport *t)
{
	if (t->is_tcp)
		free_tcp((struct tcpsock *)t);
	else
		free_udp((struct udpsock *)t);
}

static struct udpsock *
alloc_udp(int fd)
{
	static const struct udpsock udp0;
	struct udpsock *udp;

	STAT_OBJ_UDPSOCK_REQ();
	if ((udp = (struct udpsock *)malloc(sizeof(*udp))) == NULL) {
		LERR("%m");
		STAT_OBJ_UDPSOCK_FAIL();
		return NULL;
	}
	STAT_OBJ_UDPSOCK_ALLOC();
	*udp = udp0;
	udp->fd = fd;

	if (serve_new_fd(fd, &rfds0) != 0) {
		free(udp);
		STAT_OBJ_UDPSOCK_FREE();
		return NULL;
	}
	TAILQ_INSERT_TAIL(&udp_head, udp, next);

	LDEBUG2("udp %d allocated", fd);
	return udp;
}

static void
free_udp(struct udpsock *udp)
{
	LDEBUG2("udp %d freeing", udp->fd);
	if (udp_next == udp)
		udp_next = TAILQ_NEXT(udp_next, next);
	TAILQ_REMOVE(&udp_head, udp, next);
	FD_CLR(udp->fd, &rfds0);
	FD_CLR(udp->fd, &rfds);
	close(udp->fd);
	free(udp);
	STAT_OBJ_UDPSOCK_FREE();
}

static struct tcpsock *
alloc_tcp(int fd, int allocate_pbuf)
{
	static const struct tcpsock tcp0;
	struct tcpsock *tcp;

	STAT_OBJ_TCPSOCK_REQ();
	if ((tcp = (struct tcpsock *)malloc(sizeof(*tcp))) == NULL) {
		LERR("%m");
		STAT_OBJ_TCPSOCK_FAIL();
		return NULL;
	}
	STAT_OBJ_TCPSOCK_ALLOC();
	*tcp = tcp0;

	if (allocate_pbuf) {
		if ((tcp->sendbuf = alloc_pbuf(DNS_MSG_BUF_SIZE)) == NULL)
			goto fail;
		if ((tcp->recvbuf = alloc_pbuf(DNS_MSG_BUF_SIZE)) == NULL)
			goto fail;
	}

	if (serve_new_fd(fd, &rfds0) != 0)		/* XXX or wfds0? */
		goto fail;
	tcp->t.is_tcp = 1;
	tcp->fd = fd;
	TAILQ_INSERT_TAIL(&tcp_head, tcp, next);

	LDEBUG2("tcp %d allocated", fd);
	return tcp;
fail:
	free_pbuf(tcp->recvbuf);
	free_pbuf(tcp->sendbuf);
	free(tcp);
	STAT_OBJ_TCPSOCK_FREE();
	return NULL;
}

static void
free_tcp(struct tcpsock *tcp)
{
	LDEBUG2("tcp %d freeing", tcp->fd);
	if (tcp_next == tcp)
		tcp_next = TAILQ_NEXT(tcp_next, next);
	TAILQ_REMOVE(&tcp_head, tcp, next);
	FD_CLR(tcp->fd, &rfds0);
	FD_CLR(tcp->fd, &wfds0);
	FD_CLR(tcp->fd, &rfds);
	FD_CLR(tcp->fd, &wfds);
	close(tcp->fd);
	free_pbuf(tcp->recvbuf);
	free_pbuf(tcp->sendbuf);
	free(tcp);
	STAT_OBJ_TCPSOCK_FREE();
}

static int
serve_new_fd(int fd, fd_set *set)
{
	if (fd >= FD_SETSIZE) {
		LERR("descriptor too big: %d", fd);
		return -1;
	}
	if (fd > maxfd)
		maxfd = fd;
	FD_SET(fd, set);
	return 0;
}

static int
is_errno_operational(void)
{
	switch (errno) {
	case ENETDOWN:
	case ENETUNREACH:
	case ECONNREFUSED:
	case EHOSTDOWN:
	case EHOSTUNREACH:
		return 1;
	default:
		return 0;
	}
}


int
load_allow_query(const char *addr, int prefixlen)
{
	struct allow_query *alq;

	if ((alq = (struct allow_query *)malloc(sizeof(*alq))) == NULL) {
		LERR("%m");
		return -1;
	}
	if ((alq->sa = str2sadup(addr)) == NULL) {
		free(alq);
		return -1;
	}
	switch (alq->sa->sa_family) {
	case AF_INET:
		if (prefixlen > 32)
			prefixlen = 32;
		break;
	case AF_INET6:
		if (prefixlen > 128)
			prefixlen = 128;
		break;
	}
	alq->prefixlen = prefixlen;
	alq->next = conf_load.allow_query;
	conf_load.allow_query = alq;
	return 0;
}

int
fixate_allow_query(int cancel)
{
	struct allow_query *alq;

	if (!cancel) {
		alq = conf_run.allow_query;
		conf_run.allow_query = conf_load.allow_query;
		conf_load.allow_query = alq;
	}
	while ((alq = conf_load.allow_query) != NULL) {
		conf_load.allow_query = conf_load.allow_query->next;
		free(alq->sa);
		free(alq);
	}
	return 0;
}

static int
check_allow_query(const struct sockaddr *sa)
{
	const struct allow_query *alq;

	for (alq = conf_run.allow_query; alq != NULL; alq = alq->next) {
		if (sa->sa_family != alq->sa->sa_family)
			continue;
		switch (sa->sa_family) {
		case AF_INET:
		    {
			const struct sockaddr_in *s1, *s2;
			s1 = (const struct sockaddr_in *)sa;
			s2 = (const struct sockaddr_in *)alq->sa;
			if (cmp_bits_be(&s1->sin_addr, &s2->sin_addr,
			    alq->prefixlen) != 0)
				continue;
			break;
		    }
		case AF_INET6:
		    {
			const struct sockaddr_in6 *s1, *s2;
			s1 = (const struct sockaddr_in6 *)sa;
			s2 = (const struct sockaddr_in6 *)alq->sa;
			if (s1->sin6_scope_id != s2->sin6_scope_id)
				continue;
			if (cmp_bits_be(&s1->sin6_addr, &s2->sin6_addr,
			    alq->prefixlen) != 0)
				continue;
			break;
		    }
		default:
			continue;
		}
		return 0;
	}
	return -1;
}

/* compare _without_ order */
static int
cmp_bits_be(const void *p1, const void *p2, int bits)
{
	const unsigned char *a1, *a2;

	/* XXX How sockaddr is stored if CHAR_BIT != 0? */
	a1 = p1;
	a2 = p2;
	while (bits >= 8) {
		if (*a1++ != *a2++)
			return 1;
		bits -= 8;
	}
	if (bits == 0)
		return 0;
	return (*a1 ^ *a2) & (0xff00U >> bits);
}


struct sockaddr *
sadup(const struct sockaddr *sa)
{
	struct sockaddr *newsa;

	if ((newsa = (struct sockaddr *)malloc(SA_LEN(sa))) == NULL)
		return NULL;
	return memcpy(newsa, sa, SA_LEN(sa));
}

int
sacmp(const struct sockaddr *sa1, const struct sockaddr *sa2)
{
	int diff;

	if ((diff = sa1->sa_family - sa2->sa_family) != 0)
		return diff;
	if ((diff = SA_LEN(sa1) - SA_LEN(sa2)) != 0)
		return diff;

	switch (sa1->sa_family) {
	case AF_INET:
	    {
		const struct sockaddr_in *s1, *s2;
		s1 = (const struct sockaddr_in *)sa1;
		s2 = (const struct sockaddr_in *)sa2;
		if ((diff = CMP16(s1->sin_port, s2->sin_port)) != 0)
			return diff;
		return memcmp(&s1->sin_addr, &s2->sin_addr,
		    sizeof(struct in_addr));
	    }
	case AF_INET6:
	    {
		const struct sockaddr_in6 *s1, *s2;
		s1 = (const struct sockaddr_in6 *)sa1;
		s2 = (const struct sockaddr_in6 *)sa2;
		if ((diff = CMP16(s1->sin6_port, s2->sin6_port)) != 0)
			return diff;
		if ((diff = CMP16(s1->sin6_scope_id, s2->sin6_scope_id)) != 0)
			return diff;
		return memcmp(&s1->sin6_addr, &s2->sin6_addr,
		    sizeof(struct in6_addr));
	    }
	default:
		LERR("SANITY: unknown address family %d", sa1->sa_family);
		abort();
		/* NOTREACHED */
	}
}

#ifdef NOTUSED
int
sadnscmp(const struct sockaddr *sa, int type, const char *addr)
{
	switch (sa->sa_family) {
	case AF_INET:
		if (type != TYPE_A)
			return 1;		/* XXX */
		return memcmp(&((const struct sockaddr_in *)sa)->sin_addr,
		    addr, 4);
	case AF_INET6:
		if (type != TYPE_AAAA)
			return 1;
		return memcmp(&((const struct sockaddr_in6 *)sa)->sin6_addr,
		    addr, 16);
	default:
		abort();
		/* NOTREACHED */
	}
}

int
serv2num(const char *serv)
{
	struct servent *ent;
	char *endptr;
	long l;

	if ((ent = getservbyname(serv, "udp")) != NULL)
		return ntohs(ent->s_port);
	l = strtol(serv, &endptr, 0);
	if (*endptr == '\0' && 0 < l && l < 65536L)
		return l;
	return 0;
}
#endif

/* XXX thread unsafe */
const char *
sa2str(const struct sockaddr *sa)
{
	static char buf[64];

	if (sa == NULL)
		return "(null: TCP?)";
	if (getnameinfo(sa, SA_LEN(sa), buf, sizeof(buf), NULL, 0,
	    NI_NUMERICHOST) != 0)
		return "(error)";
	return buf;
}

#ifdef DEBUG
/* XXX thread unsafe */
const char *
addr2str(int family, const char *addr)
{
	static char buf[64];

	if (inet_ntop(family, addr, buf, sizeof(buf)) == NULL)
		return "(error)";
	return buf;
}
#endif

/* salen should be socklen_t */
struct sockaddr *
addr2sa(struct sockaddr *sa, size_t salen, unsigned int type, const char *addr)
{
	struct sockaddr_in *sa4;
	struct sockaddr_in6 *sa6;

	switch (type) {
	case TYPE_A:
		if (salen < sizeof(*sa4)) {
			LERR("SANITY: shorter buffer than sockaddr_in");
			return NULL;
		}
		sa4 = (struct sockaddr_in *)sa;
		memset(sa4, 0, sizeof(*sa4));
#ifdef HAVE_SA_LEN
		sa4->sin_len = sizeof(*sa4);
#endif
		sa4->sin_family = AF_INET;
		memcpy(&sa4->sin_addr, addr, sizeof(sa4->sin_addr));
		sa4->sin_port = htons(DNS_DEFAULT_PORT);
		break;
	case TYPE_AAAA:
		if (salen < sizeof(*sa6)) {
			LERR("SANITY: shorter buffer than sockaddr_in6");
			return NULL;
		}
		sa6 = (struct sockaddr_in6 *)sa;
		memset(sa6, 0, sizeof(*sa6));
#ifdef HAVE_SA_LEN
		sa6->sin6_len = sizeof(*sa6);
#endif
		sa6->sin6_family = AF_INET6;
		memcpy(&sa6->sin6_addr, addr, sizeof(sa6->sin6_addr));
		sa6->sin6_port = htons(DNS_DEFAULT_PORT);
		break;
	default:
		LERR("SANITY: invalid type for address %u", type);
		return NULL;
	}
	return sa;
}

struct sockaddr *
str2sadup(const char *addr)
{
	struct addrinfo hints, *res0;
	struct sockaddr *sa;
	int gairet;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_DGRAM;
	hints.ai_protocol = IPPROTO_UDP;
#ifdef AI_NUMERICSERV
	hints.ai_flags = AI_NUMERICHOST | AI_NUMERICSERV;
#else
	hints.ai_flags = AI_NUMERICHOST;
#endif
	gairet = getaddrinfo(addr, NUM2STR(DNS_DEFAULT_PORT), &hints, &res0);
	if (gairet != 0) {
		LERR("%s: %s", addr, gai_strerror(gairet));
		return NULL;
	}
	if ((sa = sadup(res0->ai_addr)) == NULL)
		LERR("%m");
	freeaddrinfo(res0);
	return sa;
}

#ifndef HAVE_SA_LEN
/*
 * Use unsigned int so that net.h has no dependency on sys/socket.h.
 * SUSv2 says socklen_t is unsigned but SUSv3 pulled it back and says nothing
 * about the signedness.  Historically, (signed) int has been used in BSD
 * socket APIs.
 */
unsigned int
compat_sa_len(const struct sockaddr *sa)
{
	switch (sa->sa_family) {
	case AF_INET:
		return sizeof(struct sockaddr_in);
	case AF_INET6:
		return sizeof(struct sockaddr_in6);
	default:
		return 0;
	}
}
#endif
