/* $Id: preload.c 328 2010-04-10 01:46:17Z ken $ */
/*
 * Copyright (c) 2006 KAMADA Ken'ichi.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/param.h>
#include <errno.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>

#include "version.h"
#include "var.h"
#include "pathnames.h"
#include "util.h"
#include "format.h"
#include "net.h"
#include "cache.h"
#include "db.h"
#include "preload.h"

#include "yy_proto.h"

/*
 * One <character-string> + temporary '\0'; multiple <character-string>s
 * with MAXHOSTNAMELEN is overdone.
 */
#define ID_SERVER_LEN (256 + 1)

static int setup_local_zone(struct hint *h);

static int load_conf_file(const char *fn, int must);

static union {
	struct rrset r;
	char strut[offsetof(struct rrset, datav[1])];
} version, hostname;
static char vdbuf[] = " " PROGVERSION;
static char hdbuf[ID_SERVER_LEN];

int
init_builtin_rrset(void)
{
	vdbuf[0] = sizeof(vdbuf) - 2;

	version.r.c.eol = 0;			/* should be time(NULL) */
	version.r.domain = "\007version\006server";
	version.r.type = TYPE_TXT;
	version.r.class = CLASS_CH;
	version.r.datac = 1;
	version.r.datav[0].p = vdbuf;
	version.r.datav[0].l = sizeof(vdbuf) - 1;

	if (gethostname(hdbuf + 1, sizeof(hdbuf) - 1) == -1) {
		LERR("gethostname: %m");
		return -1;
	}
	hdbuf[sizeof(hdbuf) - 1] = '\0';
	hdbuf[0] = strlen(hdbuf + 1);

	hostname.r.c.eol = 0;			/* should be time(NULL) */
	hostname.r.domain = "\002id\006server";
	hostname.r.type = TYPE_TXT;
	hostname.r.class = CLASS_CH;
	hostname.r.datac = 1;
	hostname.r.datav[0].p = hdbuf;
	hostname.r.datav[0].l = (unsigned char)hdbuf[0] + 1;

	return 0;
}

struct rrset *
get_builtin_rrset(const char *domain, unsigned int type, unsigned int class)
{
	if (type == TYPE_TXT && class == CLASS_CH) {
		if (domcmp(version.r.domain, domain) == 0)
			return &version.r;
		if (domcmp(hostname.r.domain, domain) == 0)
			return &hostname.r;
	}

	return NULL;
}

int
load_local_zones(void)
{
	static const char *local_zones[] = {
		/* draft-ietf-dnsop-default-local-zones-12 */
		"10.in-addr.arpa.",
		"16.172.in-addr.arpa.",
		"17.172.in-addr.arpa.",
		"18.172.in-addr.arpa.",
		"19.172.in-addr.arpa.",
		"20.172.in-addr.arpa.",
		"21.172.in-addr.arpa.",
		"22.172.in-addr.arpa.",
		"23.172.in-addr.arpa.",
		"24.172.in-addr.arpa.",
		"25.172.in-addr.arpa.",
		"26.172.in-addr.arpa.",
		"27.172.in-addr.arpa.",
		"28.172.in-addr.arpa.",
		"29.172.in-addr.arpa.",
		"30.172.in-addr.arpa.",
		"31.172.in-addr.arpa.",
		"168.192.in-addr.arpa.",
		"0.in-addr.arpa.",
		"127.in-addr.arpa.",
		"254.169.in-addr.arpa.",
		"2.0.192.in-addr.arpa.",
		"100.51.198.in-addr.arpa.",
		"113.0.203.in-addr.arpa.",
		"255.255.255.255.in-addr.arpa.",
		"0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.ip6.arpa.",
		"1.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.ip6.arpa.",
		"d.f.ip6.arpa.",
		"8.e.f.ip6.arpa.",
		"9.e.f.ip6.arpa.",
		"a.e.f.ip6.arpa.",
		"b.e.f.ip6.arpa.",
		"8.b.d.0.1.0.0.2.ip6.arpa.",
		/* others */
		"localhost.",
		NULL
	};
	char zone[DNS_MAX_NAME_LEN];
	const char **l;
	struct hint *h;

	for (l = local_zones; *l != NULL; l++) {
		if (str2dom(zone, sizeof(zone), *l) != 0) {
			LERR("%s: invalid zone", *l);
			return -1;
		}
		if ((h = alloc_hint(zone, CLASS_IN)) == NULL)
			return -1;
		if (setup_local_zone(h) != 0) {
			LERR("%s: setup_local_zone failed", *l);
			return -1;
		}
	}
	return 0;
}

static int
setup_local_zone(struct hint *h)
{
	struct rrset *r;
	char rdbuf[128];
	size_t rdlen;
	char *p;

	h->local = 1;

	if (h->local_soa == NULL) {
		/* [draft-ietf-dnsop-default-local-zones-12 3] */
		rdlen = enc_make_soa(rdbuf, sizeof(rdbuf),
		    h->zone, "\006nobody\007invalid",
		    1U, 3600U, 1200U, 604800U, 10800U);
		if (rdlen == 0)
			return -1;
		p = malloc(offsetof(struct rrset, datav[1]) + rdlen);
		if (p == NULL)
			return -1;
		r = (struct rrset *)p;
		p += offsetof(struct rrset, datav[1]);
		memcpy(p, rdbuf, rdlen);

		*r = (struct rrset) {
			.domain = h->zone,		/* share */
			.class = h->class,
			.type = TYPE_SOA,
			.datac = 1,
		};
		r->datav[0].p = p;
		r->datav[0].l = rdlen;

		h->local_soa = r;
	}

	return 0;
}

int
load_config(void)
{
	/* Initialize conf_load. */
	conf_load = (struct conf) {
		.cache_size = MIN_CACHE_SIZE,
	};

	/* Load configuration into conf_load. */
	if (load_local_zones() != 0 ||
	    load_conf_file(HINTS_FILE, 1) != 0 ||
	    load_conf_file(CONFIG_FILE, 0) != 0)
		goto fail;

	if (conf_load.listen_on == NULL) {
		if (conf_load.allow_query == NULL) {
			if (load_listen_on("127.0.0.1", option.bind_port) != 0 ||
			    load_listen_on("::1", option.bind_port) != 0)
				goto fail;
		} else {
			if (load_listen_on(NULL, option.bind_port) != 0)
				goto fail;
		}
	}
	if (conf_load.allow_query == NULL) {
		if (load_allow_query("127.0.0.0", 8) != 0 ||
		    load_allow_query("::1", 128) != 0)
			goto fail;
	}

	/* Fixate configuration into conf_run. */
	if (fixate_hints(0) != 0)
		(void)fixate_hints(1);
	if (fixate_listen_on(0) != 0)
		(void)fixate_listen_on(1);
	if (fixate_allow_query(0) != 0)
		(void)fixate_allow_query(1);
	if (conf_run.cache_size == 0)
		conf_run.cache_size = conf_load.cache_size;
	else if (conf_run.cache_size != conf_load.cache_size)
		LWARN("reloading cache-size not supported");
	return 0;
fail:
	(void)fixate_hints(1);
	(void)fixate_listen_on(1);
	(void)fixate_allow_query(1);
	return -1;
}

static int
load_conf_file(const char *fn, int must)
{
	FILE *fp;
	int ret;

	if ((fp = fopen(fn, "r")) == NULL) {
		if (errno == ENOENT && !must)
			return 0;
		LERR("%s: %m", fn);
		return -1;
	}
	rc_set_file(fp, fn);
	ret = yyparse();
	rc_cleanup_tmpstr();
	fclose(fp);
	if (ret != 0)
		return -1;
	return 0;
}
