/* $Id: rcparser.y 341 2011-06-02 00:50:59Z ken $ */
/*
 * Copyright (c) 2006 KAMADA Ken'ichi.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

%{
#include <sys/types.h>
#ifdef HAVE_SYS_QUEUE_H
# include <sys/queue.h>
#else
# include "missing/queue.h"
#endif
#include <sys/socket.h>
#include <errno.h>
#include <inttypes.h>
#include <netdb.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <syslog.h>

#include "var.h"
#include "util.h"
#include "format.h"
#include "net.h"
#include "cache.h"
#include "db.h"

#include "yy_proto.h"

struct tmpstr {
	LIST_ENTRY(tmpstr) next;
	char str[];
};

struct binstr;
static struct tmpstr *alloc_tmpstr(const struct binstr *binstr);
static void free_tmpstr(struct tmpstr *tmpstr);
static const char *str2umax(uintmax_t *um, const char *str);

static LIST_HEAD(, tmpstr) tmpstr_head = LIST_HEAD_INITIALIZER(&tmpstr_head);

%}

%union
{
	const char *str;
	struct binstr {
		const char *p;
		int l;
	} binstr;

	/* XXX too large */
	struct tmpstr *tmpstr;
	char domain[DNS_MAX_NAME_LEN];
	uintmax_t number;
	unsigned int class;
}
%token ZONE HINT CACHE_SIZE LISTEN_ON PORT ALLOW_QUERY
%token <str> NUMBER
%token <binstr> STRING QUOTED

%type <str> port
%type <tmpstr> word address
%type <number> number
%type <domain> domain
%type <class> class

%%
sections
	:	section
	|	sections '\n' section
	;
section
	:	/* empty */
	|	zone_section
	|	cache_size_section
	|	listen_on_section
	|	allow_query_section
	;
zone_section
	:	ZONE domain class '{' hint_list '}'
	;
hint_list
	:	empty empty hint	/* adjust the depth of the stack */
	|	hint_list '\n' hint
	;
hint
	:	/* empty */
	|	HINT address
		{
			struct hint *h;

			/* -4, -3: peeking at the zone_section rule */
			if ((h = alloc_hint($<domain>-4, $<class>-3)) == NULL)
				return -1;
			h->local = 0;
			if ((h->sav[h->sac] = str2sadup($2->str)) == NULL)
				return -1;
			h->sac++;
			free_tmpstr($2);
		}
	;
cache_size_section
	:	CACHE_SIZE number
		{
			conf_load.cache_size = $2;
			if (conf_load.cache_size != $2) {
				yyerror("out of range");
				return -1;
			}
		}
	;
listen_on_section
	:	LISTEN_ON '{' listen_on_list '}'
	;
listen_on_list
	:	listen_on
	|	listen_on_list '\n' listen_on
	;
listen_on
	:	/* empty */
	|	address port
		{
			if (load_listen_on($1->str, $2) != 0) {
				yyerror("load_listen_on failed");
				return -1;
			}
			free_tmpstr($1);
		}
	;
port
	:	/* empty */
		{
			$$ = option.bind_port;
		}
	|	PORT NUMBER
		{
			$$ = $2;
		}
	;
allow_query_section
	:	ALLOW_QUERY '{' allow_query_list '}'
	;
allow_query_list
	:	allow_query
	|	allow_query_list '\n' allow_query
	;
allow_query
	:	/* empty */
	|	address '/' number
		{
			if (load_allow_query($1->str, $3) != 0) {
				yyerror("load_allow_query failed");
				return -1;
			}
			free_tmpstr($1);
		}
	;

class
	:	STRING
		{
			const char *estr;
			uintmax_t um;

			if (strncasecmp($1.p, "IN", 3) == 0)
				$$ = CLASS_IN;
			else if (strncasecmp($1.p, "CLASS", 5) == 0) {
				if ((estr = str2umax(&um, $1.p + 5)) != NULL) {
					yyerror(estr);
					return -1;
				}
				if (um != (uint16_t)um) {
					yyerror("out of range");
					return -1;
				}
				$$ = um;
			} else {
				yyerror("unknown class");
				return -1;
			}
		}
	;
domain
	:	word
		{
			if (str2dom($$, sizeof($$), $1->str) != 0) {
				yyerror("failed to parse as a domain");
				return -1;
			}
			free_tmpstr($1);
		}
	;
address
	:	word
	;
word
	:	STRING
		{
			if (($$ = alloc_tmpstr(&$1)) == NULL) {
				yyerror("out of memory");
				return -1;
			}
		}
	|	QUOTED
		{
			if (($$ = alloc_tmpstr(&$1)) == NULL) {
				yyerror("out of memory");
				return -1;
			}
		}
	;
number
	:	NUMBER
		{
			const char *estr;

			if ((estr = str2umax(&$$, $1)) != NULL) {
				yyerror(estr);
				return -1;
			}
		}
	;
empty
	:	/* empty */
	;

%%

static struct tmpstr *
alloc_tmpstr(const struct binstr *binstr)
{
	struct tmpstr *tmpstr;

	if ((tmpstr = (struct tmpstr *)
	    malloc(sizeof(*tmpstr) + binstr->l + 1)) == NULL)
		return NULL;
	if (*binstr->p == '"') {
		memcpy(tmpstr->str, binstr->p + 1, binstr->l - 2);
		tmpstr->str[binstr->l - 2] = '\0';
	} else
		memcpy(tmpstr->str, binstr->p, binstr->l + 1);
	LIST_INSERT_HEAD(&tmpstr_head, tmpstr, next);
	return tmpstr;
}

void
rc_cleanup_tmpstr(void)
{
	struct tmpstr *tmpstr;
	int i;

	i = 0;
	while ((tmpstr = LIST_FIRST(&tmpstr_head)) != NULL) {
		free_tmpstr(tmpstr);
		i++;
	}
	if (i != 0)
		LDEBUG2("cleanup_tmpstr: %d", i);
}

static void
free_tmpstr(struct tmpstr *tmpstr)
{
	LIST_REMOVE(tmpstr, next);
	free(tmpstr);
}

static const char *
str2umax(uintmax_t *um, const char *str)
{
	char *ep;

	if (str[0] == '0' && str[1] != '\0')
		return "not decimal";

	errno = 0;
	*um = strtoumax(str, &ep, 10);
	if (str[0] == '\0' || *ep != '\0')
		return "not a number";
	/* SUSv3 says strtol() keeps errno on success.  See also C99 7.5 #3. */
	if (errno == ERANGE)
		return "out of range";
	return NULL;
}
