/* $Id: serve.c 341 2011-06-02 00:50:59Z ken $ */
/*
 * Copyright (c) 2006 KAMADA Ken'ichi.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#ifdef HAVE_SYS_QUEUE_H
# include <sys/queue.h>
#else
# include "missing/queue.h"
#endif
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <inttypes.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <time.h>

#include "missing/missing.h"

#include "var.h"
#include "util.h"
#include "stats.h"
#include "format.h"
#include "net.h"
#include "cache.h"
#include "db.h"
#include "preload.h"
#include "serve.h"

/* query from client */
struct qtask {
	TAILQ_ENTRY(qtask) next;	/* XXX inefficient */
	uint32_t tick;			/* XXX tick of explicit queries */

	struct alias {
		char name[DNS_MAX_NAME_LEN];
		time_t eol;		/* eol of the last entry is unused */
	} alias[MAX_ALIAS + 1], *canon;
	/* Current query was sent to the authority of this zone. */
	const char *zone;		/* point into canon->name[] */

	uint16_t flags;
	uint16_t qid;
	uint16_t class;
	uint16_t type;

	enum { STATE_WAITING, STATE_RUNNING, STATE_FINAL } state;
	/* # of outstanding queries (i.e., active servv[].t entries). */
	int nactive;
	int nsent;
	/* # of depending queries (i.e., active depend[] entries). */
	int ndepend;
	struct timespec timeout;
	size_t last_zonelen;
	int loop;

	/* 0x00 = 512 bytes, 0x01 = EDNS0, 0x02 = TCP */
	unsigned int pktsize_level;

	struct serv {
		union {
			struct sockaddr sa;
			struct sockaddr_in sa_in;
			struct sockaddr_in6 sa_in6;
		} addr;		/* not sockaddr_storage to minimize the size */
		unsigned int level;		/* see pktsize_level */
#define PSL_EDNS	0x01
#define PSL_TCP		0x02
		int edns_version;
		struct {
			unsigned int udp_reachable	: 1;
			unsigned int edns_probing	: 1;
			unsigned int edns_retry		: 1;
		} state;
		unsigned int edns_unreach;
#define EDNS_UNREACH_LIMIT		2
		struct timespec start;
		unsigned int rtt;
		struct transport *t;		/* NULL = not used */
	} servv[30];
	int servi, servc;
	struct {
		char ns[DNS_MAX_NAME_LEN];
		int search4;
		int search6;
	} nsv[10];
	int nsi, nsc;

	/* client information */
	struct transport *ct;		/* NULL is spontaneous qtask */
	struct sockaddr *csa;		/* point to caddr (UDP) or NULL (TCP) */
	struct sockaddr_storage caddr;
	uint16_t cid;
	struct {
		int version;
		uint16_t udpsize;
	} edns;

	/* pending operation */
	struct qtask_depend *pending_head;
	struct qtask_depend **pending_tailp;
	struct qtask_depend {
		struct qtask *target;
		struct qtask *this;
		struct qtask_depend *next;
		struct qtask_depend **prevp;
	} depend[4];			/* XXX magical2 */

	/* work area */
	struct {
		unsigned int ns_sub	: 1;
		unsigned int ns_self	: 1;
		unsigned int ns_ooz	: 1;
		unsigned int soa	: 1;
		unsigned int answer	: 1;
	} has;
	uint32_t neg_ttl;
};


static void sched_set_timeout(struct qtask *q, int sec);

static int parse_query(struct qtask *q, const char *pkt, size_t len);
static struct qtask *parse_reply(struct transport *t, struct sockaddr *sa,
    const char *pkt, size_t len);

static int parse_answer(struct qtask *q, struct pktdec *pd);
static int parse_authority(struct qtask *q, struct pktdec *pd,
    const char **soa_owner);
static int parse_additional(struct qtask *q, struct pktdec *pd);
static struct rrset *negative_cache(struct qtask *q, int rcode,
    const char *apex);
static void disable_edns(struct serv *sv, const char *reason);
static void cache_edns_version(struct serv *sv, int edns_version);

static int new_spontaneous(struct qtask *parent,
    const char *domain, unsigned int type, unsigned int class);

static struct qtask *alloc_qtask(struct qtask *parent);
static void free_qtask(struct qtask *q, int fail, int nodata);
static struct qtask *get_qtask(uint16_t id);
static struct qtask *get_qtask_to_depend(struct qtask *parent,
    const char *domain, unsigned int type, unsigned int class);
static struct serv *get_sv_by_t(struct qtask *q, struct transport *t);
static int tick_lessthan(uint32_t d, uint32_t s);
static int qtask_depend(struct qtask *parent, struct qtask *child);
static void qtask_undepend(struct qtask *q);

static void qtask_start_iteration(struct qtask *q);
static void qtask_unblock(struct qtask *q, int fail, int nodata);
static void qtask_next_server(struct qtask *q, int timeout);
static int query_try_next(struct qtask *q, int timeout);
static int make_query(struct qtask *q, struct serv *sv);
static void abort_current_queries(struct qtask *q, int edns_retry);
static void stop_query(struct serv *sv, int cancel);
static int query_to_hint(struct qtask *q, struct hint *h);
static int query_to_authority(struct qtask *q, struct authority *a);
static int find_ns_glue(struct qtask *q,
    const char *name, unsigned int type, unsigned int class);
static int find_ns_addr(struct qtask *q,
    const char *name, unsigned int type, unsigned int class);
static int add_addr_to_qtask(struct qtask *q, const struct sockaddr *sa);
static int add_ns_to_qtask(struct qtask *q, const char *ns, int s4, int s6);
static uint16_t pickup_new_id(void);

static int response(struct qtask *q, struct rrset *r, int rcode,
    struct rrset *soa);
static int add_rrset(struct rrset *r, void *arg);

static TAILQ_HEAD(, qtask) qtask_head = TAILQ_HEAD_INITIALIZER(qtask_head);
static TAILQ_HEAD(, qtask) larval_head = TAILQ_HEAD_INITIALIZER(larval_head);
static uint32_t curtick = 0;


struct timeval *
schedule(struct timeval *tv)
{
	struct timespec cur;
	struct qtask *q, *next;

	(void)clock_gettime(CLOCK_MONOTONIC, &cur);
redo:
	next = NULL;
	TAILQ_FOREACH(q, &qtask_head, next) {
		if (q->state == STATE_WAITING)
			continue;
		if (cur.tv_sec < q->timeout.tv_sec ||
		    (cur.tv_sec == q->timeout.tv_sec &&
		    cur.tv_nsec < q->timeout.tv_nsec)) {
			if (next == NULL ||
			    q->timeout.tv_sec < next->timeout.tv_sec ||
			    (q->timeout.tv_sec == next->timeout.tv_sec &&
			    q->timeout.tv_nsec < next->timeout.tv_nsec))
				next = q;
			continue;
		}
		LDEBUG("[timeout] id=%u <%s,%u,%u>",
		    q->qid, dom2str(q->canon->name), q->type, q->class);
		c_bumptime();
		qtask_next_server(q, 1);
		goto redo;
	}
	if (next == NULL)
		return NULL;
	/* Calculate timeout with converting from timespec to timeval. */
	tv->tv_sec = next->timeout.tv_sec - cur.tv_sec;
	tv->tv_usec = next->timeout.tv_nsec - cur.tv_nsec; /* tv_usec is long */
	if (tv->tv_usec < 0) {
		tv->tv_sec--;
		tv->tv_usec += 1000000000L;
	}
	tv->tv_usec /= 1000;
	if (tv->tv_sec == 0 && tv->tv_usec == 0)
		tv->tv_usec++;				/* avoid spin */
	return tv;
}

static void
sched_set_timeout(struct qtask *q, int sec)
{
	(void)clock_gettime(CLOCK_MONOTONIC, &q->timeout);
	q->timeout.tv_sec += sec;
}

/*
 * Upcall to handle queries from clients.
 * Return value:
 *   0  Success in the aspect of transport (answer or error is returned).
 *  -1  Error (no message is returned).
 */
int
upcall_query(struct transport *t, struct sockaddr *sa, char *buf, size_t len)
{
	struct qtask *q;
	int ret;

	if ((q = alloc_qtask(NULL)) == NULL)
		return -1;
	q->ct = t;
	t->tag = q;
	if (sa != NULL) {
		memcpy(&q->caddr, sa, SA_LEN(sa));
		q->csa = (struct sockaddr *)&q->caddr;
	}
	c_bumptime();
	ret = parse_query(q, buf, len);
	if (ret == -1) {
		free_qtask(q, 1, 0);
		return -1;
	} else if (ret > 0) {
		response(q, NULL, ret, NULL);
		free_qtask(q, 1, 0);
	}
	return 0;
}

/*
 * Upcall to handle responses from authoritative servers.
 */
void
upcall_reply(struct transport *t, struct sockaddr *sa, char *buf, size_t len)
{
	struct qtask *q;

	c_bumptime();
	if ((q = parse_reply(t, sa, buf, len)) != NULL)
		qtask_next_server(q, 0);
}

/*
 * Upcall to close a server transport because of a network error.
 */
void
upcall_drop_server(struct transport *t)
{
	struct qtask *q;
	struct serv *sv;

	q = (struct qtask *)t->tag;
	if ((sv = get_sv_by_t(q, t)) == NULL) {
		LERR("SANITY: dropping unknown server");
		return;
	}
	sv->t = NULL;
	q->nactive--;
	if (!t->is_tcp)			/* XXX penalty */
		add_servcap_rtt(&sv->addr.sa, 200);
	c_bumptime();
	qtask_next_server(q, 0);
}

/*
 * Upcall to drop TCP clients.
 */
void
upcall_drop_tcp_client(struct transport *t)
{
	struct qtask *q;

	if (t->tag == NULL)	/* connected but not yet received a query */
		return;

	/*
	 * We can cancel the qtask here, but completing the current
	 * outstanding query seems better than stopping it.  So change it
	 * to a "spontaneous" qtask.
	 */
	q = (struct qtask *)t->tag;
	q->ct = NULL;
}

/*
 * return -1 for early error, 0 for success, and RCODE.
 */
static int
parse_query(struct qtask *q, const char *pkt, size_t len)
{
	struct pktdec pd;
	struct rrset *reply;
	int ret, rcode;

	LDEBUG2_PKT("query:");
	LDUMP2_PKT(pkt, len);

	if (dec_init(&pd, pkt, len, "[query]") == -1) {
		LDUMP2(pkt, len);
		return -1;
	}

	q->cid = ntohs(pd.hd->id);
	q->flags = ntohs(pd.hd->flags);
	if (q->flags & QR_BIT) {
		LDEBUG("Don't respond to a response");
		return -1;
	}

	if (dec_question(&pd) == -1)	/* XXX no [query] tag on failure */
		return RCODE_FORMERR;
	q->canon = &q->alias[0];
	domcpy(q->canon->name, pd.name);
	q->type = pd.type;
	q->class = pd.class;
	LDEBUG("[query] <%s,%u,%u> from %s",
	    dom2str(q->alias[0].name), q->type, q->class, sa2str(q->csa));
	STAT_D_QUERY(q->alias[0].name, q->type, q->class);

	ret = dec_preparse(&pd);
	rcode = GET_RCODE(q->flags) + (pd.opt.extrcode << EXT_RCODE_SHIFT);
	q->edns.version = pd.opt.version;
	q->edns.udpsize = pd.opt.udpsize;
	if (ret != 0)
		return RCODE_FORMERR;
	if (pd.opt.version > 0) {
		LDEBUG("EDNS%d client is not supported", pd.opt.version);
		return RCODE_BADVERS;		/* [RFC2671 4.6] */
	}
	if (pd.opt.version >= 0)
		LDEBUG2("EDNS%d client", pd.opt.version);

	if (rcode != RCODE_NOERROR) {
		LDEBUG("invalid RCODE: %d", rcode);
		return RCODE_FORMERR;
	}
	if (GET_OPCODE(q->flags) != OPCODE_QUERY) {
		LDEBUG("OPCODE not implemented: %u", GET_OPCODE(q->flags));
		return RCODE_NOTIMP;
	}
	if ((q->flags & (AA_BIT | TC_BIT | RA_BIT | Z_MASK)) != 0)
		LDEBUG2("strange flags: 0x%04x", q->flags);	/* ignore */

	if (pd.countv[SECTION_AN] != 0 || pd.countv[SECTION_NS] != 0 ||
	    pd.countv[SECTION_AR] != 0) {
		LDEBUG("extra section: %d, %d, %d",
		    pd.countv[SECTION_AN], pd.countv[SECTION_NS],
		    pd.countv[SECTION_AR]);
		return RCODE_FORMERR;
	}

	reply = get_builtin_rrset(q->alias[0].name, q->type, q->class);
	if (reply != NULL) {
		LDEBUG2("built-in RRset");
		reply->c.eol = curtime;
		response(q, reply, 0, NULL);
		free_qtask(q, 0, 0);
		return 0;
	}

	if (!(q->flags & RD_BIT)) {
		LDEBUG("non-recursive query is not implemented");
		return RCODE_NOTIMP;
	}
	if (q->type == QTYPE_AXFR || q->type == QTYPE_IXFR ||
	    q->type == QTYPE_MAILB || q->type == QTYPE_MAILA)
		return RCODE_NOTIMP;
	if (q->class == QCLASS_ANY)
		return RCODE_REFUSED;

	qtask_start_iteration(q);
	return 0;
}

/*
 * Return value:
 *  NULL  success or error (ignore silently)
 *  q     error (try next)
 *
 * In the case of TCP, this function must call free_transport().  This is
 * required by recv_tcp_reply() in net.c.
 */
static struct qtask *
parse_reply(struct transport *t, struct sockaddr *sa,
    const char *pkt, size_t len)
{
	struct pktdec pd;
	struct qtask *q;
	struct serv *sv;
	const char *soa_owner;		/* points into q->canon->name */
	uint16_t flags;
	int rcode;

	soa_owner = NULL;

	LDEBUG2_PKT("reply:");
	LDUMP2_PKT(pkt, len);

	q = (struct qtask *)t->tag;
	if ((sv = get_sv_by_t(q, t)) == NULL) {
		LERR("SANITY: transport not found");
		abort();
	}

	if (dec_init(&pd, pkt, len, "[reply]") == -1) {
		LDUMP2(pkt, len);
		goto reject_early;
	}

	LDEBUG("[reply] id=%u <%s,%u,%u>",
	    q->qid, dom2str(q->canon->name), q->type, q->class);

	if (q->qid != ntohs(pd.hd->id)) {
		LDEBUG("qid mismatch %u", ntohs(pd.hd->id));
		goto reject_early;
	}
	/* If UDP (sa != NULL), check the address; if TCP, the stack did. */
	if (sa != NULL && sacmp(&sv->addr.sa, sa) != 0) {
		LDEBUG2("reply from unexpected address: %s", sa2str(sa));
		goto reject_early;	/* TCP does not come here */
	}

	/* update RTT of the address that I received the reply from. */
	stop_query(sv, 0);
	q->nactive--;

	if (sv->level == 0)
		sv->state.udp_reachable = 1;
	if (sv->state.udp_reachable && sv->edns_unreach >= EDNS_UNREACH_LIMIT) {
		disable_edns(sv, "timeout1");
		sv->state.edns_probing = 0;
		sv->edns_unreach = 0;
		/* XXX how to handle RTT? */
	}

	/* Parse the question section and check if it matches the query. */
	if (dec_question(&pd) == -1)
		goto malformed_accepted;
	if (domcmp(q->canon->name, pd.name) != 0 ||
	    q->type != pd.type || q->class != pd.class) {
		LDEBUG("question mismatch <%s,%u,%u>",
		    dom2str(pd.name), pd.type, pd.class);
		return q;
	}
	if (dec_preparse(&pd) != 0)
		goto malformed_accepted;
	flags = ntohs(pd.hd->flags);
	rcode = GET_RCODE(flags) + (pd.opt.extrcode << EXT_RCODE_SHIFT);
	if (pd.opt.version >= 0)
		LDEBUG2("EDNS%d server", pd.opt.version);

	if (GET_OPCODE(flags) != OPCODE_QUERY) {
		LDEBUG("invalid OPCODE: %u", GET_OPCODE(flags));
		return q;
	}
	if (!(flags & QR_BIT)) {
		LDEBUG("invalid QR bit");
		return q;
	}
	if ((flags & (RD_BIT | Z_MASK)) != 0)
		LDEBUG2("unexpected flags: 0x%04x", flags);	/* ignore */
	/* TC_BIT is checked later. */
	/* I don't care RA_BIT. */

	if (sv->level & PSL_EDNS &&
	    (rcode == RCODE_FORMERR || rcode == RCODE_SERVFAIL ||
	     rcode == RCODE_NOTIMP) &&			/* [RFC2671 5.3] */
	    pd.opt.version < 0) {
		disable_edns(sv, "error");
		(void)make_query(q, sv);		/* retry now */
		return NULL;
	}
	if (rcode == RCODE_FORMERR && pd.opt.version >= 0) {
		/*
		 * Workaround for a bogus zone, which returns an FORMERR
		 * _and_ reflects an OPT in response to an EDNS0 query.
		 * (observed around mozilla.org on 2007-02-20)
		 */
		disable_edns(sv, "bogus OPT");
		(void)make_query(q, sv);		/* retry now */
		return NULL;
	}
	if (pd.opt.version >= 0 && sv->edns_version != -1) {
		if (sv->level & PSL_EDNS)
			cache_edns_version(sv, pd.opt.version);
		else
			LDEBUG("OPT replied without request");
	}
	if (flags & TC_BIT) {
		if (sv->level & PSL_TCP) {
			LDEBUG("TC bit in TCP mode");
			return q;
		}
		if (sv->level < q->pktsize_level) {
			LERR("SANITY: level reversal");
			return NULL;
		}
		q->pktsize_level = sv->level + 1;
		LDEBUG2("TC retry %u", q->pktsize_level);
		abort_current_queries(q, 1);
		q->servi = 0;
		return q;
	}
	/* Check this after TC_BIT in case of truncation of OPT. */
	if (pd.opt.version < 0 && sv->level & PSL_EDNS)
		disable_edns(sv, "OPT ignored");

	q->has.ns_sub = q->has.ns_self = q->has.ns_ooz = 0;
	q->has.soa = q->has.answer = 0;
	q->neg_ttl = DNS_NEGCACHE_MAX_TTL;

	dec_from_section(&pd, SECTION_NS);
	if (parse_authority(q, &pd, &soa_owner) != 0)
		goto malformed_accepted;
	if (rcode == RCODE_NXDOMAIN || !q->has.ns_sub) {
		if (!(flags & AA_BIT))
			LDEBUG2("no AA flag");
	} else {
		if (flags & AA_BIT)
			LDEBUG2("unexpected AA flag");
	}

	dec_from_section(&pd, SECTION_AN);
	if (parse_answer(q, &pd) != 0)
		goto malformed_accepted;

	if (rcode == RCODE_NXDOMAIN && q->has.answer) {
		/*
		 * We could know which name is NXDOMAIN, but this
		 * optimization is not yet implemented;
		 * run through qtask_start_iteration() one more time.
		 */
		LDEBUG2("NXDOMAIN with data");
	} else if (rcode != RCODE_NOERROR) {
		LDEBUG2("rcode=%d", rcode);
		/* XXX how to cache servfail? */
		/*
		 * NXDOMAIN is type-agnostic, but I heard some server
		 * (load balancer?) returns NXDOMAIN for AAAA query,
		 * while there is valid A records.  So cache NXDOMAIN with
		 * the type.
		 */
		if (rcode == RCODE_NXDOMAIN) {
			(void)negative_cache(q, RCODE_NXDOMAIN, soa_owner);
			goto reply_cached;
		}
		return q;
	}

	dec_from_section(&pd, SECTION_AR);
	if (parse_additional(q, &pd) != 0)
		goto malformed_accepted;

	/*
	 * If a CNAME that points to an in-zone name is returned
	 * but its RR is not, it may be No Data.
	 * But to say for sure, we need to confirm that is is really in
	 * the zone (not in a child zone) by checking SOA.
	 * In addition, some server returns an answer of this type when
	 * it should be NXDOMAIN [RFC2308 2.2.1] [NSD 3.0.7].
	 * So if there is some data, don't make No Data cache.
	 */
	if (!q->has.answer) {
		if (q->has.ns_self && !q->has.soa) {
			LDEBUG("self referral; ignore this response");
			return q;
		} else if (q->has.ns_ooz) {
			LDEBUG("out-of-zone referral; ignore this response");
			return q;
		}
		/*
		 * If the reply has NS(s) and no SOA, it is not No Data
		 * [RFC2308 2].
		 */
		if (q->has.soa || !q->has.ns_sub) {
			(void)negative_cache(q, 0, soa_owner);
			goto reply_cached;
		}
	}

	/* was a valid reply; proceed to the next iteration */
reply_cached:
	/*
	 * We can do something like below for negative responses, but
	 * qtask_start_iteration() is enough.
	 *	response(q, neg, 0,
	 *	    neg->zone != NULL ?
	 *	    get_rrset(neg->zone, TYPE_SOA, neg->class) : NULL);
	 *	free_qtask(q, 0, 1);
	 *	return NULL;
	 */
	qtask_start_iteration(q);
	return NULL;
reject_early:
	if (sv->t->is_tcp) {
		stop_query(sv, 1);
		q->nactive--;
		return q;		/* try next if TCP */
	}
	return NULL;			/* ignore if UDP */
malformed_accepted:
	/*
	 * If the response for an EDNS query is broken, the
	 * authoritative server, the F/W in-between, or something
	 * may be broken.
	 */
	if (sv->level & PSL_EDNS)
		disable_edns(sv, "malformed");
	(void)make_query(q, sv);		/* retry now */
	return NULL;		/* or dump packet? */
}

static int
parse_answer(struct qtask *q, struct pktdec *pd)
{
	struct rrset *r;

	if (q->has.ns_sub && !dec_is_end(pd))
		LDEBUG2("parse_answer in referral mode");

	while (!dec_is_end(pd)) {
		if (dec_next_rrset(pd) == -1)
			return -1;

		if (domparent(q->zone, pd->name) == NULL) {
			LDEBUG2("<r:%s> out of zone", dom2str(pd->name));
			continue;
		}

		/*
		 * If the requested type is A or AAAA and it is a name
		 * server (i.e., the parent knows its address as a glue),
		 * it will be returned in the answer section.
		 */
		if (q->has.ns_sub &&
		    !(pd->type == TYPE_A || pd->type == TYPE_AAAA || pd->type == TYPE_CNAME)) {
			LDEBUG2("ignore type %u (referral mode)", pd->type);
			continue;
		}

		if (pd->type != q->type && pd->type != TYPE_CNAME &&
		    q->type != QTYPE_ANY) {
			LDEBUG2("answer type mismatch: %d", pd->type);
			continue;
		}
		if (pd->class != q->class) {
			LDEBUG2("answer class mismatch: %d", pd->class);
			continue;
		}
		if (domcmp(q->canon->name, pd->name) == 0)
			q->has.answer = 1;

		r = new_rrset(pd->name, pd->type, pd->class, pd->ttl,
		    pd->count);
		do {
			if (dec_parse_rr(pd) == -1)
				return -1;
			c_checkttl(&r->c, pd->ttl);

			/* CNAME is validated in dec_parse_rdata(). */
			if (dec_parse_rdata(pd) == -1)
				return -1;
			r->datav[r->datac].l = pd->rdlen;
			r->datav[r->datac].p = c_extcpy(pd->rd, pd->rdlen);
			r->datac++;
			LDEBUG2("<r:%s,%u,%u>",
			    dom2str(r->domain), pd->type, pd->class);
		} while (pd->count > 0);
		if (q->has.ns_sub && r->type != TYPE_CNAME)
			replace_rrset_as_glue(r);
		else
			replace_rrset(r);
	}
	return 0;
}

static int
parse_authority(struct qtask *q, struct pktdec *pd, const char **soa_owner)
{
	struct authority *a;
	struct rrset *r;

	while (!dec_is_end(pd)) {
		if (dec_next_rrset(pd) == -1)
			return -1;

		/* XXX ooz glue is useful when asking to this zone... */
		if (domparent(q->zone, pd->name) == NULL) {
			LDEBUG2("<a:%s> out of zone", dom2str(pd->name));
			if (pd->type == TYPE_NS)
				q->has.ns_ooz = 1;
			continue;
		}
		if (pd->class != q->class) {
			LDEBUG2("answer class mismatch: %d", pd->class);
			continue;
		}

		switch (pd->type) {
		case TYPE_NS:
			if (domcmp(q->zone, pd->name) == 0) {
				LDEBUG2("<a:%s:-> arrived", dom2str(pd->name));
				q->has.ns_self = 1;
				continue;
			} else
				q->has.ns_sub = 1;

			a = new_authority(pd->name, pd->class, pd->ttl,
			    pd->count);
			do {
				if (dec_parse_rr(pd) == -1)
					return -1;
				c_checkttl(&a->c, pd->ttl);

				/* Validate NS (pd->rd). */
				if (dec_parse_rdata(pd) == -1)
					continue;	/* XXX or fail? */
				a->nsv[a->nsc] = c_extcpy(pd->rd, pd->rdlen);
				a->nsc++;
				LDEBUG2("<a:%s:%s>",
				    dom2str(a->zone), dom2str(pd->rd));
			} while (pd->count > 0);
			replace_authority(a);
			break;
		case TYPE_SOA:
			/* SOA is a singleton, so no loop */
			if (dec_parse_rr(pd) == -1)
				return -1;
			if (dec_parse_soa_minimum(pd) == -1)
				continue;		/* XXX or fail? */
			if (dec_parse_rdata(pd) == -1)
				continue;

			if (q->has.soa)
				continue;
			q->has.soa = 1;
			q->neg_ttl = pd->soa_minimum <= pd->ttl ?
			    pd->soa_minimum : pd->ttl;
			/*
			 * Remember SOA's owner name as a pointer into
			 * q->canon->name, because pd->name is destroyed soon.
			 * It is used only for negative cache.
			 * In the current implementation, ncache is created
			 * only when there is no answer returned, in which
			 * case, soa_owner must be q->canon->name.
			 * Otherwise, soa_owner may be incorrect or NULL,
			 * but it is not used later.
			 */
			*soa_owner = domparent(pd->name, q->canon->name);

			/* A heuristic to prevent SOAs from wasting cache... */
			/* XXX This causes SOA to expire earlier. */
			r = get_rrset(pd->name, pd->type, pd->class);
			if (r != NULL && r->c.eol - curtime > pd->ttl / 2) {
				LDEBUG2("keep cached SOA for %s",
				    dom2str(pd->name));
				continue;
			}

			r = new_rrset(pd->name, pd->type, pd->class, pd->ttl,
			    1);
			r->datav[0].l = pd->rdlen;
			r->datav[0].p = c_extcpy(pd->rd, pd->rdlen);
			r->datac++;
			LDEBUG2("<r:%s,%u,%u> SOA",
			    dom2str(pd->name), pd->type, pd->class);
			replace_rrset(r);
			break;
		default:
			LDEBUG2("ignore type=%d in authority", pd->type);
			break;
		}
	}
	return 0;
}

static int
parse_additional(struct qtask *q, struct pktdec *pd)
{
	struct glue *g;
	int addrlen, addrfamily, addrchar;

	while (!dec_is_end(pd)) {
		if (dec_next_rrset(pd) == -1)
			return -1;

		if (domparent(q->zone, pd->name) == NULL) {
			LDEBUG2("<g:%s> out of zone", dom2str(pd->name));
			continue;
		}
		if (pd->class != q->class) {
			LDEBUG2("answer class mismatch: %d", pd->class);
			continue;
		}

		/*
		 * Don't treat additional data (especially glues) as
		 * authoritative.  They may be truncated without the TC
		 * bit set [draft-ietf-dnsop-respsize-10 2.1], though
		 * it is forbidden [RFC2181 9].
		 * The parent server may not know all A/AAAA RRs.
		 */
		switch (pd->type) {
		case TYPE_A:
			addrlen = 4;
			addrfamily = AF_INET;
			addrchar = '4';
			break;
		case TYPE_AAAA:
			addrlen = 16;
			addrfamily = AF_INET6;
			addrchar = '6';
			break;
		default:
			LDEBUG2("glue type mismatch: %d", pd->type);
			continue;
		}

		g = new_glue(pd->name, pd->type, pd->class, pd->ttl,
		    pd->count);
		do {
			if (dec_parse_rr(pd) == -1)
				return -1;
			c_checkttl(&g->c, pd->ttl);

			if (pd->rdlen != addrlen) {
				LDEBUG("size of address != %d", addrlen);
				continue;
			}
			g->datav[g->datac].l = pd->rdlen;
			g->datav[g->datac].p = c_extcpy(pd->rd, pd->rdlen);
			g->datac++;
			LDEBUG2("<g%c:%s:%s>", addrchar, dom2str(g->domain),
			    addr2str(addrfamily, pd->rd));
		} while (pd->count > 0);
		replace_glue(g);
	}
	return 0;
}

/*
 * Create negative cache for q->canon->name in a zone denoted by apex.
 * That is, don't call this function if you received some CNAMEs (You
 * need to reflect CNAMEs into q and to make sure the apex is correct).
 */
static struct rrset *
negative_cache(struct qtask *q, int rcode, const char *apex)
{
	struct rrset *r;
	uint32_t ttl;

	if (q->has.answer)
		LERR("SANITY: might be creating a wrong negative cache");

	ttl = q->neg_ttl < DNS_NEGCACHE_MIN_TTL ? DNS_NEGCACHE_MIN_TTL :
	    DNS_NEGCACHE_MAX_TTL < q->neg_ttl ? DNS_NEGCACHE_MAX_TTL :
	    q->neg_ttl;
	r = new_rrset(q->canon->name, q->type, q->class, ttl, 0);
	r->rcode = rcode;
	if (apex != NULL) {
		/* Remember apex as a pointer into r->domain. */
		r->zone = domparent(apex, r->domain);
		if (r->zone == NULL)
			LERR("SANITY: invalid apex for negative cache");
	} else {
		/*
		 * A zone is expected below q->zone, but negative response
		 * without SOA was received.  Leave r->zone being NULL.
		 */
		LDEBUG2("negative caching without SOA for %s in %s",
		     dom2str(q->canon->name), dom2str(q->zone));
		/*
		 * Use shorter TTL.  RFC 2308 discourage caching of
		 * itself [RFC2308 5], though.
		 */
		r->c.eol = curtime + DNS_NEGCACHE_TTL_WITHOUT_SOA;
	}
	replace_rrset(r);
	return r;
}

static void
disable_edns(struct serv *sv, const char *reason)
{
	LDEBUG("disable EDNS (%s): %s", reason, sa2str(&sv->addr.sa));
	cache_edns_version(sv, -1);
}

static void
cache_edns_version(struct serv *sv, int edns_version)
{
	sv->edns_version = edns_version;
	set_servcap_edns_version(&sv->addr.sa, edns_version);
}


void
invoke_spontaneous(void)
{
	struct qtask *q;

	while ((q = TAILQ_FIRST(&larval_head)) != NULL) {
		TAILQ_REMOVE(&larval_head, q, next);
		TAILQ_INSERT_TAIL(&qtask_head, q, next);
		LDEBUG("[spontaneous] %p <%s,%u,%u>",
		    q, dom2str(q->alias[0].name), q->type, q->class);
		c_bumptime();
		qtask_start_iteration(q);
	}
}

/* XXX we may need depend distance limit. */
static int
new_spontaneous(struct qtask *parent,
    const char *domain, unsigned int type, unsigned int class)
{
	struct qtask *q;

	LDEBUG2("new_spontaneous: %s,%u,%u", dom2str(domain), type, class);

	if ((q = get_qtask_to_depend(parent, domain, type, class)) != NULL) {
		if (q == parent)
			return 1;
		LDEBUG2("--> only add depend (already q)");
		return qtask_depend(parent, q);
	}

	if ((q = alloc_qtask(parent)) == NULL)
		return -1;
	q->canon = &q->alias[0];
	domcpy(q->canon->name, domain);
	q->type = type;
	q->class = class;
	/*
	 * Even if qtask_depend failed, don't stop the new qtask <q>
	 * for the next trial.
	 */
	return qtask_depend(parent, q);
}


static struct qtask *
alloc_qtask(struct qtask *parent)
{
	static const struct qtask q0;
	struct qtask *q;

	STAT_OBJ_QTASK_REQ();
	if ((q = (struct qtask *)malloc(sizeof(*q))) == NULL) {
		LERR("alloc_qtask: %m");
		STAT_OBJ_QTASK_FAIL();
		return NULL;
	}
	STAT_OBJ_QTASK_ALLOC();
	*q = q0;
	q->tick = parent == NULL ? curtick++ : parent->tick;
	q->loop = parent == NULL ? MAX_LOOP_COUNT : parent->loop;
	q->pending_tailp = &q->pending_head;

	if (parent == NULL)
		TAILQ_INSERT_TAIL(&qtask_head, q, next);
	else
		TAILQ_INSERT_TAIL(&larval_head, q, next);
	return q;
}

/*
 * We should almost always call response() before free_qtask() in order
 * not to leave a TCP client waiting for an unsent reply.  The only
 * exception is when the query was severely broken, in which case
 * upcall_query() returns -1 and the caller closes the connection.
 */
static void
free_qtask(struct qtask *q, int fail, int nodata)
{
	struct qtask_depend *p;

	LDEBUG2("freeing id=%u (%s)", q->qid, fail ? "fail" : "success");

	/* update RTT for un-answered queries */
	abort_current_queries(q, 0);

	if (q->ct != NULL)
		q->ct->tag = NULL;

	qtask_undepend(q);
	TAILQ_REMOVE(&qtask_head, q, next);
	/* invoke parent(s) */
	while ((p = q->pending_head) != NULL) {
		/* Unlink the first parent, */
		if ((q->pending_head = q->pending_head->next) == NULL)
			q->pending_tailp = &q->pending_head;
		p->target = NULL;
		p->this->ndepend--;
		/* and invoke it. */
		LDEBUG("[invoked by free (%s)] id=%u",
		    fail ? "fail" : "success", p->this->qid);
		qtask_unblock(p->this, fail, nodata);
	}
	free(q);
	STAT_OBJ_QTASK_FREE();
}

void
info_query(void)
{
	struct qtask *q;
	int i;

	LINFO("Tick: %" PRIu32, curtick);
	LINFO("Query Info:");
	TAILQ_FOREACH(q, &qtask_head, next) {
		LINFO(" T%" PRIu32 ": %p q%d d%d [0]%s [%td]%s",
		    q->tick, q, q->nactive, q->ndepend,
		    dom2str(q->alias[0].name),
		    q->canon - q->alias, dom2str(q->canon->name));
		for (i = 0; i < lengthof(q->depend); i++)
			if (q->depend[i].target != NULL)
				LINFO("  dep:%p", q->depend[i].target);
	}
	TAILQ_FOREACH(q, &larval_head, next)
		LINFO(" L: %p", q);
}

static struct qtask *
get_qtask(uint16_t id)
{
	struct qtask *q;

	TAILQ_FOREACH(q, &qtask_head, next)
		if (q->qid == id)
			return q;
	return NULL;
}

/*
 * Get a qtask that has the triplet <domain, type, class>.
 * Return value:
 *  != NULL && != parent  Found.
 *  == parent             Same triplet in the same tick is found.  Dead end.
 *  == NULL               Not found.
 */
static struct qtask *
get_qtask_to_depend(struct qtask *parent,
    const char *domain, unsigned int type, unsigned int class)
{
	struct qtask *q, *retq;

	retq = NULL;
	TAILQ_FOREACH(q, &qtask_head, next) {
		if (q == parent)
			continue;
		if (domcmp(q->canon->name, domain) == 0 &&
		    q->type == type &&
		    q->class == class) {
			if (q->tick == parent->tick) {
				LDEBUG2("duplicate dependency: %p <%s,%u,%u>",
				    q, dom2str(domain), type, class);
				return parent;
			}
			if (q->qid == 0)
				continue;
			if (tick_lessthan(q->tick, parent->tick))
				retq = q;
			else
				LDEBUG2("reversed tick: %u<--%u: %p <%s,%u,%u>",
				    q->tick, parent->tick,
				    q, dom2str(domain), type, class);
		}
	}
	return retq;
}

static struct serv *
get_sv_by_t(struct qtask *q, struct transport *t)
{
	struct serv *sv;

	for (sv = q->servv; sv < q->servv + q->servc; sv++)
		if (sv->t == t)
			return sv;
	return NULL;
}

/* check tick in get_qtask_to_depend() to avoid cyclic dependencies */
static int
tick_lessthan(uint32_t d, uint32_t s)
{
	return (uint32_t)(d - s) > 0x80000000UL;
	/* return s != d && (uint32_t)(s - d) < 0x80000000UL; */
}

/* Of course, a parent is a dependant :-) */
static int
qtask_depend(struct qtask *parent, struct qtask *child)
{
	int i;

	if (parent == child) {
		LERR("SANITY: bad dependence");
		return -1;
	}

	/* If already depending, do nothing. */
	for (i = 0; i < lengthof(parent->depend); i++)
		if (parent->depend[i].target == child)
			return 0;

	/* search a vacant slot */
	for (i = 0; i < lengthof(parent->depend); i++)
		if (parent->depend[i].target == NULL)
			break;
	if (i == lengthof(parent->depend)) {
		LERR("SANITY: too many depends");
		return -1;
	}
	/* insert to the pending list */
	parent->depend[i].target = child;
	parent->depend[i].this = parent;
	parent->depend[i].prevp = child->pending_tailp;
	parent->depend[i].next = NULL;
	*child->pending_tailp = &parent->depend[i];
	child->pending_tailp = &parent->depend[i].next;
	parent->ndepend++;
	return 0;
}

static void
qtask_undepend(struct qtask *q)
{
	int i;

	/* unregister myself from the child(s) */
	for (i = 0; i < lengthof(q->depend); i++) {
		if (q->depend[i].target == NULL)
			continue;
		/* adjust link (pending_head is also adjusted) */
		*q->depend[i].prevp = q->depend[i].next;
		if (q->depend[i].next != NULL)
			q->depend[i].next->prevp = q->depend[i].prevp;
		else
			q->depend[i].target->pending_tailp = q->depend[i].prevp;
		q->depend[i].target = NULL;
	}
	q->ndepend = 0;
}


static void
qtask_start_iteration(struct qtask *q)
{
	struct qtask *sameq;
	struct alias *alias;
	struct hint *h;
	struct authority *a;
	struct rrset *r, *soa;
	const char *h_zone, *a_zone;
	size_t zonelen;
	int ret;

	soa = NULL;

restart_iteration:
	/* this should be the last resort and internal error */
	if (--q->loop <= 0) {
		LINFO("loop count exceeded");
		goto servfail;
	}

	/* Search the answer in the cache (follow CNAMEs if needed). */
	for (;;) {
		/* search answer */
		r = get_rrset(q->canon->name, q->type, q->class);
		if (r != NULL)
			goto success;
		/* search CNAME */
		r = get_rrset(q->canon->name, TYPE_CNAME, q->class);
		if (r == NULL)
			break;	/* No answer nor CNAME found, need query */
		LDEBUG2("%s is an alias for %s",
		    dom2str(q->canon->name), dom2str(r->datav[0].p));
		for (alias = &q->alias[0]; alias <= q->canon; alias++)
			if (domcmp(alias->name, r->datav[0].p) == 0)
				goto cname_loop;	/* loop detected */
		if (q->canon + 1 >= endof(q->alias)) {
			LDEBUG("too long CNAME chain for %s",
			    dom2str(q->alias[0].name));
			goto servfail;
		}
		q->canon->eol = r->c.eol;
		q->canon++;
		domcpy(q->canon->name, r->datav[0].p);
		q->last_zonelen = 0;
	}

	q->qid = 0;
	q->state = STATE_WAITING;
	abort_current_queries(q, 0);
	qtask_undepend(q);
	q->servi = q->servc = 0;
	q->nsi = q->nsc = 0;
	q->pktsize_level = 0;
	q->nsent = 0;

	/* If the same outstanding query is found, piggyback */
	sameq = get_qtask_to_depend(q, q->canon->name, q->type, q->class);
	if (sameq != NULL) {
		if (sameq == q)
			goto servfail;
		if (qtask_depend(q, sameq) == 0) {
			LDEBUG2("piggyback on id=%u", sameq->qid);
			return;
		}
	}

	/* search authority and hint */
	a = get_authority(&a_zone, q->canon->name, q->class);
	h = get_hint(&h_zone, q->canon->name, q->class);
	if (h == NULL && a == NULL) {
		/* Root hints should always match. */
		LINFO("hint not found");
		goto servfail;
	}

	/*
	 * Both h_zone and a_zone point into q->canon->name,
	 * so we can compare them.  Use the longer one.
	 */
	LDEBUG2("query to %s (a/%p: %s, h/%p: %s)",
	    h_zone <= a_zone ? "hint" : "authority",
	    a, a ? dom2str(a->zone) : "-", h, h ? dom2str(h->zone) : "-");
	if (h_zone <= a_zone) {
		if (h->local) {
			LDEBUG("%s: unconfigured local zone", dom2str(h->zone));
			h->local_soa->c.eol = curtime + LOCALZONE_SOA_TTL;
			if (q->canon->name < h_zone) {
				soa = h->local_soa;
				goto nxdomain_local;
			}
			/*
			 * XXX Currently, multiple RRsets cannot be returned
			 * on QTYPE_ANY in local zones due to limitation of
			 * response().
			 */
			if (q->type == TYPE_SOA || q->type == QTYPE_ANY)
				r = h->local_soa;
			else
				soa = h->local_soa;
			goto success_local;
		}
		q->zone = h_zone;
		ret = query_to_hint(q, h);
	} else {
		q->zone = a_zone;
		ret = query_to_authority(q, a);
		if (ret == 1) {		/* XXX magic return code */
			LDEBUG("renew authority for zone %s", dom2str(a->zone));
			a->c.eol = 0;
			goto restart_iteration;
		}
	}
	if (ret != 0)
		goto servfail;

	zonelen = domlen(q->zone);
	if (zonelen <= q->last_zonelen) {
		LINFO("zonelen didn't get longer");
		goto servfail;
	}
	q->last_zonelen = zonelen;
	if ((q->qid = pickup_new_id()) == 0) {
		LINFO("failed to get qid");
		goto servfail;
	}
	if (query_try_next(q, 0) == -1)
		goto servfail;
	return;

success:
	response(q, r, 0,
	    r->datac == 0 && r->zone != NULL ?
	    get_rrset(r->zone, TYPE_SOA, r->class) : NULL);
	free_qtask(q, 0, r->datac == 0);
	return;
success_local:
	response(q, r, 0, soa);
	free_qtask(q, 0, 1);	/* locally served RRsets have no A/AAAA */
	return;
cname_loop:
	response(q, r, 0, NULL);
	free_qtask(q, 0, 1);
	return;
nxdomain_local:
	response(q, NULL, RCODE_NXDOMAIN, soa);
	free_qtask(q, 0, 1);
	return;
servfail:
	response(q, NULL, RCODE_SERVFAIL, NULL);
	free_qtask(q, 1, 0);
}

static void
qtask_unblock(struct qtask *q, int fail, int nodata)
{
	int i, ret;

	if (q->servc == 0 && q->nsc == 0) {
		/* this is a piggyback */
		if (fail)
			goto servfail;
		qtask_start_iteration(q);
		return;
	}
	if ((fail || nodata) && (q->nactive > 0 || q->ndepend > 0))
		return;			/* wait success */
	for (i = 0; i < q->nsi; i++) {
		/* See if some of nsv[] have been resolved. */
		if (q->nsv[i].search4 || q->nsv[i].search6)
			LDEBUG2("checking addresses for %s",
			    dom2str(q->nsv[i].ns));
		if (q->nsv[i].search4) {
			/* search glue as well? */
			ret = find_ns_addr(q, q->nsv[i].ns,
			    TYPE_A, q->class);
			q->nsv[i].search4 = ret == -1;
		}
		if (q->nsv[i].search6) {
			ret = find_ns_addr(q, q->nsv[i].ns,
			    TYPE_AAAA, q->class);
			q->nsv[i].search6 = ret == -1;
		}
	}
	if (query_try_next(q, 0) == -1)
		goto servfail;
	return;
servfail:
	response(q, NULL, RCODE_SERVFAIL, NULL);
	free_qtask(q, 1, 0);
}

static void
qtask_next_server(struct qtask *q, int timeout)
{
	struct serv *sv;
	int i;

	if (timeout) {
		/* For servers/firewalls that ignore EDNS messages. */
		for (i = 0; i < q->servc; i++) {
			sv = &q->servv[i];
			if (sv->t == NULL ||
			    sv->level & PSL_TCP ||
			    !sv->state.edns_probing)
				continue;
			if (sv->edns_unreach >= EDNS_UNREACH_LIMIT) {
				sv->state.edns_probing = 0;
				continue;
			}
			if (++sv->edns_unreach < EDNS_UNREACH_LIMIT ||
			    !sv->state.udp_reachable) {
				LDEBUG2("EDNS retry 0x%x", sv->edns_unreach);
			} else {
				disable_edns(sv, "timeout2");
				sv->state.edns_probing = 0;
				sv->edns_unreach = 0;
				/* Stop UDP and retry with TCP. */
				/* XXX how to handle RTT? */
				stop_query(sv, 1);
				q->nactive--;
			}
			(void)make_query(q, sv);
		}
	}

	if (query_try_next(q, timeout) == -1) {
		response(q, NULL, RCODE_SERVFAIL, NULL);
		free_qtask(q, 1, 0);
	}
}

/*
 * Return value:
 *   0  Success (and timeout was set.)
 *  -1  Fail (and timeout was not set, thus the caller MUST free_qtask().)
 */
static int
query_try_next(struct qtask *q, int timeout)
{
	struct serv *sv;
	int i, ntrial, ret;

	ntrial = 0;
try_next:
	if (q->nsent >= MAX_CONTACT) {
		/* When contacted enough servers, give up before trying all. */
		/* If there are some depending job, wait them. */
		for (i = 0; i < lengthof(q->depend); i++) {
			if (q->depend[i].target != NULL) {
				LDEBUG("still waiting for id=%u",
				    q->depend[i].target->qid);
				sched_set_timeout(q, TIMEOUT_PENDING);
				return 0;
			}
		}
		/* XXX should be before the above clause? */
		/* If there are some job for ENDS probing, wait them. */
		for (i = 0; i < q->servc; i++) {
			if (q->servv[i].t != NULL &&
			    q->servv[i].state.edns_probing) {
				sched_set_timeout(q, TIMEOUT_TRY_NEXT);
				return 0;
			}
		}
		if (q->state == STATE_FINAL && !timeout && q->nactive > 0)
			return 0;		/* wait more */
		if (q->state != STATE_FINAL && q->nactive > 0) {
			q->state = STATE_FINAL;
			sched_set_timeout(q, TIMEOUT_FINAL);
			return 0;
		}
		LDEBUG("all servers failed <%s,%u,%u> zone=%s",
		    dom2str(q->canon->name), q->type, q->class,
		    dom2str(q->zone));
		return -1;
	}

	q->state = STATE_RUNNING;

	if (q->servi >= q->servc - 1 && q->nsi < q->nsc &&
	    q->ndepend <= lengthof(q->depend) - 2) {
		int hope;

		/* If remaining addr is 1, start resolving ooz NS in advance. */
		/* XXX
		 * if this is an in-zone name, we may not be able to
		 * resolve this (because we are searching the NS to ask to ...),
		 * so we need to ask NS to the parent and get glue.
		 */
		hope = 0;
		if (q->nsv[q->nsi].search4 &&
		    new_spontaneous(q, q->nsv[q->nsi].ns, TYPE_A,
		    q->class) == 0)
			hope++;
		if (q->nsv[q->nsi].search6 &&
		    new_spontaneous(q, q->nsv[q->nsi].ns, TYPE_AAAA,
		    q->class) == 0)
			hope++;
		q->nsi++;
		if (hope == 0) {
			LDEBUG("new_spontaneous failed; next");
			goto try_next;
		}
	}
	if (q->servc == 0 && q->ndepend == 0) {
		LDEBUG("no qtask to wait");
		return -1;
	}

	if (q->servi >= q->servc && q->nsi >= q->nsc) {
		/* retransmit from the start */
		LDEBUG2("one more round");
		q->servi = 0;
	}
	if (q->servi >= q->servc) {
		/* I cannot query to any server for now, wait to be resolved. */
		sched_set_timeout(q, TIMEOUT_TRY_NEXT);
		return 0;
	}
	if (ntrial++ >= q->servc) {
		/*
		 * Immediate retries more than q->servc will be useless,
		 * or worse, may cause an infinite loop, so wait for
		 * resending (and resolution of spontaneous qtasks if any).
		 * Increment q->nsent to avoid waiting forever.
		 */
		q->nsent++;
		sched_set_timeout(q, TIMEOUT_TRY_NEXT);
		return 0;
	}
	sv = &q->servv[q->servi++];

	if (sv->t != NULL) {
		if (sv->t->is_tcp ||
		    sv->state.edns_retry ||
		    sv->state.edns_probing) {
			sv->state.edns_retry = 0;
			goto try_next;
		}
	}
	sv->state.edns_retry = 0;

	ret = make_query(q, sv);
	if (ret < 0)
		return -1;
	if (ret > 0)
		goto try_next;		/* early retry is better than timeout */
	if (!(sv->level & PSL_TCP)) {
		(void)clock_gettime(CLOCK_MONOTONIC, &sv->start);
		start_servcap_rtt(&sv->addr.sa, &sv->start);
	}
	sched_set_timeout(q, TIMEOUT_TRY_NEXT);
	/*
	 * nsent is incremented here instead of in make_query()
	 * to count only successful send, not count EDNS retry, etc.
	 */
	q->nsent++;
	return 0;
}

/*
 * Return value:
 *   1  error (try next)
 *   0  success
 *  -1  error (fatal)
 */
static int
make_query(struct qtask *q, struct serv *sv)
{
	struct pbuf *pb;
	struct pktenc pe;

	sv->level = q->pktsize_level;
#if 1
	/* Use EDNS if necessary. */
	if (sv->edns_version >= 0 && sv->level == 0)
		sv->level |= PSL_EDNS;
#else
	/* Use EDNS by default. */
	if (sv->edns_version != -1 && sv->level == 0)
		sv->level |= PSL_EDNS;
#endif
	if (sv->edns_version == -1 && sv->level & PSL_EDNS)
		sv->level = PSL_TCP;
	if (sv->edns_unreach >= EDNS_UNREACH_LIMIT)
		sv->level &= ~PSL_EDNS;

	pb = get_pbuf_to_query();
	if (enc_init(&pe, PBUF_DATA(pb), pb->size,
	    sv->level & PSL_EDNS ? 0 : -1, DNS_MAX_UDP_SIZE,
	    sv->level & PSL_TCP) != 0)
		goto nobuf;

	/* Add header. */
	pe.hd->id = htons(q->qid);
	enc_header(&pe, RCODE_NOERROR);

	/* Add question section. */
	if (enc_question(&pe, q->canon->name, q->type, q->class) != 0)
		goto nobuf;

	if (enc_fini(&pe) != 0)
		goto nobuf;

	LDEBUG("sending id=%u <%s,%u,%u> lv=%u/%u to %s",
	    q->qid, dom2str(q->canon->name), q->type, q->class,
	    sv->level, q->pktsize_level, sa2str(&sv->addr.sa));

	if (sv->t == NULL) {
		sv->t = open_query(q, &sv->addr.sa, sv->level & PSL_TCP);
		if (sv->t == NULL)
			goto fail_try_next;
		q->nactive++;
	} else if (sv->t->is_tcp || sv->level & PSL_TCP) {
		LERR("SANITY: TCP retry");
		return -1;
	}
	if (send_query(sv->t, pe.p - pe.buf) != 0) {
		free_transport(sv->t);
		sv->t = NULL;
		q->nactive--;
		goto fail_try_next;
	}
	LDEBUG2_PKT("query sent:");
	LDUMP2_PKT(pe.buf, pe.p - pe.buf);

	return 0;
fail_try_next:
	add_servcap_rtt(&sv->addr.sa, 100);	/* XXX penalty */
	return 1;
nobuf:
	/* should not occur */
	LINFO("query too large");
	return -1;
}

static void
abort_current_queries(struct qtask *q, int edns_retry)
{
	int i;

	for (i = 0; i < q->servc; i++) {
		if (q->servv[i].t == NULL) {
			/* If this address is unused, decay the RTT. */
			if (q->servv[i].start.tv_sec == 0 &&
			    !(q->servv[i].level & PSL_TCP))
				decay_servcap_rtt(&q->servv[i].addr.sa);
			continue;
		}
		if (edns_retry && q->pktsize_level <= q->servv[i].level) {
			q->servv[i].state.edns_retry = 1;
			continue;
		}
		stop_query(&q->servv[i], 1);
		q->nactive--;
	}
}

static void
stop_query(struct serv *sv, int cancel)
{
	free_transport(sv->t);
	sv->t = NULL;

	if (sv->level & PSL_TCP)
		return;
	stop_servcap_rtt(&sv->addr.sa, &sv->start, cancel);
}

/*
 * Return value:
 *   0  Success.
 *  -1  Fail (no server was found).
 */
static int
query_to_hint(struct qtask *q, struct hint *h)
{
	int i;

	for (i = 0; i < h->sac; i++)
		if (add_addr_to_qtask(q, h->sav[i]) != 0)
			break;
	if (q->servc == 0) {
		LINFO("no server found");
		return -1;
	}
	return 0;
}

/*
 * To protect <a>, no cache allocation functions should be called
 * in this function (and subsequent ones).
 * Return value:
 *   0  Success.
 *  -1  Fail (no server was found).
 *   1  This authority needs to be updated to renew glues.
 */
static int
query_to_authority(struct qtask *q, struct authority *a)
{
	int i, ret4, ret6, glue_count;
	int glueless;

	glue_count = 0;

	/* find glue and ooz-server */
	for (i = 0; i < a->nsc; i++) {
		LDEBUG2("searching addresses for %s%s", dom2str(a->nsv[i]),
		    domparent(a->zone, a->nsv[i]) != NULL ? "" : " (ooz-NS)");
		ret4 = find_ns_glue(q, a->nsv[i], TYPE_A, a->class);
		ret6 = find_ns_glue(q, a->nsv[i], TYPE_AAAA, a->class);
		glueless = ret4 == -1 && ret6 == -1;
		if (ret4 == -1)
			ret4 = find_ns_addr(q, a->nsv[i], TYPE_A, a->class);
		else
			glue_count += ret4;
		if (ret6 == -1)
			ret6 = find_ns_addr(q, a->nsv[i], TYPE_AAAA, a->class);
		else
			glue_count += ret6;
		if (glueless && (ret4 == -1 || ret6 == -1)) {
			if (add_ns_to_qtask(q, a->nsv[i],
			    ret4 == -1, ret6 == -1) != 0)
				break;
		}
	}
	if (q->servc == 0 && q->nsc == 0) {
		LINFO("no server found");
		return -1;
	}
	if (glue_count < a->glue_count) {
		/*
		 * Some glues have expired and the authority needs to be
		 * renewed.  This seems a bit inefficient at the first glance,
		 * but actually, better than querying for each NS.
		 * In addition, this is somewhat robust against broken (or
		 * misconfigured) authoritative servers.
		 */
		LDEBUG("%s: glue expiration detected", dom2str(a->zone));
		return 1;	/* should be before updating last_zonelen */
	}
	a->glue_count = glue_count;

	return 0;
}

/*
 * Don't call cache allocation functions, since this function is called
 * from query_to_authority().
 * Return value:
 *  -1  not cached
 *  >0  cached
 */
static int
find_ns_glue(struct qtask *q,
    const char *name, unsigned int type, unsigned int class)
{
	struct sockaddr_storage ss;
	struct sockaddr *sa;
	struct glue *g;
	int i, found;

	g = get_glue(name, type, class);
	if (g == NULL)
		return -1;
	found = 0;
	LDEBUG2("%d %s(s) found (glue)",
	    g->datac, type == TYPE_A ? "A" : "AAAA");
	for (i = 0; i < g->datac; i++) {
		sa = addr2sa((struct sockaddr *)&ss, sizeof(ss),
		    type, g->datav[i].p);
		if (sa == NULL)
			continue;
		if (add_addr_to_qtask(q, sa) != 0)
			break;
		found++;
	}
	return found;
}

/*
 * Don't call cache allocation functions, since this function is called
 * from query_to_authority().
 * Return value:
 *  -1  not cached
 *   0  negatively cached (No Data or NXDOMAIN)
 *  >0  cached
 */
static int
find_ns_addr(struct qtask *q,
    const char *name, unsigned int type, unsigned int class)
{
	struct sockaddr_storage ss;
	struct sockaddr *sa;
	struct rrset *r;
	int i, found;

	/* follow CNAME to rescue broken zone */
	for (i = 0; i < MAX_ALIAS; i++) {
		r = get_rrset(name, type, class);
		if (r != NULL)		/* don't check r->datac */
			break;
		r = get_rrset(name, TYPE_CNAME, class);
		if (r == NULL || r->datac == 0)
			return -1;
		name = r->datav[0].p;
		LDEBUG2("CNAME (--> %s) found", dom2str(name));
	}
	if (i == MAX_ALIAS)
		return -1;
	if (r == NULL)
		return -1;

	found = 0;
	LDEBUG2("%d %s(s) found (non-glue)",
	    r->datac, type == TYPE_A ? "A" : "AAAA");
	for (i = 0; i < r->datac; i++) {
		sa = addr2sa((struct sockaddr *)&ss, sizeof(ss),
		    type, r->datav[i].p);
		if (sa == NULL)
			continue;
		if (add_addr_to_qtask(q, sa) != 0)
			break;
		found++;
	}
	return found;
}

static int
add_addr_to_qtask(struct qtask *q, const struct sockaddr *sa)
{
	unsigned int rtt;
	int i;

	if (q->servc >= lengthof(q->servv)) {
		LDEBUG("too many server addresses");
		return -1;
	}

	/* sort by rtt */
	rtt = get_servcap_rtt(sa) + random_stat(50);
	for (i = q->servi; i < q->servc; i++)
		if (rtt < q->servv[i].rtt)
			break;
	memmove(&q->servv[i + 1], &q->servv[i],
	    (char *)&q->servv[q->servc] - (char *)&q->servv[i]);
	q->servc++;
	q->servv[i].edns_version = get_servcap_edns_version(sa);
	q->servv[i].edns_unreach = 0;
	q->servv[i].state.udp_reachable = 0;
	q->servv[i].state.edns_probing = 0;
	q->servv[i].state.edns_retry = 0;
	q->servv[i].level = 0;
	q->servv[i].rtt = rtt;
	q->servv[i].start.tv_sec = 0;
	q->servv[i].t = NULL;
	if (sizeof(q->servv[i].addr) < SA_LEN(sa)) {
		LERR("SANITY: shorter buffer than SA_LEN");
		return -1;
	}
	memcpy(&q->servv[i].addr, sa, SA_LEN(sa));
	return 0;
}

static int
add_ns_to_qtask(struct qtask *q, const char *ns, int s4, int s6)
{
	int i;

	if (q->nsc >= lengthof(q->nsv)) {
		LDEBUG("too many server names");
		return -1;
	}

	if (q->nsc > 0) {
		/* randomize */
		i = random_stat(q->nsc);
		memmove(&q->nsv[i + 1], &q->nsv[i],
		    (char *)&q->nsv[q->nsc] - (char *)&q->nsv[i]);
	} else
		i = 0;
	q->nsc++;
	domcpy(q->nsv[i].ns, ns);
	q->nsv[i].search4 = s4;
	q->nsv[i].search6 = s6;
	return 0;
}

static uint16_t
pickup_new_id(void)
{
	uint16_t id;
	int i;

	/* XXX inefficient */
	for (i = 0; i < 10; i++) {
		/* Don't use 0 as an outstanding ID. */
		id = random_secure(65535) + 1;
		if (get_qtask(id) == NULL)
			return id;
	}
	return 0;
}

/*
 * Add answer section if <r> != NULL.
 * <r->rcode> can be overridden by <rcode>.
 * If <soa> and <r> are given, negative cache TTL (SOA TTL) is taken
 * from <r> not <soa>.
 */
static int
response(struct qtask *q, struct rrset *r, int rcode, struct rrset *soa)
{
	struct pbuf *pb;
	struct pktenc pe;
	struct alias *alias;
	uint32_t ttl;
	const char *have_answer;

	if (r != NULL && rcode == 0)
		rcode = r->rcode;
	LDEBUG2("%s %d %s", r != NULL ? "answer" : "error", rcode,
	    q->ct != NULL ? "to client" : "(spontaneous)");
	if (q->ct == NULL)		/* spontaneous (or dropped TCP) */
		return 0;

	pb = get_pbuf_to_reply(q->ct);
	if (enc_init(&pe, PBUF_DATA(pb), pb->size,
	    q->edns.version, q->edns.udpsize, q->csa == NULL) != 0)
		goto nobuf;

	/* Add header. */
	pe.hd->id = htons(q->cid);
	pe.hd->flags = htons((q->flags & (OPCODE_MASK | RD_BIT)) |
	    QR_BIT | RA_BIT);
	enc_header(&pe, rcode);

	/* Add question section. */
	if (enc_question(&pe, q->alias[0].name, q->type, q->class) != 0)
		goto nobuf;

	/* Add answer section. */
	/* Preface CNAMEs [RFC1034 4.3.1] */
	for (alias = &q->alias[0]; alias < q->canon; alias++) {
		/* CNAMEs can be expired. */
		ttl = alias->eol >= curtime ? alias->eol - curtime : 0;
		if (enc_rr(&pe, SECTION_AN,
		    alias->name, TYPE_CNAME, q->class, ttl,
		    (alias + 1)->name, domlen((alias + 1)->name)) != 0)
			goto nobuf;
	}
	/* Answers */
	have_answer = pe.p;
	if (r != NULL) {
		/*
		 * XXX r->c.destructor == NULL means locally generated RRsets,
		 * which are not managed by db.c/splaytree.c.
		 */
		if (q->type == QTYPE_ANY && r->c.destructor != NULL) {
			if (foreach_rrset_any(r, &add_rrset, &pe) != 0)
				goto nobuf;
		} else {
			if (add_rrset(r, &pe) != 0)
				goto nobuf;
		}
	}
	if (have_answer == pe.p)
		have_answer = NULL;

	/* Return SOA to clients [RFC2308 6]. */
	if (soa != NULL) {
		if (soa->datac != 0) {
			/* TTL of negative cache (not SOA) is signaled. */
			if (r != NULL)
				ttl = r->c.eol - curtime;	/* negcache */
			else
				ttl = soa->c.eol - curtime;	/* localzone */
			pe.ctl.optional = 1;
			if (enc_rr(&pe, SECTION_NS,
			    soa->domain, soa->type, soa->class, ttl,
			    soa->datav[0].p, soa->datav[0].l) != 0)
				goto nobuf;
		} else
			LDEBUG2("SOA absent where apex expected (%s in %s)",
			    dom2str(q->canon->name), dom2str(soa->domain));
	}

	if (enc_fini(&pe) != 0)
		return -1;
	(void)send_reply(q->ct, q->csa, pe.p - pe.buf);
	LDEBUG2_PKT("response sent:");
	LDUMP2_PKT(pe.buf, pe.p - pe.buf);
	STAT_D_RESPONSE(q->alias[0].name, q->type, q->class, rcode,
	    have_answer != NULL);
	return 0;
nobuf:
	LINFO("response too large");
	return -1;
}

static int
add_rrset(struct rrset *r, void *arg)
{
	struct pktenc *pe;
	uint32_t ttl;
	int i;

	pe = (struct pktenc *)arg;
	ttl = r->c.eol - curtime;
	for (i = 0; i < r->datac; i++)
		if (enc_rr(pe, SECTION_AN, r->domain, r->type, r->class, ttl,
		    r->datav[i].p, r->datav[i].l) != 0)
			return -1;
	return 0;
}
