/* $Id: splaytree.c 209 2007-11-09 23:22:18Z ken $ */
/*
 * Copyright (c) 2006 KAMADA Ken'ichi.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Top-down splay tree.
 *
 * References
 * [1] D.D. Sleator and R.E. Tarjan, "Self-Adjusting Binary Search Trees",
 *     JACM, Volume 32, Issue 3, pp. 652--686, July 1985.
 */

#include <stdlib.h>

#include "splaytree.h"

static int splay(struct splay_descr *d, struct splay_node *key);

struct splay_node *
splay_find(struct splay_descr *d, struct splay_node *s)
{
	if (d->root == NULL)
		return NULL;
	if (splay(d, s) == 0)
		return d->root;
	else
		return NULL;
}

/* This function must not be called with an empty tree. */
int
splay_find_leq(struct splay_descr *d, struct splay_node *s)
{
	int ret;

#if 0
	if (d->root == NULL)
		return;
#endif
	if ((ret = splay(d, s)) <= 0)
		return ret;
	return splay(d, s);
}

struct splay_node *
splay_replace(struct splay_descr *d, struct splay_node *n)
{
	struct splay_node *old;
	int ret;

	old = NULL;
	if (d->root != NULL) {
		ret = splay(d, n);
		if (ret == 0) {
			/* replace */
			old = d->root;
			n->left = d->root->left;
			n->right = d->root->right;
		} else if (ret < 0) {
			/* insert */
			n->left = d->root;
			n->right = d->root->right;
			d->root->right = NULL;
		} else {
			/* insert */
			n->right = d->root;
			n->left = d->root->left;
			d->root->left = NULL;
		}
	} else
		n->left = n->right = NULL;
	d->root = n;
	return old;
}

struct splay_node *
splay_remove(struct splay_descr *d, struct splay_node *s)
{
	struct splay_node *t;

	if (d->root == NULL)
		return NULL;
	if (splay(d, s) == 0) {
		t = d->root;
		if (t->left == NULL)
			d->root = t->right;
		else if (t->right == NULL)
			d->root = t->left;
		else {
			d->root = t->left;
			(void)splay(d, s);
			d->root->right = t->right;
		}
		return t;
	}
	return NULL;
}

struct splay_node *
splay_range_start(struct splay_descr *d,
    struct splay_range *range, struct splay_node *min, struct splay_node *max)
{
	if (d->root == NULL)
		return NULL;

	/* Search the first and last nodes in the range. */
	if (splay(d, max) > 0 && splay(d, max) > 0)
		return NULL;
	max = d->root;
	if (splay(d, min) < 0 && splay(d, min) < 0)
		return NULL;
	min = d->root;

	if ((*d->cmp)(min, max) > 0)
		return NULL;		/* no node in the range */

	/* min SHOULD be d->root (not a MUST thanks to re-splaying). */
	range->stack[0] = min;
	range->max = max;
	range->top = range->bottom = 0;
	return range->stack[0];
}

struct splay_node *
splay_range_next(struct splay_descr *d, struct splay_range *range)
{
	struct splay_node *t;

	if ((t = range->stack[range->top]) == range->max)
		return NULL;

	if (t->right == NULL) {
		if (range->top != range->bottom) {
			/* pop */
			range->top--;
			range->top %= SPLAY_RANGE_STACKSIZE;
			return range->stack[range->top];
		}
		/* Buried too deep in the stack to remember; re-splay. */
		(void)splay(d, t);
	}
	range->stack[range->top] = t = t->right;
	while (t->left != NULL) {
		/* push */
		range->top++;
		range->top %= SPLAY_RANGE_STACKSIZE;
		range->stack[range->top] = t = t->left;
		/* If the stack is full, forget the oldest entry. */
		if (range->top == range->bottom) {
			range->bottom++;
			range->bottom %= SPLAY_RANGE_STACKSIZE;
		}
	}
	return t;
}

static int
splay(struct splay_descr *d, struct splay_node *key)
{
	struct splay_node *left, *right;
	struct splay_node **lt, **gt;
	struct splay_node *t;
	int ret;

	t = d->root;
	lt = &left;
	gt = &right;
	while ((ret = (*d->cmp)(t, key)) != 0) {
		if (ret > 0) {
			if (t->left == NULL)
				break;
			ret = (*d->cmp)(t->left, key);
			if (ret > 0 && t->left->left != NULL) {
				/* zig-zig */
				*gt = t->left;
				t->left = (*gt)->right;
				(*gt)->right = t;
				t = (*gt)->left;
				gt = &(*gt)->left;
			} else if (ret < 0 && t->left->right != NULL) {
				/* zig-zag */
				*gt = t;
				gt = &t->left;
				*lt = t->left;
				lt = &t->left->right;
				t = t->left->right;
			} else {
				/* zig & break */
				*gt = t;
				gt = &t->left;
				t = t->left;
				break;
			}
		} else {
			if (t->right == NULL)
				break;
			ret = (*d->cmp)(t->right, key);
			if (ret < 0 && t->right->right != NULL) {
				/* zig-zig */
				*lt = t->right;
				t->right = (*lt)->left;
				(*lt)->left = t;
				t = (*lt)->right;
				lt = &(*lt)->right;
			} else if (ret > 0 && t->right->left != NULL) {
				/* zig-zag */
				*lt = t;
				lt = &t->right;
				*gt = t->right;
				gt = &t->right->left;
				t = t->right->left;
			} else {
				/* zig & break */
				*lt = t;
				lt = &t->right;
				t = t->right;
				break;
			}
		}
	}
	*lt = t->left;
	*gt = t->right;
	t->left = left;
	t->right = right;
	d->root = t;
	return ret;
}
