/* $Id: trivium.c 275 2009-01-09 03:15:50Z ken $ */
/*
 * Copyright (c) 2007 KAMADA Ken'ichi.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Trivium [1] stream cipher.
 *
 * In this implementation, all I/O values are in network byte order.  Bits
 * are numbered from the most significant bit of the most significant byte.
 * For example, trivium_init() takes K_1 from the MSB(it) of key[0].
 *
 * IP status <http://www.ecrypt.eu.org/stream/phase3ip.html#trivium>:
 *     "To the author's knowledge Trivium is not covered by any patent.
 *     Trivium is also free available for any use."
 *
 * Reference
 * [1] De Canni\`ere, C. and Preneel, B., "Trivium -- Specifications",
 *     ECRYPT Stream Cipher Project (eSTREAM), Report 2005/030, 2005,
 *     <http://www.ecrypt.eu.org/stream/ciphers/trivium/trivium.pdf>.
 */

#include <sys/types.h>
#include <stdint.h>
#include <string.h>

#include "trivium.h"

/*
 * SETUP_PAPER_ORIGINAL
 *   - Load (K_1 .. K_80) into (s_1 .. s_80).
 *   - This initialization is specified in the original paper [1].
 * SETUP_PAPER_REVERSE
 *   - Load (K_80 .. K_1) into (s_1 .. s_80).
 *   - This is a little simpler for software implementations,
 *     because K/IV and z have the same bit order with this.
 *
 * SETUP_PAPER_REVERSE generates the same result with the test vector if
 * 10-byte key, 10-byte IV, and each 4-byte word in the output are all
 * byte-swapped.  However, that byte ordering seems strange to me and
 * SETUP_PAPER_REVERSE does not do that.
 */
#define SETUP_PAPER_ORIGINAL		0
#define SETUP_PAPER_REVERSE		1

struct trivium_ctx {
	/* s1 contains s_1 (in LSB) to s_32 (in MSB), and so on. */
	uint32_t s1, s33, s65;			/* s_1 .. s_93 */
	uint32_t s94, s126, s158;		/* s_94 .. s_177 */
	uint32_t s178, s210, s242, s274;	/* s_178 .. s_288 */
};

/*
 * When n is a constant, a decent compiler will optimize out unused S_?to?()s.
 */
#define S(n)								\
	(65 <= n && n <= 93 ? S_65to93(n) :				\
	158 <= n && n <= 177 ? S_158to177(n) :				\
	242 <= n && n <= 272 ? S_242to272(n) :				\
	274 <= n && n <= 288 ? S_274to288(n) :				\
	trivium_no_such_func() /* error */)
extern uint32_t trivium_no_such_func(void);
#define S_65to93(n)	((ctx.s65 << (96 - n)) ^ (ctx.s33 >> (n - 64)))
#define S_158to177(n)	((ctx.s158 << (189 - n)) ^ (ctx.s126 >> (n - 157)))
#define S_242to272(n)	((ctx.s242 << (273 - n)) ^ (ctx.s210 >> (n - 241)))
#define S_274to288(n)	((ctx.s274 << (305 - n)) ^ (ctx.s242 >> (n - 273)))

#if SETUP_PAPER_ORIGINAL
static uint32_t load32le_bitrev(const char *p);
#endif
#if SETUP_PAPER_REVERSE
static uint32_t load32(const char *p);
#endif

static struct trivium_ctx ctx;

void
trivium_init(const char *key, size_t keylen, const char *iv, size_t ivlen)
{
	char buf[12];
	int i;

	memset(buf, 0, sizeof(buf));
#if SETUP_PAPER_ORIGINAL
	memcpy(buf, key, keylen > 10 ? 10 : keylen);
	ctx.s1 = load32le_bitrev(buf);
	ctx.s33 = load32le_bitrev(buf + 4);
	ctx.s65 = load32le_bitrev(buf + 8);
#elif SETUP_PAPER_REVERSE
	memcpy(buf + 2, key, keylen > 10 ? 10 : keylen);
	ctx.s1 = load32(buf + 8);
	ctx.s33 = load32(buf + 4);
	ctx.s65 = load32(buf);
#endif

	memset(buf, 0, sizeof(buf));
#if SETUP_PAPER_ORIGINAL
	memcpy(buf, iv, ivlen > 10 ? 10 : ivlen);
	ctx.s94 = load32le_bitrev(buf);
	ctx.s126 = load32le_bitrev(buf + 4);
	ctx.s158 = load32le_bitrev(buf + 8);
#elif SETUP_PAPER_REVERSE
	memcpy(buf + 2, iv, ivlen > 10 ? 10 : ivlen);
	ctx.s94 = load32(buf + 8);
	ctx.s126 = load32(buf + 4);
	ctx.s158 = load32(buf);
#endif

	ctx.s178 = ctx.s210 = ctx.s242 = 0;
	ctx.s274 = 0x7000;

	/* Discard 4 full cycles (288 bits == 9 32-bit words). */
	for (i = 0; i < 4 * 9; i++)
		(void)trivium_random32();
}

uint32_t
trivium_random32(void)
{
	uint32_t t1, t2, t3;
	uint32_t z;

	/*
	 * for i = 1 to N do
	 *    t_1 <-- s_66 ^ s_93
	 *    t_2 <-- s_162 ^ s_177
	 *    t_3 <-- s_243 ^ s_288
	 *    z_i <-- t_1 ^ t_2 ^ t_3
	 *    t_1 <-- t_1 ^ s_91 s_92 ^ s_171
	 *    t_2 <-- t_2 ^ s_175 s_176 ^ s_264
	 *    t_3 <-- t_3 ^ s_286 s_287 ^ s_69
	 *    (s_1, s_2, ..., s_93) <-- (t_3, s_1, ..., s_92)
	 *    (s_94, s_95, ..., s_177) <-- (t_1, s_94, ..., s_176)
	 *    (s_178, s_179, ..., s_288) <-- (t_2, s_178, ..., s_287)
	 * end for
	 */

	/*
	 * Note that s_i is rotated as s_{i+1} <-- s_i, so the
	 * variables t? and z has the opposite bit ordering to ctx.s?.
	 */
	t1 = S(66) ^ S(93);
	t2 = S(162) ^ S(177);
	t3 = S(243) ^ S(288);

	/* z contains z_n (in MSB) to z_{n+31} (in LSB). */
	z = t1 ^ t2 ^ t3;

	t1 ^= (S(91) & S(92)) ^ S(171);
	t2 ^= (S(175) & S(176)) ^ S(264);
	t3 ^= (S(286) & S(287)) ^ S(69);

	ctx.s65 = ctx.s33;
	ctx.s33 = ctx.s1;
	ctx.s1 = t3;

	ctx.s158 = ctx.s126;
	ctx.s126 = ctx.s94;
	ctx.s94 = t1;

	ctx.s274 = ctx.s242;
	ctx.s242 = ctx.s210;
	ctx.s210 = ctx.s178;
	ctx.s178 = t2;

	return z;
}

#if SETUP_PAPER_ORIGINAL
static uint32_t
load32le_bitrev(const char *p)
{
	uint32_t r;
	unsigned int c;
	int i, j;

	r = 0;
	for (i = 0; i < 4; i++) {
		c = (unsigned char)p[3 - i];
		for (j = 0; j < 8; j++) {
			r = (r << 1) | (c & 1);
			c >>= 1;
		}
	}
	return r;
}
#endif

#if SETUP_PAPER_REVERSE
static uint32_t
load32(const char *p)
{
	return ((uint32_t)(unsigned char)p[0] << 24) +
	    ((uint32_t)(unsigned char)p[1] << 16) +
	    ((uint32_t)(unsigned char)p[2] << 8) +
	    (unsigned char)p[3];
}
#endif


#if 0
/*
 * Set 1, vector#  0:
 *                       key = 80000000000000000000
 *                        IV = 00000000
 *             stream[0..63] = 7B75CECC2079BD99885A239A9FFC5112
 *                             55A6F0AF4EEEC87E2821D4BF08E6DA86
 *                             EBA76E8002C1AE58F488FA163EAAC3C5
 *                             F2C9BCB13F0B44BB9F9C34C43C7E0ABE
 */
#include <stdio.h>
int
main()
{
	uint32_t z;
	int i, j;

	trivium_init("\0\0\0\0\0\0\0\0\0\x80", 10, "", 0);
	for (i = 0; i < 16; i++) {
		z = trivium_random32();
		for (j = 0; j < 4; j++) {
			printf("%02x", z & 0xff);
			z >>= 8;
		}
		if (i % 4 == 3)
			printf("\n");
	}
	return 0;
}
#endif
