/* $Id: util.c 341 2011-06-02 00:50:59Z ken $ */
/*
 * Copyright (c) 2006 KAMADA Ken'ichi.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <pwd.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>

#include "missing/missing.h"

#include "var.h"
#include "pathnames.h"
#include "trivium.h"
#include "util.h"

static void lose_pidfile(void);

static char *pidfile = NULL;
static int pidfile_fd = -1;

int
run_as(const char *user)
{
	struct passwd *pw;

	if ((pw = getpwnam(user)) == NULL) {
		LERR("%s: user not found", user);
		return -1;
	}
	if (setgid(pw->pw_gid) == -1) {
		LERR("setgid: %m");
		return -1;
	}
	if (setgroups(1, &pw->pw_gid) == -1) {
		LERR("setgroups: %m");
		return -1;
	}
	if (setuid(pw->pw_uid) == -1) {
		LERR("setuid: %m");
		return -1;
	}
	return 0;
}

int
make_pidfile(void)
{
	const char *myname;
	size_t len;

	if (pidfile != NULL || pidfile_fd != -1) {
		LERR("make_pidfile() is called twice");
		return -1;
	}

	myname = getprogname();
	len = strlen(PIDFILE_DIR) + 1 + strlen(myname) + 4 + 1;
	if ((pidfile = (char *)malloc(len)) == NULL) {
		LERR("out of memory");
		return -1;
	}
	snprintf(pidfile, len, "%s/%s.pid", PIDFILE_DIR, myname);

	/*
	 * On BSD systems, use flock(2); otherwize, use fcntl(2).
	 * The fcntl locks are not inherited by a child process, so
	 * fcntl locking is done in write_pidfile().
	 */
#ifdef O_EXLOCK
	if ((pidfile_fd = open(pidfile,
	    O_WRONLY | O_CREAT | O_EXLOCK | O_NONBLOCK, 0644)) == -1) {
		LERR("%s: open: %m", pidfile);
		lose_pidfile();
		return -1;
	}
#else
	if ((pidfile_fd = open(pidfile, O_WRONLY | O_CREAT, 0644)) == -1) {
		LERR("%s: open: %m", pidfile);
		lose_pidfile();
		return -1;
	}
#endif

	return 0;
}

int
write_pidfile(void)
{
	char pidstr[16];
	pid_t pid;

#ifndef O_EXLOCK
	/* .l_start and .l_len are 0 to lock the entire file. */
	if (fcntl(pidfile_fd, F_SETLK,
	    &(struct flock) { .l_whence = SEEK_SET, .l_type = F_WRLCK }
	    ) == -1) {
		LERR("%s: fcntl(F_SETLK): %m", pidfile);
		lose_pidfile();
		return -1;
	}
#endif

	if (ftruncate(pidfile_fd, 0) == -1) {
		LERR("%s: ftruncate: %m", pidfile);
		return -1;
	}
	pid = getpid();
	snprintf(pidstr, sizeof(pidstr), "%d\n", (int)pid);
	if (write(pidfile_fd, pidstr, strlen(pidstr)) == -1) {
		LERR("%s: write: %m", pidfile);
		return -1;
	}
	return 0;
}

void
cleanup_pidfile(void)
{
	struct stat sb1, sb2;

	if (pidfile == NULL && pidfile_fd == -1)
		return;
	if (pidfile == NULL || pidfile_fd == -1) {
		LERR("inconsistent PID file info");
		return;
	}

	/* NB: race between stat() and unlink() */
	if (fstat(pidfile_fd, &sb1) == -1)
		LERR("fstat PID file failed: %m");
	else if (stat(pidfile, &sb2) == -1)
		LERR("%s: stat: %m", pidfile);
	else if (sb1.st_dev != sb2.st_dev || sb1.st_ino != sb2.st_ino)
		LWARN("PID file is replaced; exiting without unlinking it");
	else {
		/* If root priv has been dropped, unlink will fail. */
		if (unlink(pidfile) == -1 &&
		    ftruncate(pidfile_fd, 0) == -1)
			LERR("%s: ftruncate: %m", pidfile);
	}

	free(pidfile);
	pidfile = NULL;
	close(pidfile_fd);
	pidfile_fd = -1;
}

/* Unlike cleanup_pidfile(), this will never unlink the file. */
static void
lose_pidfile(void)
{
	if (pidfile_fd != -1)
		close(pidfile_fd);
	pidfile_fd = -1;
	free(pidfile);
	pidfile = NULL;
}


void
init_random(void)
{
	struct timeval cur;
	uint32_t key[5];

	gettimeofday(&cur, NULL);
	key[0] = htonl(cur.tv_sec);
	key[1] = htonl(cur.tv_usec);
	key[2] = htonl(getpid());
	key[3] = htonl(getppid());
	/* key[4] */

	trivium_init((char *)key, 10, (char *)key + 10, 10);
}

unsigned int
random_stat(unsigned int n)
{
	return trivium_random32() % n;
}

unsigned int
random_secure(unsigned int n)
{
	return trivium_random32() % n;
}


#ifdef LOG_PERROR
#define compat_vsyslog vsyslog
#else
#include <errno.h>

static void compat_vsyslog(int priority, const char *fmt, va_list ap);
static int log_to_stderr = 0;

void
logmsg_stderr(int enable)
{
	log_to_stderr = enable;
}

static void
compat_vsyslog(int priority, const char *fmt, va_list ap)
{
	char buf[1024], *p;
	va_list ap_copy;
	const char *estr, *e;

	if (!log_to_stderr) {
		vsyslog(priority, fmt, ap);
		return;
	}

	estr = strerror(errno);
	va_copy(ap_copy, ap);
	vsyslog(priority, fmt, ap_copy);
	va_end(ap_copy);

	p = buf;
	*p++ = ' ';
	while (*fmt != '\0') {
		if (*fmt == '%' && *(fmt + 1)== 'm') {
			/* embed the error message with escape */
			for (e = estr; *e != '\0'; e++) {
				if (p > endof(buf) - 1)
					goto nobuf;
				if (*e == '%') {
					*p++ = '%';
					if (p > endof(buf) - 1)
						goto nobuf;
				}
				*p++ = *e;
			}
			fmt += 2;
		} else if (*fmt == '%' && *(fmt + 1) == '%') {
			if (p > endof(buf) - 2)
				goto nobuf;
			*p++ = *fmt++;
			*p++ = *fmt++;
		} else {
			if (p > endof(buf) - 1)
				goto nobuf;
			*p++ = *fmt++;
		}
	}
	if (*(p - 1) != '\n') {
		if (p > endof(buf) - 1)
			goto nobuf;
		*p++ = '\n';
	}
	if (p > endof(buf) - 1)
		goto nobuf;
	*p = '\0';
	fprintf(stderr, "%s:", getprogname());
	vfprintf(stderr, buf, ap);
	return;
nobuf:
	fprintf(stderr, "%s: compat_vsyslog: out of buffer\n", getprogname());
}

void
logmsg(int priority, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	compat_vsyslog(priority, fmt, ap);
	va_end(ap);
}
#endif

#ifdef DEBUG
#include <ctype.h>

void
logmsg2(int priority, const char *fmt, ...)
{
	va_list ap;

	if (option.debug_level < 2)
		return;

	va_start(ap, fmt);
	compat_vsyslog(priority, fmt, ap);
	va_end(ap);
}

void
dump2(const char *ptr, size_t len)
{
	char bufhex[56], *b, bufprint[17];
	const char *p;
	int i, thislen, c, used, blen;

	if (option.debug_level < 2)
		return;

	p = ptr;
	while (len > 0) {
		thislen = len < 16 ? len : 16;
		b = bufhex;
		blen = sizeof(bufhex);

		/* hex */
		for (i = 0; i < 16; i++) {
			used = i < thislen ?
			    snprintf(b, blen, " %02x", (unsigned char)p[i]) :
			    snprintf(b, blen, "   ");
			if (used < blen) {
				b += used;
				blen -= used;
			}
		}
		/* printable ASCII */
		for (i = 0; i < thislen; i++) {
			c = (unsigned char)p[i];
			bufprint[i] = isascii(c) && isprint(c) ? c : '.';
		}
		bufprint[i] = '\0';

		LDEBUG2("%p %s  %s", p, bufhex, bufprint);
		p += thislen;
		len -= thislen;
	}
}
#endif
